#ifndef PROTO_H
#define PROTO_H

#include "flags.h"
#include "../config.h"

#ifdef HAVE_PCRE
#include <pcre.h>
#include <pcreposix.h>
#else
#include "../pcre/pcre.h"
#include "../pcre/pcreposix.h"
#endif /* HAVE_PCRE */

#include <pthread.h>

#ifdef HAVE_STDINT_H
#include <stdint.h>
#else
#include <types.h>
#endif /* HAVE_STDINT_H */

#ifdef HAVE_POLL
#include <sys/poll.h>
#endif

#include "compat.h"
#include "settings.h"
#include "types.h"
#include "macros.h"

#undef strncpy
#define strncpy(a,b,c) dont_use_strncpy()
#undef strncat
#define strncat(a,b,c) dont_use_strncat()

/* main.c */
void config();
int config_load(int, char *);
int config_save(char *);
int process_new(CONNECTION *);
int thread_create(void *, void *);
void process_entry(CONNECTION *);
void mainloop();
void show_help(char **);
int pid_check(char *);
void bypass_check(CONNECTION *);

/* global.c */
void global_load(XML_LIST *xml_list);
XML_LIST *global_xml(XML_LIST *xml_list);
void global_general_insert(char *, char *, char *, char *, char *, char *, char *, char *, char *, char *);
void global_ftp_insert(char *, char *, char *, char *, char *);
void global_dnsbl_insert(char *, char *, char *);
void global_prefetch_insert(char *, char *, char *);

/* mem.c */
void *xmalloc(int size);
void *xrealloc(void *, int);
void xfree(void *);

/* xml.c */
XML_LIST *xml_load(XML_LIST *, char *);
XML_LIST *xml_list_add(XML_LIST *, char *, int);
void xml_list_free(XML_LIST *);
XML_LIST *xml_section(XML_LIST *, char *);
int xml_save(XML_LIST *, char *);

/* access.c */
ACCESS_LIST *access_load(ACCESS_LIST *, XML_LIST *);
void access_free(ACCESS_LIST *);
void access_ll_free(struct ACCESS_LIST_LIST *);
struct ACCESS_LIST_LIST *access_check(ACCESS_LIST *, CONNECTION *, char *, char *);
XML_LIST *access_xml(ACCESS_LIST *, XML_LIST *);
void access_ll_insert(struct ACCESS_LIST_LIST *, char *, char *, char *, char *, char *, char *, char *);
struct ACCESS_LIST_LIST *access_ll_new(struct ACCESS_LIST_LIST *);
struct ACCESS_LIST_LIST *access_ll_delete(struct ACCESS_LIST_LIST *);
void access_setup(CONNECTION *, struct ACCESS_LIST_LIST *);

/* header.c */
HEADER *header_new();
HEADER_LIST *header_load(HEADER_LIST *, XML_LIST *);
XML_LIST *header_xml(HEADER_LIST *, XML_LIST *);
void header_free(HEADER_LIST *);
void header_ll_free(struct HEADER_LIST_LIST *);
void header_ll_insert(struct HEADER_LIST_LIST *, char *, char *, char *, char *, char *);
struct HEADER_LIST_LIST *header_ll_new(struct HEADER_LIST_LIST *);
struct HEADER_LIST_LIST *header_ll_delete(struct HEADER_LIST_LIST *);
struct HTTP_HEADER_LIST *header_list_insert(struct HTTP_HEADER_LIST *, char *, char *);
struct HTTP_HEADER_LIST *header_list_dup(struct HTTP_HEADER_LIST *);
HEADER *http_header_parse_request(char *);
HEADER *http_header_parse_response(char *);
HEADER *http_header_dup(HEADER *);
int header_length(char *, size_t);
void http_header_list_parse(HEADER *, char *);
void http_header_free(HEADER *);
void http_header_list_free(struct HTTP_HEADER_LIST *);
char *header_extract(char *, int);
struct HTTP_HEADER_LIST *header_filter(HEADER_LIST *, CONNECTION *, int);
char *header_get(CONNECTION *, int, int);
void header_send(HEADER *, CONNECTION *, int, int);
void header_send_filebuf(FILEBUF *, HEADER *, CONNECTION *, int, int);
struct url_command_t **url_command_parse(char *);
char *url_command_create(struct url_command_t **);
void url_command_free(struct url_command_t **);
char *url_command_add(CONNECTION *, char *);
struct url_command_t *url_command_find(struct url_command_t **, char *);
void xcache_header_update(CONNECTION *, int flags);
void conditional_header(HEADER *, CACHEMAP *);
int url_command_extract(CONNECTION *);

/* filebuf.c */
void filebuf_free(FILEBUF *);
FILEBUF *filebuf_new();
FILEBUF *filebuf_dup(FILEBUF *);
void filebuf_add(FILEBUF *, const char *, int);
void filebuf_resize(FILEBUF *, int);
void filebuf_addf(FILEBUF *, const char *, ...);
void filebuf_replace(FILEBUF *, int, int, char *);
void filebuf_shorten(FILEBUF *);
int filebuf_compress(FILEBUF *, int);
int filebuf_uncompress(FILEBUF *, int);
int filebuf_save(FILEBUF *, char *);
int filebuf_read(FILEBUF *, char *);
STACK *filebuf_to_stack(FILEBUF *);
int filebuf_decode(CONNECTION *, FILEBUF *);
int filebuf_encode(CONNECTION *, FILEBUF *);
void zlib_error(int);

/* filter.c */
FILTER_LIST *filter_load(FILTER_LIST *, XML_LIST *);
void filter_free(FILTER_LIST *);
void filter_ll_free(struct FILTER_LIST_LIST *);
XML_LIST *filter_xml(FILTER_LIST *, XML_LIST *);
struct FILTER_LIST_LIST *filter_check(FILTER_LIST *, CONNECTION *);
FILEBUF *filter_get(FILTER_LIST *, char *);
void filter_ll_insert(struct FILTER_LIST_LIST *, char *, char *, char *, char *, char *);
struct FILTER_LIST_LIST *filter_ll_delete(struct FILTER_LIST_LIST *);
struct FILTER_LIST_LIST *filter_ll_new(struct FILTER_LIST_LIST *);

/* cookies.c */
COOKIE_LIST *cookie_load(COOKIE_LIST *, XML_LIST *);
XML_LIST *cookie_xml(COOKIE_LIST *, XML_LIST *);
void cookie_free(COOKIE_LIST *);
void cookie_ll_free(struct COOKIE_LIST_LIST *);
void cookie_ll_insert(struct COOKIE_LIST_LIST *, char *, char *, char *);
struct COOKIE_LIST_LIST *cookie_ll_new(struct COOKIE_LIST_LIST *);
struct COOKIE_LIST_LIST *cookie_ll_delete(struct COOKIE_LIST_LIST *);
int cookie_check(COOKIE_LIST *, int, CONNECTION *);

/* misc.c */
char *s_strncpy(char *, char *, size_t);
char *s_strncat(char *, char *, size_t);
char *s_strcasestr(char *, char *);
size_t s_strnlen(char *, size_t);
char *code_to_error(int);
char *xstrdup(char *);
char *xstrndup(char *, size_t);
char *error_to_template(int);
void hostent_free(HOSTENT *);
int isip(char *);
void string_tolower(char *);
char *string_to_html(char *, int);
char *url_encode(char *, size_t);
char *url_decode(char *, size_t);
char *string_to_xml(char *);
char **string_break(char *, char);
int range_check(char *, unsigned int);
void array_free(char **);
char *array_merge(char **, int);
int array_find(char **, char *);
int array_length(char **);
char **array_dup(char **);
char **array_add(char **, char *);
char **array_delete(char **, char *);
void stack_free(STACK *);
mode_t mode_parse(char *);
char *mode_create(mode_t);
struct sockaddr_in *ftpport_parse(char *);
char *ftpport_create(struct sockaddr_in *);
char *filesize(size_t);
time_t local_mktime(struct tm *);
time_t utc_time(time_t *);
char **cmdline_parse(char *);
char *string_append(char *, char *);
char *path_to_url(char *);
int in_absolutetimerange(struct tm *, struct tm *, struct tm *);
int in_alltimerange(struct tm *, struct tm *, struct tm *);
char *month_name(int);
char *weekday_name(int);
URL *url_parse(char *);
char *url_create(URL *);
void url_free(URL *);
int isempty(char *, size_t);
int urlcmp(URL *, URL *);
char *resolve_path(char *);
char *url_path_fix(char *);
void url_strip(char *);
int protocol_port(char *);
size_t string_to_size(char *);

/* base64.c */
void to64frombits(unsigned char *, const unsigned char *, int);
int from64tobits(char *, const char *);

/* pool.c */
POOL *pool_init();
void pool_add(POOL *, SOCKET *, char *, char *, int, char *, char *);
SOCKET *pool_find(POOL *, char *, char *, int, char *, char *);
void pool_clean(POOL *);
void pool_clean_unlocked(POOL *);

/* regexp.c */
regex_t *reg_compile(char *, int);
void reg_free(regex_t *);
int reg_exec(regex_t *, char *);
regex_sub_t *reg_sub_compile(char *, int);
regmatch_sub_t *reg_sub_exec(regex_sub_t *, char *);
char *reg_sub_parse(regmatch_sub_t *, char *);
void reg_sub_match_free(regmatch_sub_t *);
void reg_sub_free(regex_sub_t *);
void reg_replace(FILEBUF *, regex_sub_t *, char *);
void string_replace(FILEBUF *, char *, char *);

/* network.c */
void net_init();
CONNECTION *connection_new();
void connection_free(CONNECTION *);
int net_listen(int, char *);
CONNECTION *net_accept(int);
void net_close(CONNECTION *);
void net_listen_closeall();
int net_connect(char *, int, int);
int net_proxy(CONNECTION *, int);
int net_transfer(CONNECTION *, int, int);
NETWORK *network_load(NETWORK *, XML_LIST *);
XML_LIST *network_xml(NETWORK *, XML_LIST *);
void network_free(NETWORK *);
void listen_list_insert(struct LISTEN_LIST *, char *, char *, char *);
struct LISTEN_LIST *listen_list_delete(struct LISTEN_LIST *);
struct LISTEN_LIST *listen_list_new(struct LISTEN_LIST *);
void network_check(NETWORK *);
int net_filebuf_read(FILEBUF *, CONNECTION *, int, int);
int net_filebuf_send(FILEBUF *, CONNECTION *, int);
int net_cache_send(CACHEMAP *, CONNECTION *, int, int);
int proxy_socks4(CONNECTION *);
int putsock(SOCKET *, char *, ...);
HOSTENT *net_dns(char *);

/* socket.c */
SOCKET *sock_new(int);
void sock_close(SOCKET *);
int sock_read(SOCKET *, void *, int, int);
int sock_write(SOCKET *, void *, int);
int sock_getline(SOCKET *, char *, int, int);
int sock_pushfront(SOCKET *, char *, int);
void sock_flush(SOCKET *);
void sock_resize(SOCKET *, int, int);
void sock_clear(SOCKET *, int);

/* protocol.c */
int protocol_start(CONNECTION *);
int protocol_reconnect(CONNECTION *);
int protocol_http(CONNECTION *);
int protocol_ftp(CONNECTION *);
int protocol_connect(CONNECTION *);
unsigned int next_chunksize(CONNECTION *, int);
FILEBUF *http_transfer_filebuf(CONNECTION *, int, int);
void http_transfer_discard(CONNECTION *, int);
void http_transfer(CONNECTION *, int);
int proxy_authenticate(CONNECTION *);
int send_ntlm_response(CONNECTION *);
int send_basic_response(CONNECTION *);
void keepalive_check(CONNECTION *);
SOCKET *ftp_transfer_setup(CONNECTION *, char *, ...);
int ftp_transfer(CONNECTION *, int, SOCKET *, int);
FILEBUF *ftp_transfer_filebuf(CONNECTION *, int, SOCKET *, int);
int ftp_code_wait(CONNECTION *, int, int);
int buffer_check(CONNECTION *);
int transfer_limit_check(CONNECTION *);
int proxy_connect(CONNECTION *);
void proxy_test(CONNECTION *);
int accepts_encoded_content(CONNECTION *);

/* dirlist.c */
struct dir_t *dir_list_parse(FILEBUF *);
struct dir_t *dir_list_add(struct dir_t *, int, mode_t, char *, char *, size_t, struct tm, char *);
struct dir_t *dir_list_sort(struct dir_t *, int, int);
void dir_list_free(struct dir_t *);
struct dir_t *dir_list_filter(struct dir_t *, char *);

/* template.c */
TEMPLATES *templates_load(TEMPLATES *, XML_LIST *);
XML_LIST *templates_xml(TEMPLATES *, XML_LIST *);
void templates_free(TEMPLATES *);
void templates_list_free(struct TEMPLATE_LIST *);
void template_list_insert(struct TEMPLATE_LIST *, char *, char *, char *, char *, char *, char *, char *);
struct TEMPLATE_LIST *template_list_new(struct TEMPLATE_LIST *);
struct TEMPLATE_LIST *template_list_delete(struct TEMPLATE_LIST *);
FILEBUF *template_get(TEMPLATES *, CONNECTION *, char *);
int template_send(TEMPLATES *, char *, CONNECTION *, int);
void template_parse(CONNECTION *, FILEBUF *);

/* rewrite.c */
REWRITE_LIST *rewrite_load(REWRITE_LIST *, XML_LIST *);
XML_LIST *rewrite_xml(REWRITE_LIST *, XML_LIST *);
void rewrite_free(REWRITE_LIST *);
void rewrite_list_free(struct REWRITE_LIST_LIST *);
int rewrite_do(REWRITE_LIST *, CONNECTION *, FILEBUF *, int, int);
void rewrite_list_insert(struct REWRITE_LIST_LIST *, char *, char *, char *, char *, char *, char *, char *);
struct REWRITE_LIST_LIST *rewrite_list_delete(struct REWRITE_LIST_LIST *);
struct REWRITE_LIST_LIST *rewrite_list_new(struct REWRITE_LIST_LIST *);

/* profiles.c */
PROFILES_LIST *profiles_load(PROFILES_LIST *, XML_LIST *);
XML_LIST *profiles_xml(PROFILES_LIST *, XML_LIST *);
void profiles_free(PROFILES_LIST *);
void profiles_list_free(struct PROFILES_LIST_LIST *);
void profiles_update(PROFILES_LIST *, CONNECTION *);
void profiles_list_insert(struct PROFILES_LIST_LIST *, char *, char *, char *, char *, char *, char *, char *, char *);
struct PROFILES_LIST_LIST *profiles_list_delete(struct PROFILES_LIST_LIST *);
struct PROFILES_LIST_LIST *profiles_list_new(struct PROFILES_LIST_LIST *);
int profile_find(char **, char **);

/* log.c */
void putlog(int, char *, ...);
void logbuffer_add(LOGBUFFER *, char *);
void logbuffer_clear(LOGBUFFER *);
void logbuffer_resize(LOGBUFFER *, int);

/* signal.c */
void signal_setup();

/* compat.c */
HOSTENT *p_gethostbyname(char *);
int p_poll(struct pollfd *, int, int);
int p_setenv(char *, char *, int);
void p_clearenv();
void *p_mremap(void *, size_t, size_t, int);

/* external.c */
EXTERNAL *external_load(EXTERNAL *, XML_LIST *);
XML_LIST *external_xml(EXTERNAL *, XML_LIST *);
void external_free(EXTERNAL *);
void external_list_free(struct EXTERNAL_LIST_LIST *);
void external_list_insert(struct EXTERNAL_LIST_LIST *, char *, char *, char *, char *, char *, char *, char *);
struct EXTERNAL_LIST_LIST *external_list_delete(struct EXTERNAL_LIST_LIST *);
struct EXTERNAL_LIST_LIST *external_list_new(struct EXTERNAL_LIST_LIST *);
struct EXTERNAL_LIST_LIST *external_find(EXTERNAL *, CONNECTION *);
FILEBUF *external_exec(CONNECTION *, char *, FILEBUF *, int);
char *external_getmime(FILEBUF *);
char *env_header_format(char *, char *);

/* mime.c */
MIME_LIST *mime_load(MIME_LIST *, XML_LIST *);
XML_LIST *mime_xml(MIME_LIST *, XML_LIST *);
void mime_free(MIME_LIST *);
void mime_ll_free(struct MIME_LIST_LIST *);
void mime_ll_insert(struct MIME_LIST_LIST *, char *, char *, char *, char *, char *, char *);
struct MIME_LIST_LIST *mime_ll_delete(struct MIME_LIST_LIST *);
struct MIME_LIST_LIST *mime_ll_new(struct MIME_LIST_LIST *);
struct MIME_LIST_LIST *mime_check(MIME_LIST *, CONNECTION *);

/* hash.c */
unsigned int hash_key(int, char *);
HASH_TABLE *hash_create(int);
struct HASH_LIST *hash_insert(HASH_TABLE *, char *, void *);
void hash_purge(HASH_TABLE *, char *);
struct HASH_LIST *hash_search(HASH_TABLE *, char *);
void hash_destroy(HASH_TABLE *);
int hash_expire(HASH_TABLE *, unsigned int);

/* redirect.c */
REDIRECT_LIST *redirect_load(REDIRECT_LIST *, XML_LIST *);
XML_LIST *redirect_xml(REDIRECT_LIST *, XML_LIST *);
struct REDIRECT_LIST_LIST *redirect_list_new(struct REDIRECT_LIST_LIST *);
void redirect_list_insert(struct REDIRECT_LIST_LIST *, char *, char *, char *, char *, char *, char *, char *, char *);
void redirect_free(REDIRECT_LIST *);
void redirect_list_free(struct REDIRECT_LIST_LIST *);
struct REDIRECT_LIST_LIST *redirect_list_delete(struct REDIRECT_LIST_LIST *);
int redirect_do(REDIRECT_LIST *, CONNECTION *, int);

/* forward.c */
FORWARD_LIST *forward_load(FORWARD_LIST *, XML_LIST *);
XML_LIST *forward_xml(FORWARD_LIST *, XML_LIST *);
struct FORWARD_LIST_LIST *forward_list_new(struct FORWARD_LIST_LIST *);
void forward_list_insert(struct FORWARD_LIST_LIST *, char *, char *, char *, char *, char *, char *, char *, char *, char *, char *, char *);
struct FORWARD_LIST_LIST *forward_list_delete(struct FORWARD_LIST_LIST *);
void forward_list_free(struct FORWARD_LIST_LIST *);
void forward_free(FORWARD_LIST *);
int forward_do(FORWARD_LIST *, CONNECTION *);

/* keywords.c */
KEYWORD_LIST *keyword_load(KEYWORD_LIST *, XML_LIST *);
XML_LIST *keyword_xml(KEYWORD_LIST *, XML_LIST *);
struct KEYWORD_LIST_LIST *keyword_list_new(struct KEYWORD_LIST_LIST *);
void keyword_list_insert(struct KEYWORD_LIST_LIST *, char *, char *, char *, char *, char *, char *);
struct KEYWORD_LIST_LIST *keyword_list_delete(struct KEYWORD_LIST_LIST *);
void keyword_list_free(struct KEYWORD_LIST_LIST *);
void keyword_free(KEYWORD_LIST *);
int keyword_check(KEYWORD_LIST *, CONNECTION *, FILEBUF *, int);

/* cache.c */
struct cache_store *cache_load(struct cache_store *, XML_LIST *);
struct cache_store *stores_list_new(struct cache_store *);
void global_cache_insert(char *, char *, char *, char *, char *, char *, char *, char *, char *, char *, char *, char *, char *);
void stores_list_insert(struct cache_store *, char *, char *, char *, char *, char *);
void select_store(CONNECTION *, CACHE *, CACHEMAP *);
struct cache_store *stores_list_delete(struct cache_store *);
void stores_list_free(struct cache_store *);
XML_LIST *cache_xml(struct cache_store *, XML_LIST *);
void stores_update(CACHE *);

CACHE *cache_init();
CACHEMAP *cache_hash_find(CACHE *, char *);
int cache_filename(CACHE *, CACHEMAP *, char *, int);
void cache_hash_add(CACHE *, CACHEMAP *);
void cache_hash_remove(CACHE *, CACHEMAP *);
CACHEMAP *cache_open(CONNECTION *connection, CACHE *, char *, char *, int);
void cache_close(CACHE *, CACHEMAP *);
void cache_close_unlocked(CACHE *, CACHEMAP *);
int cache_add_unlocked(CACHE *, CACHEMAP *, void *, size_t);
int cache_add(CACHE *, CACHEMAP *, void *, size_t);
void cache_scandisk(CACHE *, struct cache_store *);
void cache_clean(CACHE *);
char *cache_header(CACHEMAP *);
void cache_check(CACHE *, HEADER *);
FILEBUF *cache_to_filebuf(CACHEMAP *, int);
void cache_invalidate(CACHE *, CACHEMAP *);
void cache_invalidate_unlocked(CACHE *, CACHEMAP *);
void cache_disk_delete(CACHE *, CACHEMAP *);
void cache_mem_delete(CACHE *, CACHEMAP *);
void cache_free(CACHE *, CACHEMAP *);
int cache_flag_set(CACHE *, CACHEMAP *, int);
int cache_flag_unset(CACHE *, CACHEMAP *, int);
void cache_touch(CACHE *, CACHEMAP *, int);
void cache_touch_unlocked(CACHE *, CACHEMAP *, int);
void cache_check_header(CACHE *, CACHEMAP *);
int cacheable(HEADER *);
void cache_validated(CACHE *, CACHEMAP *);
int cache_invalidate_key(CACHE *, char *);
size_t cache_resize(CACHE *, CACHEMAP *, size_t);
int cache_exists(CACHE *, char *);
void cache_journal_write(CACHE *, struct cache_store *);
void cache_journal_read(CACHE *, struct cache_store *);
int cache_journal_add(CACHE *, CACHEMAP *);
CACHEMAP *cachemap_new(CACHE *);
void unlink_list_add(char *);
void unlink_thread();

/* timers.c */
TIMERS *timers_init();
int timers_add(TIMERS *, int interval, int flags, void *callback, void *arg);
int timers_remove(TIMERS *, int tid);
void timers_check(TIMERS *);
struct TIMERS_LIST *timers_find(TIMERS *, int tid);
void timers_insert(TIMERS *, struct TIMERS_LIST *);


/* limits.c */
LIMITS_LIST *limits_load(LIMITS_LIST *, XML_LIST *);
XML_LIST *limits_xml(LIMITS_LIST *, XML_LIST *);
void limits_ll_insert(struct LIMITS_LIST_LIST *x, char *, char *, char *, struct tm *, struct tm *, char *, char *, char *);
struct LIMITS_LIST_LIST *limits_ll_new(struct LIMITS_LIST_LIST *);
struct LIMITS_LIST_LIST *limits_ll_delete(struct LIMITS_LIST_LIST *);
void limits_ll_free(struct LIMITS_LIST_LIST *);
void limits_free(LIMITS_LIST *);
struct LIMITS_LIST_LIST *limits_check(LIMITS_LIST *, CONNECTION *);
void limits_update(LIMITS_LIST *, CONNECTION *);
void limits_reset(LIMITS_LIST *);

/* prefetch.c */
PREFETCH *prefetch_init();
void prefetch_init_threads(PREFETCH *);
void prefetch_thread(PREFETCH *);
int protocol_prefetch(URL *, int);
void prefetch_queue_add(PREFETCH *, URL *, int);
void prefetch_queue_free(struct PREFETCH_QUEUE *);
void prefetch_setup_callbacks(CONNECTION *, PREFETCH *);
void prefetch_callback(HTMLSTREAM *, struct htmlstream_node *, CONNECTION *);

/* hstream.c */
HTMLSTREAM *htmlstream_new();
void htmlstream_free(HTMLSTREAM *);
void htmlstream_tree_free(struct htmlstream_node *);
void htmlstream_tag_free(struct htmlstream_tag *);
void htmlstream_add(HTMLSTREAM *, char *, int);
struct htmlstream_tag *htmlstream_tag_parse(char *, int);
int htmlstream_callback_add(HTMLSTREAM *, char *, void *, void *);

/* icp.c */
struct proxy_candidate *plist_add(struct proxy_candidate *, struct FORWARD_LIST_LIST *);
void plist_free(struct proxy_candidate *);
struct FORWARD_LIST_LIST *proxy_select(CONNECTION *, struct proxy_candidate *);
struct proxy_candidate *plist_remove(struct proxy_candidate *, struct proxy_candidate *);
int icp_packet_send(int fd, struct sockaddr_in *, struct icp_packet *);
int icp_packet_receive(int fd, struct sockaddr_in *, struct icp_packet *);
char *icp_opcode_name(int);
void icp_thread();

#ifdef HAVE_PAM
int auth_pam(char *, char *);
#endif /* HAVE_PAM */

/* interface.c */
struct cgi_args_t *cgi_parse_request(char *);
void cgi_args_free(struct cgi_args_t *);
int interface_check_referer(CONNECTION *);
void interface_handle_request(CONNECTION *);
void interface_page_connections(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_dnscache(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_headers(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_save(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_load(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_log(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_cache(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_pool(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_prefetch(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_general(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_general_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_ftp(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_ftp_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_dnsbl(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_dnsbl_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_prefetch(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_global_prefetch_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_profiles(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_profiles_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_profiles_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_cache(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_cache_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_cache_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_filter(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_filter_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_filter_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_header(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_header_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_header_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_access(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_access_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_access_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_rewrite(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_rewrite_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_rewrite_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_mime(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_mime_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_mime_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_cookies(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_cookies_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_cookies_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_external(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_external_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_external_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_templates(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_templates_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_templates_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_network(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_network_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_network_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_redirect(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_redirect_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_redirect_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_forward(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_forward_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_forward_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_keywords(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_keywords_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_keywords_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_limits(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_limits_dialog(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void interface_page_config_limits_action(FILEBUF *, struct cgi_args_t *, CONNECTION *);
void filter_check_show(CONNECTION *);
void mime_check_show(CONNECTION *);
void score_show(CONNECTION *, int);
void cache_show(CONNECTION *);
void show_filebuf_diff(CONNECTION *, FILEBUF *, FILEBUF *);
void show_dirlist(CONNECTION *, struct dir_t *);
void show_headers(CONNECTION *);
void show_filebuf(CONNECTION *, FILEBUF *);
void show_proxytest(CONNECTION *);
void show_htmltree(CONNECTION *);
void show_htmltree_node(FILEBUF *, struct htmlstream_node *, int);
void show_profiles(CONNECTION *);

#endif				/* PROTO_H */
