#ifndef TYPES_H
#define TYPES_H

struct unlink_list {
	struct unlink_list *next;
	char *path;
};

typedef struct {
	int code;
	char *type;
	unsigned int size, realsize;
	char *data;
} FILEBUF;

struct HASH_LIST {
	struct HASH_LIST *next;
	time_t age;
	char *ref;
	void *data;
	int flags;
};

typedef struct {
	int hit;
	int miss;
	int size;
	struct HASH_LIST **hash_list;
} HASH_TABLE;

struct htmlstream_tag_property {
	struct htmlstream_tag_property *next, *prev;
	char *name;
	char *value;
};

struct htmlstream_tag {
	char *name;
	struct htmlstream_tag_property *properties;
};

struct htmlstream_node {
	struct htmlstream_node *down, *prev, *next, *up;
	int level;
	char *text;
	struct htmlstream_tag *tag;
};

typedef struct {
	HASH_TABLE *callbacks;
	struct htmlstream_node *tree;
	struct htmlstream_node *tail;
	int startpos, incomment;
	FILEBUF *buffer;
} HTMLSTREAM;

struct htmlstream_callback {
	void (*func)(HTMLSTREAM *, struct htmlstream_node *, void *);
	void *arg;
};

typedef struct {
	int fd, frozen;
	int inbuf_len, outbuf_len;
	int inbuf_reallen, outbuf_reallen;
	void *inbuf, *outbuf;
} SOCKET;

typedef struct {
	char *username;
	char *password;
	char *proto;
	char *host;
	char *file;
	int port;
} URL;

struct url_command_t {
	char *command;
	char *options;
};

typedef struct {
	pcre *pattern;
	pcre_extra *extra;
} regex_sub_t;

typedef struct {
	int rm_so, rm_eo, subcount;
	const char **substrings;
} regmatch_sub_t;

typedef struct XML_LIST_T {
	struct XML_LIST_T *prev, *next;
	int type;
	char *item;
} XML_LIST;

typedef struct STACK_T {
	struct STACK_T *next, *prev;
	char *data;
} STACK;

struct HTTP_HEADER_LIST {
	struct HTTP_HEADER_LIST *next, *prev;
	int send;
	char *type;
	char *value;
};

struct LOGBUFFER_LIST {
	struct LOGBUFFER_LIST *next, *prev;
	char *msg;
};

typedef struct {
	pthread_rwlock_t lock;
	int entries, size;
	struct LOGBUFFER_LIST *head, *tail;
} LOGBUFFER;	

typedef struct {
	char *raw;
	int type;
	int version;
	int port;
	int code;
	int content_length;
	int chunked;
	int keepalive;
	int proxy_keepalive;
	char *username;
	char *password;
	char *method;
	char *proto;
	char *host;
	char *file;
	char *referer;
	char *location;
	char *content_type;
	char *content_encoding;
	char *accept_encoding;
	char *authorization;
	char *range;
	char *proxy_authenticate;
	char *proxy_authorization;
	char *host_header;
	char *cache_control;
	char *expires;
	char *if_modified_since;
	char *last_modified;
	char *warning;
	char *age;
	char *x_cache;
	struct HTTP_HEADER_LIST *header;
	struct HTTP_HEADER_LIST *header_filtered;
	int flags;
} HEADER;

struct cache_store {
	struct cache_store *prev, *next;
	int id, enabled;
	char *path, *comment;
	int journalfd, journalentries, diskentries;
	size_t disksize, maxdisksize, diskextra;
	char **profiles;
};

typedef struct {
	pthread_cond_t writecompletecond;
	int id, fd;
        char *key;
        void *data;
        int refcount;
        size_t size, realsize;
	size_t offset;
	int flags;
	HEADER *header;
	struct cache_store *store;
	time_t atime, mtime, etime, ftime, lmtime;
} CACHEMAP;

struct CACHEMAPLIST_T {
        struct CACHEMAPLIST_T *prev, *next;
        CACHEMAP *cachemap;
};

typedef struct {
	pthread_mutex_t lock;
	int enabled, id;
	int violaterfc;
	int hashsize;
	int sbalancemethod;
        unsigned int memsize;
	unsigned int maxmemsize;
	unsigned int memextra;
	unsigned int maxsize, minsize;
	unsigned int maxwaitsize;
	int maxage, minage;
	int validate;
	int lmfactor;
	int mementries;
	int hit, miss;
	int prefetchwindow;
	time_t lastclean;
	int icpport, icptimeout;
	struct cache_store *stores;
        struct CACHEMAPLIST_T **cachemap;
} CACHE;


typedef struct {
	int status;
	int fd;
	int port;
	char *ip;
} SOCKLIST;

typedef struct {
	int thread;
	int flags;
	SOCKET *client, *server;
	int depth;
	char *ip;
	int port;
	char *interface;
	int request;
	int keepalive_client, keepalive_server;
	int authenticate;
	size_t transferred, transferlimit;
	char *proxy_auth;
	char *proxy_host;
	int proxy_port;
	int proxy_type;
	char *proxy_username;
	char *proxy_password;
	char *proxy_domain;
	int access;
	int bypass;
	struct url_command_t **url_command;
	char **profiles;
	CACHEMAP *cachemap;
	HEADER *header, *rheader;
	HTMLSTREAM *htmlstream;
} CONNECTION;

struct COOKIE_LIST_LIST {
	struct COOKIE_LIST_LIST *prev, *next;
	int id, enabled;
	int direction;
	char *comment;
	char **profiles;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	int policy;
	struct COOKIE_LIST_LIST *allow, *deny;
} COOKIE_LIST;

struct FILTER_LIST_LIST {
	struct FILTER_LIST_LIST *prev, *next;
	int id, enabled;
	char *comment, *template, *host, *file;
	char **profiles;
	regex_t *he, *fe;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	int policy;
	char *dtemplate;
	struct FILTER_LIST_LIST *allow, *deny, *empty;
} FILTER_LIST;

struct KEYWORD_LIST_LIST {
	struct KEYWORD_LIST_LIST *prev, *next;
	int id, enabled;
	int score;
	int size;
	char *comment, *mime, *keyword;
	regex_t *me, *ke;
	char **profiles;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	int threshold;
	char *template;
	struct KEYWORD_LIST_LIST *keyword_list;
} KEYWORD_LIST;

struct REDIRECT_LIST_LIST {
	struct REDIRECT_LIST_LIST *prev, *next;
	int id, enabled;
	int which, port, send302, options;
	regex_sub_t *up;
	char *comment, *redirect, *url;
	char **profiles;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	struct REDIRECT_LIST_LIST *redirect_list;
} REDIRECT_LIST;

struct FORWARD_LIST_LIST {
	struct FORWARD_LIST_LIST *prev, *next;
	int id, enabled;
	int which, port, type, icpport, icptype;
	char *comment, *proxy, *username, *password, *domain;
	char **profiles;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	struct FORWARD_LIST_LIST *forward_list;
} FORWARD_LIST;

struct MIME_LIST_LIST {
	struct MIME_LIST_LIST *prev, *next;
	int id, enabled;
	char *comment, *template, *mime, *host, *file;
	char **profiles;
	regex_t *me, *he, *fe;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	int policy;
	char *dtemplate;
	struct MIME_LIST_LIST *allow, *deny, *empty;
} MIME_LIST;

struct HEADER_LIST_LIST {
	struct HEADER_LIST_LIST *prev, *next;
	int id, enabled;
	char *comment, *type, *value;
	char **profiles;
	int which;
	regex_t *te, *ve;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	int policy;
	struct HEADER_LIST_LIST *allow, *deny, *insert;
} HEADER_LIST;

struct ACCESS_LIST_LIST {
	struct ACCESS_LIST_LIST *prev, *next;
	int id, enabled;
	int pamauth;
	int access, bypass;
	char *comment, *ip, *username, *password;
	char **profiles;
	regex_t *ie, *ue;
};

typedef struct {
	pthread_rwlock_t lock;
	int id;
	int policy;
	struct ACCESS_LIST_LIST *allow, *deny, *empty;
} ACCESS_LIST;

typedef struct {
	pthread_rwlock_t lock;
	int id;
	struct LISTEN_LIST *listen_list;
} NETWORK;

struct LISTEN_LIST {
	struct LISTEN_LIST *prev, *next;
	int id, enabled;
	int port;
	char *comment, *ip;
};

struct TEMPLATE_LIST {
	struct TEMPLATE_LIST *prev, *next;
	int id, enabled;
	int type, code, parsable;
	char *comment, *name, *file, *mime;
	char **profiles;
};

typedef struct {
	pthread_rwlock_t lock;
	int id;
	char *path;
	struct TEMPLATE_LIST *template_list;
} TEMPLATES;

struct EXTERNAL_LIST_LIST {
	struct EXTERNAL_LIST_LIST *prev, *next;
	int id, enabled;
	regex_t *me;
	int type;
	int size;
	char *comment, *exec, *newmime, *mime;
	char **profiles;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	struct EXTERNAL_LIST_LIST *external_list;
} EXTERNAL;

struct REWRITE_LIST_LIST {
	struct REWRITE_LIST_LIST *prev, *next;
	int id, enabled;
	int which;
	int size;
	regex_t *me;
	regex_sub_t *pe;
	char *comment, *pattern, *replace, *mime;
	char **profiles;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	struct REWRITE_LIST_LIST *rewrite;
} REWRITE_LIST;

struct PROFILES_LIST_LIST {
	struct PROFILES_LIST_LIST *prev, *next;
	int id, enabled;
	regex_t *he, *fe;
	char *comment, *host, *file, *portrange;
	char **profiles, **addprofiles, **removeprofiles, **urlcommand;
};

typedef struct {
	pthread_rwlock_t lock;
	int id;
	struct PROFILES_LIST_LIST *profiles;
} PROFILES_LIST;

struct LIMITS_LIST_LIST {
	struct LIMITS_LIST_LIST *prev, *next;
	int action, id, enabled, flags;
	char **profiles;
	char *comment, *template;
	struct tm *from, *to;
	unsigned int maxrequests;
	size_t maxbytes;
	unsigned int n_requests;
	size_t n_bytes;
};

typedef struct {
	pthread_rwlock_t lock;
	int id, enabled;
	struct LIMITS_LIST_LIST *limits_list;
} LIMITS_LIST;

typedef struct {
	struct in_addr *addr;
	int len;
} HOSTENT;

typedef struct {
	pthread_mutex_t lock;
	pthread_cond_t cond;
	void *arg;
	int flags, pid;
	int requests;
	char *ip, *host, *file, *method;
	int port;
} THREADLIST;
	
struct cgi_args_t {
	struct cgi_args_t *next;
	char *name;
	char *value;
};
 
typedef struct {
	pthread_rwlock_t lock;
	int ctimeout, timeout;
	int keeptimeout;
	int poolsize, pooltimeout;
	char *tempdir;
	int maxbuffer;
	int linelength;
	char *cportrange;
	int compressout, compressin;
	char *encodemime;
	regex_t *emp;
} GENERAL;

typedef struct {
	pthread_rwlock_t lock;
	int pasv, timeout, sortorder, sortfield;
	char *anonlogin;
	char *anonpass;
} FTP;

struct pool_t {
	struct pool_t *next, *prev;
	SOCKET *sock;
	time_t age;
	char *proto;
	char *host;
	char *username, *password;
	int port;
};

typedef struct {
	int entries;
	pthread_mutex_t lock;
	struct pool_t *pool_list;
} POOL;

typedef struct {
	pthread_rwlock_t lock;
	int enabled;
	char *template;
	char *domain;
	char **badips;
} DNSBL;

struct TIMERS_LIST {
	struct TIMERS_LIST *prev, *next;
	int tid;
	int interval;
	time_t etime;
	int flags;
	void (*callback)(void *);
	void *arg;
};

typedef struct {
	pthread_mutex_t lock;
	int tid;
	struct TIMERS_LIST *timers;
} TIMERS;

struct PREFETCH_QUEUE {
	struct PREFETCH_QUEUE *next;
	int depth;
	URL *url;
};	

typedef struct {
	int enabled;
	pthread_mutex_t lock;
	pthread_cond_t cond;
	int threads, queuesize, entries, maxdepth;
	char **tags;
	struct PREFETCH_QUEUE *queue;
} PREFETCH;

typedef struct {
	GENERAL *general;
	FTP *ftp;
	CACHE *cache;
	LOGBUFFER *logbuffer;
	HASH_TABLE *dns_cache;
	POOL *pool;
	TIMERS *timers;
	DNSBL *dnsbl;
	PREFETCH *prefetch;
	struct unlink_list *unlink_head, *unlink_tail;
	pthread_mutex_t unlink_lock;
	pthread_cond_t unlink_cond;
	pthread_mutex_t dns_cache_lock;
} GLOBAL;

struct dir_t {
	struct dir_t *next, *prev;
	int isdir;
	mode_t mode;
	char *user;
	char *group;
	size_t size;
	struct tm time;
	time_t t;
	char *name;
};

struct icp_packet {
        unsigned char opcode;
        unsigned char version;
        uint16_t mlen;
        uint32_t rnum;
        uint32_t options;
        uint32_t optiondata;
        uint32_t hostaddr;
        unsigned char payload[ICP_MESSAGE_MAX];
};

struct proxy_candidate {
	struct proxy_candidate *prev, *next;
	unsigned int reqnum;
	struct icp_packet icppacket;
	struct FORWARD_LIST_LIST *proxy;
};

#endif				/* TYPES_H */
