/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/time.h>
#include "proto.h"

extern long gmtoffset;
extern GLOBAL *global;

/* hold lock when:
	calling:
		cache_hash_add
		cache_hash_remove
		cache_clean
		cache_filename
		cache_disk_delete
		cache_mem_delete
		cache_journal_*
		*_unlocked
		select_store

	when accessing any members of a cachemap struct besides key, data, offset, and size.

	no other threads may change any cachemap members when it's open for writing.

	a thread with a cachemap open for writing may not change any member besides data without
	the mutex locked.

	cache->stores is protected with cache->lock
*/

struct cache_store *cache_load(struct cache_store *s, XML_LIST * xml_list)
{
	struct cache_store *head = s, *stores = s;

	while ((xml_list = xml_section(xml_list, "<cache>"))) {
		XML_LIST_LOOP(xml_list, "<cache>") {
			XML_LIST_CMP(xml_list, "<enabled>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "false"))
						global->cache->enabled = FALSE;
					else
						global->cache->enabled = TRUE;
				}
			}
			XML_LIST_CMP(xml_list, "<violaterfc>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "false"))
						global->cache->violaterfc = FALSE;
					else
						global->cache->violaterfc = TRUE;
				}
			}
			XML_LIST_CMP(xml_list, "<maxmemsize>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<memextra>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<maxage>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<validate>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<lmfactor>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<maxsize>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<minage>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<minsize>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<maxwaitsize>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<prefetchwindow>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<icpport>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<icptimeout>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL);

			}
			XML_LIST_CMP(xml_list, "<sbalancemethod>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item);

			}

			XML_LIST_CMP(xml_list, "<store>") {
				stores = stores_list_new(stores);
				if (head == NULL)
					head = stores;

				stores->id = global->cache->id++;

				XML_LIST_LOOP(xml_list, "<store>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								stores->enabled = FALSE;
							else
								stores->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							stores_list_insert(stores, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							stores_list_insert(stores, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<path>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							stores_list_insert(stores, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<maxdisksize>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							stores_list_insert(stores, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<diskextra>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							stores_list_insert(stores, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}
		}
	}

	return head;
}

XML_LIST *cache_xml(struct cache_store * stores, XML_LIST * xml_list)
{
	char buf[16], *ptr;
	struct cache_store *sl;

	pthread_mutex_lock(&global->cache->lock);

	xml_list = xml_list_add(xml_list, "<cache>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
	xml_list = xml_list_add(xml_list, (global->cache->enabled == TRUE) ? "true" : "false", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<violaterfc>", XML_TAG);
	xml_list = xml_list_add(xml_list, (global->cache->violaterfc == FALSE) ? "false" : "true", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</violaterfc>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<maxmemsize>", XML_TAG);
	snprintf(buf, sizeof(buf), "%u", global->cache->maxmemsize);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</maxmemsize>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<memextra>", XML_TAG);
	snprintf(buf, sizeof(buf), "%u", global->cache->memextra);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</memextra>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<maxage>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", global->cache->maxage);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</maxage>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<validate>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", global->cache->validate);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</validate>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<lmfactor>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", global->cache->lmfactor);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</lmfactor>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<maxsize>", XML_TAG);
	snprintf(buf, sizeof(buf), "%u", global->cache->maxsize);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</maxsize>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<minage>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", global->cache->minage);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</minage>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<minsize>", XML_TAG);
	snprintf(buf, sizeof(buf), "%u", global->cache->minsize);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</minsize>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<maxwaitsize>", XML_TAG);
	snprintf(buf, sizeof(buf), "%u", global->cache->maxwaitsize);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</maxwaitsize>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<prefetchwindow>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", global->cache->prefetchwindow);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</prefetchwindow>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<icpport>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", global->cache->icpport);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</icpport>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<icptimeout>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", global->cache->icptimeout);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</icptimeout>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<sbalancemethod>", XML_TAG);
	xml_list = xml_list_add(xml_list, (global->cache->sbalancemethod == SBALANCE_FILLSIZE) ? "fillsize" : "fillpercent", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</sbalancemethod>", XML_TAG);

	for (sl = global->cache->stores; sl; sl = sl->next) {
		xml_list = xml_list_add(xml_list, "<store>", XML_TAG);

		xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
		xml_list = xml_list_add(xml_list, (sl->enabled == TRUE) ? "true" : "false", XML_VALUE);
		xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

		if (sl->comment != NULL) {
			xml_list = xml_list_add(xml_list, "<comment>", XML_TAG);
			ptr = string_to_xml(sl->comment);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</comment>", XML_TAG);
		}

		if (sl->profiles != NULL) {
			xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);
			ptr = array_merge(sl->profiles, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);
		}

		if (sl->path != NULL) {
			xml_list = xml_list_add(xml_list, "<path>", XML_TAG);
			ptr = string_to_xml(sl->path);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</path>", XML_TAG);
		}

		xml_list = xml_list_add(xml_list, "<maxdisksize>", XML_TAG);
		snprintf(buf, sizeof(buf), "%u", sl->maxdisksize);
		xml_list = xml_list_add(xml_list, buf, XML_VALUE);
		xml_list = xml_list_add(xml_list, "</maxdisksize>", XML_TAG);

		xml_list = xml_list_add(xml_list, "<diskextra>", XML_TAG);
		snprintf(buf, sizeof(buf), "%u", sl->diskextra);
		xml_list = xml_list_add(xml_list, buf, XML_VALUE);
		xml_list = xml_list_add(xml_list, "</diskextra>", XML_TAG);

		xml_list = xml_list_add(xml_list, "</store>", XML_TAG);
	}

	xml_list = xml_list_add(xml_list, "</cache>", XML_TAG);

	pthread_mutex_unlock(&global->cache->lock);

	return xml_list;
}

void global_cache_insert(char *maxmemsize, char *memextra, char *maxage, char *validate, char *lmfactor, char *maxsize, char *minage, char *minsize, char *maxwaitsize, char *prefetchwindow, char *icpport, char *icptimeout, char *sbalancemethod)
{
	if (maxmemsize != NULL) {
		global->cache->maxmemsize = 0;
		if (strcmp(maxmemsize, ""))
			global->cache->maxmemsize = string_to_size(maxmemsize);
	}

	if (memextra != NULL) {
		global->cache->memextra = 0;
		if (strcmp(memextra, ""))
			global->cache->memextra = string_to_size(memextra);
	}

	if (maxage != NULL) {
		global->cache->maxage = 0;

		if (strcmp(maxage, ""))
			global->cache->maxage = atoi(maxage);
	}

	if (validate != NULL) {
		global->cache->validate = 0;

		if (strcmp(validate, ""))
			global->cache->validate = atoi(validate);
	}

	if (lmfactor != NULL) {
		global->cache->lmfactor = 0;

		if (strcmp(lmfactor, ""))
			global->cache->lmfactor = atoi(lmfactor);
	}

	if (maxsize != NULL) {
		global->cache->maxsize = 0;

		if (strcmp(maxsize, ""))
			global->cache->maxsize = string_to_size(maxsize);
	}

	if (minage != NULL) {
		global->cache->minage = 0;

		if (strcmp(minage, ""))
			global->cache->minage = atoi(minage);
	}

	if (minsize != NULL) {
		global->cache->minsize = 0;

		if (strcmp(minsize, ""))
			global->cache->minsize = string_to_size(minsize);
	}

	if (maxwaitsize != NULL) {
		global->cache->maxwaitsize = 0;

		if (strcmp(maxwaitsize, ""))
			global->cache->maxwaitsize = string_to_size(maxwaitsize);
	}

	if (prefetchwindow != NULL) {
		global->cache->prefetchwindow = 10;

		if (strcmp(prefetchwindow, ""))
			global->cache->prefetchwindow = atoi(prefetchwindow);
	}

	if (icpport != NULL) {
		global->cache->icpport = 0;

		if (strcmp(icpport, ""))
			global->cache->icpport = atoi(icpport);
	}

	if (icptimeout != NULL) {
		global->cache->icptimeout = ICP_TIMEOUT;

		if (strcmp(icptimeout, ""))
			global->cache->icptimeout = atoi(icptimeout);
	}

	if (sbalancemethod != NULL) {
		global->cache->sbalancemethod = SBALANCE_FILLSIZE;

		if (!strcasecmp(sbalancemethod, "fillsize"))
			global->cache->sbalancemethod = SBALANCE_FILLSIZE;
		else if (!strcasecmp(sbalancemethod, "fillpercent"))
			global->cache->sbalancemethod = SBALANCE_FILLPERCENT;
	}
}

struct cache_store *stores_list_new(struct cache_store *x)
{

	if (x == NULL) {
		x = xmalloc(sizeof(struct cache_store));
		x->prev = NULL;
	} else {
		while (x->next != NULL)
			x = x->next;
		x->next = xmalloc(sizeof(struct cache_store));
		x->next->prev = x;
		x = x->next;
	}

	x->enabled = TRUE;
	x->next = NULL;
	x->comment = NULL;
	x->profiles = NULL;
	x->path = NULL;
	x->journalentries = 0;
	x->journalfd = -1;
	x->diskentries = 0;
	x->disksize = x->maxdisksize = x->diskextra = 0;


	return x;
}

struct cache_store *stores_list_delete(struct cache_store *x)
{
	struct cache_store *start = x;

	while (start->prev != NULL)
		start = start->prev;

	if (x->next != NULL)
		x->next->prev = x->prev;
	if (x->prev != NULL)
		x->prev->next = x->next;
	else
		start = x->next;

	if (x->profiles != NULL)
		array_free(x->profiles);
	FREE_AND_NULL(x->comment);
	FREE_AND_NULL(x->path);

	xfree(x);

	return start;
}

void stores_list_free(struct cache_store *x)
{
	struct cache_store *tmp;

	while (x != NULL) {
		tmp = x->next;

		if (x->profiles != NULL)
			array_free(x->profiles);
		FREE_AND_NULL(x->comment);
		FREE_AND_NULL(x->path);

		x = tmp;
	}
}

void stores_list_insert(struct cache_store *x, char *comment, char *profiles, char *path, char *maxdisksize, char *diskextra)
{
	if (comment != NULL) {
		FREE_AND_NULL(x->comment);

		if (strcmp(comment, ""))
			x->comment = xstrdup(comment);
	}

	if (profiles != NULL) {
		if (x->profiles != NULL)
			array_free(x->profiles);

		if (strcmp(profiles, ""))
			x->profiles = string_break(profiles, ',');
		else
			x->profiles = NULL;
	}

	if (path != NULL) {
		FREE_AND_NULL(x->path);

		if (strcmp(path, ""))
			x->path = xstrdup(path);
	}

	if (maxdisksize != NULL) {
		x->maxdisksize = 0;

		if (strcmp(maxdisksize, ""))
			x->maxdisksize = string_to_size(maxdisksize);
	}

	if (diskextra != NULL) {
		x->diskextra = 0;

		if (strcmp(diskextra, ""))
			x->diskextra = string_to_size(diskextra);
	}
}

void select_store(CONNECTION * connection, CACHE * cache, CACHEMAP * cachemap)
{
	int ptmp;
	size_t lowest = ~0;
	struct cache_store *selected = NULL, *stores;

	ASSERT(connection != NULL);

	for (stores = cache->stores; stores; stores = stores->next) {
		if (stores->enabled == FALSE)
			continue;

		if (!profile_find(connection->profiles, stores->profiles))
			continue;

		/* check if the entry is properly configured, so assumptions can be made later */
		if (stores->path == NULL || stores->maxdisksize == 0)
			continue;

		switch (cache->sbalancemethod) {
		case SBALANCE_FILLSIZE:
			if (stores->disksize < lowest) {
				lowest = stores->disksize;
				selected = stores;
			}
			break;
		case SBALANCE_FILLPERCENT:
			ptmp = (int) (((float) stores->disksize / (float) stores->maxdisksize) * 100.0);
			if (ptmp < lowest) {
				lowest = ptmp;
				selected = stores;
			}
			break;
		}

	}

	cachemap->store = selected;
}

void stores_update(CACHE * cache)
{
	int i;
	struct cache_store *stores;
	CACHEMAP *cachemap;
	struct CACHEMAPLIST_T *cl, *tmp;

	/* this routine is called whenever the cache disk store configuration is changed, it will resynchronize everything
	   in the cache hash table. */
	for (i = 0; i < cache->hashsize; i++) {
		for (cl = cache->cachemap[i]; cl; cl = tmp) {
			tmp = cl->next;
			cachemap = cl->cachemap;

			if (cachemap->store != NULL) {
				for (stores = cache->stores; stores && cachemap->store != stores; stores = stores->next);
				if (stores == NULL || stores->path == NULL) {
					/* this cachemap's filesystem store information doesn't exist anymore */

					if (!(cachemap->flags & CACHE_MEM)) {
						cache_hash_remove(cache, cachemap);
						cache_free(cache, cachemap);
					} else {
						/* it's mapped in memory, so it may still be useful. */
						cl->cachemap->store = NULL;
						cl->cachemap->flags &= ~CACHE_DISK;
					}
				}
			}
		}
	}

	for (stores = global->cache->stores; stores; stores = stores->next) {
		if (stores->journalfd != -1) {
			close(stores->journalfd);
			stores->journalfd = -1;
		}

		if (stores->path == NULL)
			continue;

		stores->diskentries = stores->disksize = 0;

		/* re-reading the journals will reassociate each cache object in the hash table with a cache disk store. */
		cache_journal_read(global->cache, stores);
		cache_journal_write(global->cache, stores);
	}
}

CACHE *cache_init()
{
	CACHE *cache;

	cache = xmalloc(sizeof(CACHE));

	pthread_mutex_init(&cache->lock, NULL);

	cache->enabled = FALSE;
	cache->id = 0;
	cache->sbalancemethod = SBALANCE_FILLSIZE;
	cache->violaterfc = FALSE;
	cache->hashsize = CACHEMAP_HASH_SIZE;
	cache->memsize = cache->maxmemsize = cache->memextra = 0;
	cache->minage = cache->maxage = 0;
	cache->lmfactor = 0;
	cache->validate = 0;
	cache->minsize = cache->maxsize = 0;
	cache->mementries = 0;
	cache->hit = cache->miss = 0;
	cache->maxwaitsize = 0;
	cache->prefetchwindow = 10;
	cache->lastclean = 0;
	cache->icpport = 0;
	cache->icptimeout = ICP_TIMEOUT;
	cache->stores = NULL;

	cache->cachemap = xmalloc(sizeof(struct CACHEMAPLIST_T *) * cache->hashsize);
	memset(cache->cachemap, 0, sizeof(struct CACHEMAPLIST_T *) * cache->hashsize);

	return cache;
}

CACHEMAP *cache_hash_find(CACHE * cache, char *key)
{
	unsigned int k;
	struct CACHEMAPLIST_T *cachemaplist;
	CACHEMAP *cachemap = NULL;

	k = hash_key(cache->hashsize, key);

	cachemaplist = cache->cachemap[k];

	while (cachemaplist && strcmp(cachemaplist->cachemap->key, key))
		cachemaplist = cachemaplist->next;

	if (cachemaplist != NULL)
		cachemap = cachemaplist->cachemap;

	return cachemap;
}

int cache_filename(CACHE * cache, CACHEMAP * cachemap, char *buf, int len)
{
	int x;
	unsigned char k;
	char *key = cachemap->key;

	k = hash_key(256, cachemap->key);
	x = snprintf(buf, len, "%s/%.2x/", cachemap->store->path, k);
	for (; x < len - 1 && *key; x++, key++)
		buf[x] = (*key == '/') ? '\\' : *key;
	buf[x] = '\0';

	return k;
}

void cache_hash_add(CACHE * cache, CACHEMAP * cachemap)
{
	unsigned int k;
	struct CACHEMAPLIST_T *cachemaplist;

	k = hash_key(cache->hashsize, cachemap->key);

	if (cache->cachemap[k] == NULL) {
		cache->cachemap[k] = xmalloc(sizeof(struct CACHEMAPLIST_T));
		cache->cachemap[k]->cachemap = cachemap;
		cache->cachemap[k]->next = cache->cachemap[k]->prev = NULL;
	} else {
		for (cachemaplist = cache->cachemap[k]; strcmp(cachemaplist->cachemap->key, cachemap->key) && cachemaplist->next; cachemaplist = cachemaplist->next);
		ASSERT(cachemaplist->next == NULL);

		cachemaplist->next = xmalloc(sizeof(struct CACHEMAPLIST_T));
		cachemaplist->next->cachemap = cachemap;
		cachemaplist->next->prev = cachemaplist;
		cachemaplist->next->next = NULL;
	}
}

void cache_hash_remove(CACHE * cache, CACHEMAP * cachemap)
{
	unsigned int k;
	struct CACHEMAPLIST_T *cachemaplist;

	k = hash_key(cache->hashsize, cachemap->key);

	for (cachemaplist = cache->cachemap[k]; cachemaplist && strcmp(cachemaplist->cachemap->key, cachemap->key); cachemaplist = cachemaplist->next);
	if (cachemaplist != NULL) {
		if (cachemaplist->next != NULL)
			cachemaplist->next->prev = cachemaplist->prev;
		if (cachemaplist->prev != NULL)
			cachemaplist->prev->next = cachemaplist->next;
		else
			cache->cachemap[k] = cachemaplist->next;

		xfree(cachemaplist);
	} else
		ASSERT(0);
}



CACHEMAP *cache_open(CONNECTION * connection, CACHE * cache, char *key, char *header, int flags)
{
	int fd;
	time_t utctime;
	char buf[1024], *ptr;
	CACHEMAP *cachemap;
	struct stat st;

	pthread_mutex_lock(&cache->lock);

	if (cache->enabled == FALSE) {
		pthread_mutex_unlock(&cache->lock);

		return NULL;
	}

	utctime = utc_time(NULL);

	cachemap = cache_hash_find(cache, key);

	if (cachemap != NULL) {
		if ((cachemap->flags & CACHE_PREFETCHED) && utctime - cachemap->mtime > cache->prefetchwindow)
			cachemap->flags &= ~CACHE_PREFETCHED;

		if (flags & CACHE_WRITING) {
			/* this will solve a race condition where a URL gets added to the prefetch
			   queue more than once and several threads try to prefetch it at 
			   about the same time. */
			if (cachemap->refcount != 0 || (cachemap->flags & CACHE_PREFETCHED)) {
				if (cachemap->flags & CACHE_PREFETCHED)
					putlog(MMLOG_DEBUG, "attempted to recreate cached file within prefetch window");

				cachemap = NULL;

				goto out;
			}

			putlog(MMLOG_DEBUG, "recreating cache file");

			/* recreate existing cache file */
			cache_invalidate_unlocked(cache, cachemap);

			cachemap = cachemap_new(cache);
			cachemap->key = xstrdup(key);
			cachemap->size = cachemap->realsize = cachemap->offset = 0;
			cachemap->mtime = utctime;
			cachemap->flags = flags;
			cachemap->refcount = 1;

			cache_hash_add(cache, cachemap);
		} else {
			if (cachemap->flags & CACHE_INVALID) {
				if (cachemap->flags & CACHE_INVALID)
					putlog(MMLOG_DEBUG, "cache object is invalid");

				cachemap = NULL;

				goto out;
			}

			if (cachemap->flags & CACHE_WRITING) {
				/* wait for thread writing the cache file to complete.
				   this isn't a very good way to do things, I would rather
				   be able to read from the cache as it is being written;
				   but this becomes very complex with threading, and I don't
				   think it's worth the trouble to handle rare cases like this */

				if ((cachemap->header->content_length != -1 && cachemap->header->content_length > cache->maxwaitsize) || cachemap->size > cache->maxwaitsize) {
					/* file is too large to wait on */
					putlog(MMLOG_DEBUG, "cache file too large to wait on");

					cachemap = NULL;

					goto out;
				}

				/* increase refcount incase thread downloading file
				   tries to invalidate it */
				cachemap->refcount++;

				pthread_cond_wait(&cachemap->writecompletecond, &cache->lock);

				if (cachemap->flags & CACHE_INVALID) {
					cache_invalidate_unlocked(cache, cachemap);

					cachemap = NULL;

					goto out;
				}

				cachemap->refcount--;

				/* did we stop waiting because it was too large ? */
				if (cachemap->size > cache->maxwaitsize) {
					putlog(MMLOG_DEBUG, "stopped waiting on large cache file");

					cachemap = NULL;

					goto out;
				}

				putlog(MMLOG_DEBUG, "waited for cache file that was being written");
			}

			if (!(cachemap->flags & CACHE_MEM) && cachemap->store != NULL) {
				ASSERT(cachemap->flags & CACHE_DISK);
				ASSERT(cachemap->data == NULL);
				ASSERT(cachemap->fd == -1);
				ASSERT(cachemap->refcount == 0);

				cache_filename(cache, cachemap, buf, sizeof(buf));

				fd = open(buf, O_RDONLY);
				if (fd != -1) {
					fstat(fd, &st);
					ptr = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);

					if (ptr != MAP_FAILED) {
						cachemap->mtime = st.st_mtime + gmtoffset;
						cachemap->atime = st.st_atime + gmtoffset;

						if (cachemap->size != st.st_size) {
							putlog(MMLOG_CACHE, "file meta information didn't match journaled data");

							/* compensate for unexpected change in file size */
							cachemap->store->disksize += (int) (st.st_size - cachemap->size);
							cachemap->size = cachemap->realsize = st.st_size;

							/* add journal entry with updated file information */
							cache_journal_add(cache, cachemap);
						} else
							cachemap->size = cachemap->realsize = st.st_size;

						cachemap->data = ptr;
						cachemap->flags |= CACHE_MEM;

						cache->memsize += cachemap->size;
						cache->mementries++;

#ifdef HAVE_MADVISE
						/* ask the kernel to asyncronously read the file from disk so it's cached
						   by the time we need it. */
						madvise(cachemap->data, cachemap->size, MADV_WILLNEED);
#endif				/* HAVE_MADVISE */
					}

					close(fd);
				}
			}

			if (!(cachemap->flags & CACHE_MEM)) {
				cache_invalidate_unlocked(cache, cachemap);
				cachemap = NULL;

				goto out;
			}

			cachemap->refcount++;
		}
	} else if (flags & CACHE_WRITING) {
		/* new cache item */
		cachemap = cachemap_new(cache);
		cachemap->key = xstrdup(key);
		cachemap->size = cachemap->realsize = cachemap->offset = 0;
		cachemap->mtime = utctime;
		cachemap->flags = flags;
		cachemap->refcount = 1;

		cache_hash_add(cache, cachemap);
	}

	if (cachemap != NULL) {
		cachemap->atime = utctime;

		if (!(cachemap->flags & CACHE_WRITING)) {
			if (cachemap->header == NULL) {
				/* refcount must be 1 at this point, so modifying any cachemap
				   member is safe */
				cachemap->offset = header_length(cachemap->data, cachemap->size);
				if (cachemap->offset != 0)
					cachemap->header = http_header_parse_response(cachemap->data);
			}
		} else {
			select_store(connection, cache, cachemap);

			/* set header first or we open up a race condition since we unlock
			   in cache_add */
			cachemap->header = http_header_parse_response(header);
			cachemap->offset = strlen(header);

			if (cachemap->header->content_length != -1 && cachemap->header->content_length > cache->maxsize) {
				putlog(MMLOG_CACHE, "file too large to cache (size: %d)", cachemap->header->content_length);

				cache_invalidate_unlocked(cache, cachemap);

				cachemap = NULL;

				goto out;
			}

			/* NOTE: cache_add_unlocked unlocks the mutex to relieve contention...
			   there may be race conditions calling it from here if you're not careful! */
			cache_add_unlocked(cache, cachemap, header, strlen(header));

			putlog(MMLOG_DEBUG, "cache header offset: %d", cachemap->offset);
		}

		if (cachemap->offset == 0 || cachemap->header == NULL || !cacheable(cachemap->header)) {
			/* this shouldn't happen, it may be a sign that the cache file was corrupted */
			cache_invalidate_unlocked(cache, cachemap);

			cachemap = NULL;

			putlog(MMLOG_CACHE, "corrupt header: %s", key);

			goto out;
		}

		if (!(cachemap->flags & CACHE_WRITING))
			cache_check_header(cache, cachemap);

		if (!(flags & CACHE_OFFLINE) && (cachemap->flags & CACHE_VIOLATION) && cache->violaterfc == FALSE) {
			cache_close_unlocked(cache, cachemap);
			cachemap = NULL;
		}

	}

      out:

	if (cache->memsize > cache->maxmemsize || (cachemap != NULL && cachemap->store != NULL && cachemap->store->disksize > cachemap->store->maxdisksize))
		cache_clean(cache);

	pthread_mutex_unlock(&cache->lock);

	return cachemap;
}

void cache_close(CACHE * cache, CACHEMAP * cachemap)
{
	pthread_mutex_lock(&cache->lock);
	cache_close_unlocked(cache, cachemap);
	pthread_mutex_unlock(&cache->lock);
}

void cache_close_unlocked(CACHE * cache, CACHEMAP * cachemap)
{
	if (cachemap->flags & CACHE_WRITING) {
		cachemap->header->content_length = cachemap->size - cachemap->offset;

		if (cachemap->size != cachemap->realsize)
			cache_resize(cache, cachemap, 0);

		msync(cachemap->data, cachemap->realsize, MS_ASYNC);

		if (cachemap->fd != -1)
			close(cachemap->fd);

		cachemap->fd = -1;
	}

	cachemap->refcount--;

	if (cachemap->flags & CACHE_INVALID) {
		if (cachemap->refcount == 0)
			cache_invalidate_unlocked(cache, cachemap);
	} else if (cachemap->flags & CACHE_WRITING) {
		if (cachemap->size == 0 && cachemap->size < cache->minsize) {
			if (cachemap->size < cache->minsize)
				putlog(MMLOG_CACHE, "file too small");

			cache_invalidate_unlocked(cache, cachemap);
		} else {
			cachemap->flags &= ~CACHE_WRITING;

			/* cache_touch_unlocked does the journal update */
			cache_touch_unlocked(cache, cachemap, TRUE);

			pthread_cond_broadcast(&cachemap->writecompletecond);
		}
	} else
		cache_touch_unlocked(cache, cachemap, FALSE);
}

int cache_add(CACHE * cache, CACHEMAP * cachemap, void *buf, size_t len)
{
	int ret;

	pthread_mutex_lock(&cache->lock);

	ret = cache_add_unlocked(cache, cachemap, buf, len);

	pthread_mutex_unlock(&cache->lock);

	return ret;
}

int cache_add_unlocked(CACHE * cache, CACHEMAP * cachemap, void *buf, size_t len)
{
	int x;
	size_t ret;
	char path[1024], file[1024], *key;

	if (cachemap->flags & CACHE_INVALID)
		return -1;

	key = cachemap->key;

	if (!(cachemap->flags & CACHE_MEM)) {
		if (cachemap->store != NULL) {
			cache_filename(cache, cachemap, file, sizeof(file));

			x = hash_key(256, key);
			snprintf(path, sizeof(path), "%s/%.2x", cachemap->store->path, x);

			mkdir(path, 0755);

			cachemap->fd = open(file, O_RDWR | O_CREAT, 0644);
			if (cachemap->fd != -1) {
				cachemap->store->diskentries++;
				cachemap->flags |= CACHE_DISK;

				cache_journal_add(cache, cachemap);
			} else
				cachemap->store = NULL;
		}

		ret = cache_resize(cache, cachemap, len);
		if (ret == 0)
			goto error;

		memcpy(cachemap->data, buf, len);
	} else {
		ret = cache_resize(cache, cachemap, len);
		if (ret == 0)
			goto error;

		memcpy(cachemap->data + (cachemap->size - len), buf, len);

	}


	cache->memsize += len;
	if (cachemap->flags & CACHE_DISK)
		cachemap->store->disksize += len;

	if ((cachemap->store != NULL && cachemap->store->disksize > cachemap->store->maxdisksize) || cache->memsize > cache->maxmemsize)
		cache_clean(cache);

	if (cache->maxsize != 0 && cachemap->size > cache->maxsize) {
		putlog(MMLOG_CACHE, "max size exceeded (size: %u)", cachemap->size);

		goto error;
	}

	if (cachemap->size > cache->maxwaitsize)
		pthread_cond_broadcast(&cachemap->writecompletecond);

	return len;

      error:
	/* we don't want to free it yet */
	cachemap->refcount++;
	cache_invalidate_unlocked(cache, cachemap);

	pthread_cond_broadcast(&cachemap->writecompletecond);

	return -1;
}

void cache_scandisk(CACHE * cache, struct cache_store *store)
{
	int x;
	char buf[1024], buf2[1024], *ptr;
	DIR *base, *subdir;
	CACHEMAP *cachemap;
	struct dirent *basedirent, *subdirent;
	struct stat st;
	URL *url;

	base = opendir(store->path);
	if (base == NULL)
		return;

	putlog(MMLOG_CACHE, "scanning cache directory structure in %s", store->path);

	while ((basedirent = readdir(base)) != NULL) {
		if (!strcmp(basedirent->d_name, ".") || !strcmp(basedirent->d_name, ".."))
			continue;

		snprintf(buf, sizeof(buf), "%s/%s", store->path, basedirent->d_name);
		subdir = opendir(buf);
		if (subdir != NULL) {
			while ((subdirent = readdir(subdir)) != NULL) {
				snprintf(buf, sizeof(buf), "%s/%s/%s", store->path, basedirent->d_name, subdirent->d_name);
				x = stat(buf, &st);
				if (x != -1 && S_ISREG(st.st_mode)) {
					s_strncpy(buf2, subdirent->d_name, sizeof(buf2));
					for (x = 0; buf2[x]; x++)
						if (buf2[x] == '\\')
							buf2[x] = '/';


					/* this makes sure it's a valid URL */
					url = url_parse(buf2);
					if (url != NULL) {
						ptr = url_create(url);
						url_free(url);
					} else {
						putlog(MMLOG_CACHE, "%s doesn't appear to be a valid cache file", buf);
						continue;
					}

					cachemap = cache_hash_find(cache, ptr);
					if (cachemap != NULL) {
						/* this can happen if the unlink thread didn't reach this entry before the proxy
						   server was killed and the file was cached again, or if this directory
						   was used for another instance of the proxy. */
						putlog(MMLOG_CACHE, "duplicate of %s from %s found in %s", ptr, (cachemap->store->path != NULL) ? cachemap->store->path : "(memory)", store->path);

						/* this could happen if scandisk is called twice on the same directory */
						if (cachemap->store == store) {
							xfree(ptr);

							continue;
						}

						/* keep other copy if it's in use or newer */
						if (cachemap->refcount != 0 || cachemap->mtime > st.st_mtime + gmtoffset) {
							unlink_list_add(buf);

							xfree(ptr);

							continue;
						}

						if (cachemap->flags & CACHE_MEM)
							cache_mem_delete(cache, cachemap);
						if (cachemap->flags & CACHE_DISK)
							cache_disk_delete(cache, cachemap);
					}

					if (cachemap == NULL) {
						cachemap = cachemap_new(cache);
						cachemap->key = ptr;
						cache_hash_add(cache, cachemap);
					} else
						xfree(ptr);

					cachemap->store = store;
					cachemap->size = cachemap->realsize = st.st_size;
					cachemap->mtime = st.st_mtime + gmtoffset;
					cachemap->atime = st.st_atime + gmtoffset;
					cachemap->flags = CACHE_DISK;

					store->disksize += st.st_size;
					store->diskentries++;
				}
			}

			closedir(subdir);
		}
	}

	closedir(base);

	return;
}

void cache_clean(CACHE * cache)
{
	int i;
	unsigned int memsizegoal, disksizegoal;
	time_t oldest = 0, newest = 0, threshold, increment, utctime;
	struct CACHEMAPLIST_T *cachemaplist, *tmp;
	CACHEMAP *cachemap;

	utctime = utc_time(NULL);

	if (cache->lastclean != 0 && cache->lastclean + CLEAN_INTERVAL > utctime)
		return;

	cache->lastclean = utctime;

	putlog(MMLOG_CACHE, "cleaning cache");

	for (i = 0; i < cache->hashsize; i++) {
		for (cachemaplist = cache->cachemap[i]; cachemaplist; cachemaplist = cachemaplist->next) {
			if (cachemaplist->cachemap->refcount != 0)
				continue;

			if (cachemaplist->cachemap->atime > newest || newest == 0)
				newest = cachemaplist->cachemap->atime;
			if (cachemaplist->cachemap->atime < oldest || oldest == 0)
				oldest = cachemaplist->cachemap->atime;
		}
	}

	/* oops, nothing we can free */
	if (newest == 0)
		return;

	increment = (newest - oldest) / CACHE_INCREMENT_FRACTION + 1;
	threshold = oldest + increment;

	if (cache->memsize > cache->maxmemsize)
		memsizegoal = (cache->memextra > cache->maxmemsize) ? 0 : cache->maxmemsize - cache->memextra;
	else
		memsizegoal = cache->maxmemsize;

	while (threshold <= newest + increment) {
		for (i = 0; i < cache->hashsize; i++) {
			for (cachemaplist = cache->cachemap[i]; cachemaplist; cachemaplist = tmp) {
				tmp = cachemaplist->next;
				cachemap = cachemaplist->cachemap;

				if (cachemap->refcount != 0)
					continue;

				if (cachemap->atime > threshold)
					continue;

				ASSERT(cachemap->flags & (CACHE_DISK | CACHE_MEM));
				ASSERT(!(cachemap->flags & CACHE_DISK) || cachemap->store != NULL);

				if (cachemap->flags & CACHE_DISK) {
					if (cachemap->store->disksize > cachemap->store->maxdisksize)
						disksizegoal = (cachemap->store->diskextra > cachemap->store->maxdisksize) ? 0 : cachemap->store->maxdisksize - cachemap->store->diskextra;
					else
						disksizegoal = cachemap->store->maxdisksize;

					if (cachemap->store->disksize > disksizegoal) {
						cache_invalidate_unlocked(cache, cachemap);
						continue;
					}
				}

				if ((cachemap->flags & CACHE_MEM) && cache->memsize > memsizegoal) {
					if (!(cachemap->flags & CACHE_DISK))
						cache_hash_remove(cache, cachemap);

					cache_mem_delete(cache, cachemap);
				}
			}
		}

		/* keep increasing the threshold until enough memory can be free'd */
		threshold += increment;
	}
}

int cacheable(HEADER * header)
{
	if (header->type == HTTP_RESP) {
		if (header->code != 200)
			return FALSE;
	} else if (header->type == HTTP_PROXY) {
		if (strcasecmp(header->method, "GET"))
			return FALSE;
		if (header->authorization != NULL)
			return FALSE;
		if (header->username != NULL && strcasecmp(header->username, ANONLOGIN))
			return FALSE;
		if (header->range != NULL)
			return FALSE;
	} else
		return FALSE;

	return TRUE;
}

/* 
check if cache file needs validating based on request headers
*/
void cache_check(CACHE * cache, HEADER * header)
{
	int x, expired = FALSE, i;
	time_t utctime = utc_time(NULL);
	char buf[1024], **args;
	CACHEMAP *cachemap = NULL;

	snprintf(buf, sizeof(buf), "%s://%s:%d%s", header->proto, header->host, header->port, header->file);
	cachemap = cache_open(NULL, cache, buf, NULL, CACHE_OFFLINE);

	if (cachemap == NULL)
		return;

	pthread_mutex_lock(&cache->lock);

	if (cachemap->flags & CACHE_PREFETCHED)
		goto out;

	if (cachemap->flags & CACHE_EXPIRED)
		expired = TRUE;

	if (header->cache_control != NULL) {
		putlog(MMLOG_DEBUG, "request cache-control header: %s", header->cache_control);

		args = string_break(header->cache_control, ',');

		for (x = 0; args[x]; x++) {
			if (!strcasecmp(args[x], "no-cache") || !strcasecmp(args[x], "no-store"))
				expired = TRUE;
			else if (!strncasecmp(args[x], "max-age=", 8)) {
				/* file can't be any older than this */
				i = atoi(&args[x][8]);

				if (utctime - cachemap->mtime > i)
					expired = TRUE;
			} else if (!strncasecmp(args[x], "min-fresh=", 10)) {
				/* file must be fresh for at least this long */
				i = atoi(&args[x][10]);

				if ((cachemap->etime != 0 && utctime + i > cachemap->etime) || (cachemap->ftime != 0 && cachemap->etime == 0 && utctime + i > cachemap->ftime))
					expired = TRUE;
			} else if (!strncasecmp(args[x], "max-stale=", 10)) {
				/* file may be stale for this long */
				i = atoi(&args[x][10]);

				if ((cachemap->etime != 0 && utctime - cachemap->etime <= i) || (cachemap->ftime != 0 && cachemap->etime == 0 && utctime - cachemap->ftime <= i))
					expired = FALSE;

			}

			xfree(args[x]);
		}

		xfree(args);
	}

	if (expired == TRUE)
		cachemap->flags |= CACHE_VALIDATE;

      out:

	pthread_mutex_unlock(&cache->lock);
	cache_close(cache, cachemap);
}

FILEBUF *cache_to_filebuf(CACHEMAP * cachemap, int maxlen)
{
	FILEBUF *ret;

	ret = filebuf_new();

	filebuf_add(ret, cachemap->data + cachemap->offset, (maxlen < cachemap->size - cachemap->offset) ? maxlen : cachemap->size - cachemap->offset);

	return ret;
}

void cache_invalidate(CACHE * cache, CACHEMAP * cachemap)
{
	pthread_mutex_lock(&cache->lock);

	cache_invalidate_unlocked(cache, cachemap);

	pthread_mutex_unlock(&cache->lock);
}

void cache_invalidate_unlocked(CACHE * cache, CACHEMAP * cachemap)
{
	if (!(cachemap->flags & CACHE_INVALID)) {
		putlog(MMLOG_CACHE, "invalidated: %s", cachemap->key);
		cachemap->flags |= CACHE_INVALID;
		cache_journal_add(cache, cachemap);
	}

	if (cachemap->refcount > 1) {
		cachemap->refcount--;

		pthread_cond_broadcast(&cachemap->writecompletecond);

		return;
	}

	if (cachemap->flags & CACHE_MEM)
		cache_mem_delete(cache, cachemap);
	if (cachemap->flags & CACHE_DISK)
		cache_disk_delete(cache, cachemap);

	cache_hash_remove(cache, cachemap);

	cache_free(cache, cachemap);
}


void cache_disk_delete(CACHE * cache, CACHEMAP * cachemap)
{
	char buf[1024];

	ASSERT(cachemap->flags & CACHE_DISK);
	ASSERT(!(cachemap->flags & CACHE_MEM));

	if (cachemap->store->path != NULL) {
		cache_filename(cache, cachemap, buf, sizeof(buf));

		unlink_list_add(buf);
	}

	cachemap->store->disksize -= cachemap->size;
	cachemap->store->diskentries--;

	cachemap->flags &= ~CACHE_DISK;
}

void cache_mem_delete(CACHE * cache, CACHEMAP * cachemap)
{
	ASSERT((cachemap->flags & CACHE_MEM) && cachemap->data != NULL);

	munmap(cachemap->data, cachemap->realsize);
	cache->memsize -= cachemap->size;
	cache->mementries--;

	cachemap->flags &= ~CACHE_MEM;
	cachemap->data = NULL;
}



void cache_touch(CACHE * cache, CACHEMAP * cachemap, int modified)
{
	pthread_mutex_lock(&cache->lock);
	cache_touch_unlocked(cache, cachemap, modified);
	pthread_mutex_unlock(&cache->lock);
}

void cache_touch_unlocked(CACHE * cache, CACHEMAP * cachemap, int modified)
{
	time_t utctime;
	char buf[1024];
	struct utimbuf utb;

	utctime = utc_time(NULL);

	cachemap->atime = utctime;
	if (modified == TRUE)
		cachemap->mtime = utctime;

	if (cachemap->store != NULL) {
		cache_filename(cache, cachemap, buf, sizeof(buf));

		utb.actime = cachemap->atime - gmtoffset;
		utb.modtime = cachemap->mtime - gmtoffset;
		utime(buf, &utb);
	}

	cache_journal_add(cache, cachemap);
}

void cache_free(CACHE * cache, CACHEMAP * cachemap)
{
	if (cachemap->fd != -1)
		close(cachemap->fd);

	if (cachemap->header != NULL)
		http_header_free(cachemap->header);

	pthread_cond_destroy(&cachemap->writecompletecond);

	xfree(cachemap->key);
	xfree(cachemap);
}

int cache_flag_set(CACHE * cache, CACHEMAP * cachemap, int flag)
{
	int ret;

	pthread_mutex_lock(&cache->lock);
	cachemap->flags |= flag;
	ret = cachemap->flags;
	pthread_mutex_unlock(&cache->lock);

	return ret;
}

int cache_flag_unset(CACHE * cache, CACHEMAP * cachemap, int flag)
{
	int ret;

	pthread_mutex_lock(&cache->lock);
	cachemap->flags &= ~flag;
	ret = cachemap->flags;
	pthread_mutex_unlock(&cache->lock);

	return ret;
}

void cache_check_header(CACHE * cache, CACHEMAP * cachemap)
{
	int i, x;
	struct tm t;
	char **args;
	time_t utctime = utc_time(NULL);

	cachemap->ftime = 0;
	cachemap->etime = 0;
	cachemap->lmtime = 0;
	cachemap->header->chunked = FALSE;
	cachemap->flags &= ~CACHE_EXPIRED;


	if (cachemap->header->expires != NULL) {
		putlog(MMLOG_DEBUG, "cache file expires header: %s", cachemap->header->expires);

		memset(&t, 0, sizeof(t));
		strptime(cachemap->header->expires, HTTPTIMEFORMAT, &t);
		cachemap->etime = mktime(&t);
	}

	if (cachemap->header->last_modified != NULL) {
		putlog(MMLOG_DEBUG, "cache file last-modified header: %s", cachemap->header->last_modified);

		memset(&t, 0, sizeof(t));
		strptime(cachemap->header->last_modified, HTTPTIMEFORMAT, &t);
		cachemap->lmtime = mktime(&t);
	}

	/* rfc says Cache-Control takes precedence over Expires: header */
	if (cachemap->header->cache_control != NULL) {
		putlog(MMLOG_DEBUG, "cache file cache-control header: %s", cachemap->header->cache_control);

		args = string_break(cachemap->header->cache_control, ',');

		for (x = 0; args[x]; x++) {
			if (!strncasecmp(args[x], "max-age=", 8)) {
				i = atoi(&args[x][8]);

				cachemap->etime = cachemap->mtime + i;
			} else if (!strcasecmp(args[x], "must-revalidate"))
				cachemap->flags |= CACHE_EXPIRED;
			else if (!strcasecmp(args[x], "no-cache") || !strcasecmp(args[x], "no-store"))
				cachemap->flags |= (CACHE_EXPIRED | CACHE_VIOLATION);

			xfree(args[x]);
		}

		xfree(args);
	}

	if (cachemap->lmtime != 0) {
		/* sanity check */
		if (cachemap->lmtime < cachemap->mtime) {
			/* calculate how long this file should be considered fresh based
			   on a percentage of the delta between the time it was last modified
			   and when it was downloaded or validated; this is used only 
			   if the server gives us nothing else to go on.
			 */
			cachemap->ftime = cachemap->mtime + ((cachemap->mtime - cachemap->lmtime) * cache->lmfactor / 100);

			putlog(MMLOG_DEBUG, "lmtime factor: %u", cachemap->ftime - cachemap->mtime);
		}
	}

	if ((cachemap->etime == 0 && cachemap->ftime == 0) || (cachemap->mtime + cache->maxage < utctime) || (cachemap->etime != 0 && cachemap->etime < utctime) || (cachemap->ftime != 0 && cachemap->etime == 0 && cachemap->ftime < utctime) || (cachemap->lmtime != 0 && cachemap->lmtime + cache->minage > utctime) || (cachemap->etime == 0 && cachemap->mtime + cache->validate < utctime)) {

		/* cache file has expired */
		cachemap->flags |= CACHE_EXPIRED;
	}
}

void cache_validated(CACHE * cache, CACHEMAP * cachemap)
{
	pthread_mutex_lock(&cache->lock);
	cachemap->flags &= ~CACHE_VALIDATE;
	cache_touch_unlocked(cache, cachemap, TRUE);
	pthread_mutex_unlock(&cache->lock);
}

int cache_invalidate_key(CACHE * cache, char *key)
{
	int ret = FALSE;
	CACHEMAP *cachemap;

	pthread_mutex_lock(&cache->lock);

	cachemap = cache_hash_find(cache, key);
	if (cachemap != NULL) {
		cachemap->refcount++;
		cache_invalidate_unlocked(cache, cachemap);
		ret = TRUE;
	}

	pthread_mutex_unlock(&cache->lock);

	return ret;
}

size_t cache_resize(CACHE * cache, CACHEMAP * cachemap, size_t size)
{
	int x;
	void *ptr;
	size_t oldrs;

	if (size == 0 || (cachemap->size + size > cachemap->realsize)) {
		/* this routine can block under some circumstances, releasing the mutex
		   is safe. */
		pthread_mutex_unlock(&cache->lock);

		oldrs = cachemap->realsize;

		if (size == 0)
			cachemap->realsize = cachemap->size;
		else {
			/* if the content length is known, just map the whole thing now */
			if (cachemap->header->content_length != -1 && cachemap->header->content_length + cachemap->offset >= size)
				cachemap->realsize = cachemap->header->content_length + cachemap->offset;
			else {
				/* preallocation tends to hide some errors such as off-by-one's, so we won't do it if this is a debug
				   build. */
#ifdef _DEBUG
				cachemap->realsize = cachemap->size + size;
#else
				cachemap->realsize = ALIGN2(cachemap->size + size, CACHE_ALIGNMENT);
#endif
			}
		}

		if (cachemap->fd != -1) {
			x = ftruncate(cachemap->fd, cachemap->realsize);

			if (x == -1) {
				pthread_mutex_lock(&cache->lock);
				goto error;
			}
		}

		if (!(cachemap->flags & CACHE_MEM))
			ptr = mmap(NULL, cachemap->realsize, PROT_WRITE | PROT_READ, (cachemap->fd != -1) ? MAP_SHARED : MAP_PRIVATE | MAP_ANON, cachemap->fd, 0);
		else
			ptr = p_mremap(cachemap->data, oldrs, cachemap->realsize, cachemap->fd);

		pthread_mutex_lock(&cache->lock);

		if (ptr == MAP_FAILED)
			goto error;
		else {
			if (!(cachemap->flags & CACHE_MEM)) {
				cache->mementries++;
				cachemap->flags |= CACHE_MEM;
			}

			cachemap->data = ptr;
		}
	}

	cachemap->size += size;

	return cachemap->size;

      error:

	return 0;
}

int cache_exists(CACHE * cache, char *key)
{
	int ret = FALSE;

	pthread_mutex_lock(&cache->lock);
	if (cache_hash_find(cache, key) != NULL)
		ret = TRUE;
	pthread_mutex_unlock(&cache->lock);

	return ret;
}

void cache_journal_write(CACHE * cache, struct cache_store *cstore)
{
	int i;
	char buf[256], buf2[256];
	FILE *fptr;
	CACHEMAP *cachemap;
	struct CACHEMAPLIST_T *cachemaplist;

	snprintf(buf, sizeof(buf), "%s/journal.new", cstore->path);
	fptr = fopen(buf, "w+");
	if (fptr == NULL)
		return;

	putlog(MMLOG_CACHE, "writing journal for %s", cstore->path);

	for (i = 0; i < cache->hashsize; i++) {
		for (cachemaplist = cache->cachemap[i]; cachemaplist; cachemaplist = cachemaplist->next) {
			cachemap = cachemaplist->cachemap;

			if (cachemap->store != NULL && cachemap->store == cstore)
				fprintf(fptr, "%d\t%s\t%u\t%lu\t%lu\n", cachemap->flags, cachemap->key, cachemap->size, cachemap->mtime, cachemap->atime);
		}
	}

	fclose(fptr);

	snprintf(buf, sizeof(buf), "%s/journal.new", cstore->path);
	snprintf(buf2, sizeof(buf2), "%s/journal", cstore->path);

	unlink(buf2);
	rename(buf, buf2);

	return;

}

void cache_journal_read(CACHE * cache, struct cache_store *cstore)
{
	int i;
	char buf[8192], **args;
	size_t size;
	time_t mtime, atime;
	FILE *fptr;
	CACHEMAP *cachemap;
	struct CACHEMAPLIST_T *cachemaplist, *tmp;

	snprintf(buf, sizeof(buf), "%s/journal", cstore->path);
	fptr = fopen(buf, "r");
	if (fptr == NULL) {
		cache_scandisk(cache, cstore);
		return;
	}

	putlog(MMLOG_CACHE, "reading journal for %s", cstore->path);

	/* read each journal entry and determine the last state a cache file was in */
	while (fgets(buf, sizeof(buf), fptr) != NULL) {
		args = string_break(buf, '\t');
		if (array_length(args) == 5) {
			size = strtoul(args[2], NULL, 10);
			mtime = strtoul(args[3], NULL, 10);
			atime = strtoul(args[4], NULL, 10);

			cachemap = cache_hash_find(cache, args[1]);
			if (cachemap == NULL) {
				cachemap = cachemap_new(cache);
				cachemap->store = cstore;
				cachemap->key = xstrdup(args[1]);
				cache_hash_add(cache, cachemap);
			} else if (cachemap->refcount != 0)
				continue;
			else if (cachemap->store == NULL)
				cachemap->store = cstore;

			/* we have to deal with the case when the same file shows up in more than one
			   journal (i.e. it was invalidated in one cache store, then later created in another)...
			   we'll use the newest copy */
			if (cachemap->store == cstore || (cachemap->mtime <= mtime && cachemap->store != cstore)) {
				cachemap->size = size;
				cachemap->flags = atoi(args[0]);
				cachemap->mtime = mtime;
				cachemap->atime = atime;

				cachemap->store = cstore;
			}
		}

		array_free(args);
	}

	for (i = 0; i < cache->hashsize; i++) {
		for (cachemaplist = cache->cachemap[i]; cachemaplist; cachemaplist = tmp) {
			tmp = cachemaplist->next;
			cachemap = cachemaplist->cachemap;

			if (cachemap->store != cstore || cachemap->refcount != 0)
				continue;

			/* these flags don't carry over between sessions */
			cachemap->flags &= ~CACHE_SESSION_FLAGS;
			/* XXX: ugly hack so redoing cache_journal_read on an active cache works */
			if (cachemap->data != NULL)
				cachemap->flags |= CACHE_MEM;

			if ((cachemap->flags & (CACHE_WRITING | CACHE_INVALID)) || !(cachemap->flags & CACHE_DISK)) {
				/* the cache object's final state was invalid... remove it. */

				cache_hash_remove(cache, cachemap);

				if (cachemap->flags & CACHE_MEM)
					cache_mem_delete(cache, cachemap);

				if (cachemap->flags & CACHE_DISK) {
					/* cache_disk_delete subtracts from the disksize
					   accounting, but it hasn't been added yet */
					cachemap->store->disksize += cachemap->size;
					cachemap->store->diskentries++;

					cache_disk_delete(cache, cachemap);
				}
				cache_free(cache, cachemap);
			} else {
				cachemap->store->diskentries++;
				cachemap->store->disksize += cachemap->size;
			}
		}
	}

	fclose(fptr);
}

int cache_journal_add(CACHE * cache, CACHEMAP * cachemap)
{
	int len;
	char buf[8192];

	if (cachemap->store == NULL)
		return TRUE;

	if (cachemap->store->journalfd == -1) {
		snprintf(buf, sizeof(buf), "%s/journal", cachemap->store->path);
		cachemap->store->journalfd = open(buf, O_WRONLY | O_CREAT | O_APPEND);

		if (cachemap->store->journalfd == -1)
			return -1;
	}

	len = snprintf(buf, sizeof(buf), "%d\t%s\t%u\t%lu\t%lu\n", cachemap->flags, cachemap->key, cachemap->size, cachemap->mtime, cachemap->atime);

	write(cachemap->store->journalfd, buf, len);

	cachemap->store->journalentries++;

	if (cachemap->store->journalentries >= JOURNAL_SIZE) {
		/* rewrite the journal, this will remove redundant entries */
		close(cachemap->store->journalfd);
		cachemap->store->journalfd = -1;

		cache_journal_write(cache, cachemap->store);
		cachemap->store->journalentries = 0;
	}

	return TRUE;
}


CACHEMAP *cachemap_new(CACHE * cache)
{
	CACHEMAP *cachemap;

	cachemap = xmalloc(sizeof(CACHEMAP));
	cachemap->flags = 0;
	cachemap->header = NULL;
	cachemap->fd = -1;
	cachemap->refcount = 0;
	cachemap->data = NULL;
	cachemap->store = NULL;
	pthread_cond_init(&cachemap->writecompletecond, NULL);

	return cachemap;
}

void unlink_list_add(char *file)
{
	struct unlink_list *node;

	node = xmalloc(sizeof(struct unlink_list));
	node->path = xstrdup(file);

	node->next = NULL;

	pthread_mutex_lock(&global->unlink_lock);

	if (global->unlink_tail != NULL)
		global->unlink_tail->next = node;
	global->unlink_tail = node;

	if (global->unlink_head == NULL)
		global->unlink_head = node;

	pthread_cond_signal(&global->unlink_cond);

	pthread_mutex_unlock(&global->unlink_lock);
}


void unlink_thread()
{
	int ret;
	struct unlink_list *tmp;

	pthread_mutex_lock(&global->unlink_lock);

	while (1) {
		if (global->unlink_head != NULL) {
			ret = unlink(global->unlink_head->path);
			if (ret == 0)
				putlog(MMLOG_DEBUG, "unlinked %s", global->unlink_head->path);

			if (global->unlink_head == global->unlink_tail)
				global->unlink_tail = NULL;

			tmp = global->unlink_head->next;
			xfree(global->unlink_head->path);
			xfree(global->unlink_head);

			global->unlink_head = tmp;

			/* ease up on disk activity and give other threads a chance to add
			   items to the unlink list */
			pthread_mutex_unlock(&global->unlink_lock);
			usleep(10000);
			pthread_mutex_lock(&global->unlink_lock);

			continue;
		}

		pthread_cond_wait(&global->unlink_cond, &global->unlink_lock);
	}
}
