/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include "proto.h"

/*
load <filter> section of xml_list into a FILTER_LIST structure
*/
FILTER_LIST *filter_load(FILTER_LIST * filter_list, XML_LIST * xml_list)
{
	FILTER_LIST *tmp_list = filter_list;
	struct FILTER_LIST_LIST *allow = NULL, *deny = NULL;


	if (tmp_list == NULL) {
		tmp_list = xmalloc(sizeof(FILTER_LIST));
		tmp_list->dtemplate = NULL;
		tmp_list->allow = NULL;
		tmp_list->deny = NULL;
		tmp_list->policy = POLICY_ALLOW;
		tmp_list->id = 0;
		tmp_list->enabled = TRUE;
		tmp_list->empty = filter_ll_new(NULL);

		filter_list = tmp_list;

		pthread_rwlock_init(&tmp_list->lock, NULL);
	} else {
		allow = tmp_list->allow;
		deny = tmp_list->deny;
	}

	while ((xml_list = xml_section(xml_list, "<filter>"))) {
		XML_LIST_LOOP(xml_list, "<filter>") {
			XML_LIST_CMP(xml_list, "<allow>") {
				allow = filter_ll_new(allow);
				allow->id = filter_list->id++;

				if (tmp_list->allow == NULL)
					tmp_list->allow = allow;
				XML_LIST_LOOP(xml_list, "<allow>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								allow->enabled = FALSE;
							else
								allow->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(allow, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(allow, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<host>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(allow, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<file>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(allow, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<template>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(allow, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}

			XML_LIST_CMP(xml_list, "<deny>") {
				deny = filter_ll_new(deny);
				deny->id = filter_list->id++;

				if (tmp_list->deny == NULL)
					tmp_list->deny = deny;
				XML_LIST_LOOP(xml_list, "<deny>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								deny->enabled = FALSE;
							else
								deny->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(deny, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(deny, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<host>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(deny, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<file>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(deny, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<template>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							filter_ll_insert(deny, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}
			XML_LIST_CMP(xml_list, "<enabled>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "false"))
						filter_list->enabled = FALSE;
					else
						filter_list->enabled = TRUE;
				}
			}
			XML_LIST_CMP(xml_list, "<default>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					FREE_AND_NULL(tmp_list->dtemplate);
					tmp_list->dtemplate = xstrdup(xml_list->item);
				}
			}
			XML_LIST_CMP(xml_list, "<policy>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "allow"))
						tmp_list->policy = POLICY_ALLOW;
					else if (!strcasecmp(xml_list->item, "deny"))
						tmp_list->policy = POLICY_DENY;
				}
			}
		}

	}

	return filter_list;
}


XML_LIST *filter_xml(FILTER_LIST * filter_list, XML_LIST * xml_list)
{
	int i;
	char *ptr;
	struct FILTER_LIST_LIST *fl = NULL;

	if (filter_list == NULL)
		return xml_list;

	pthread_rwlock_rdlock(&filter_list->lock);

	xml_list = xml_list_add(xml_list, "<filter>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
	xml_list = xml_list_add(xml_list, (filter_list->enabled == TRUE) ? "true" : "false", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<policy>", XML_TAG);
	xml_list = xml_list_add(xml_list, (filter_list->policy == POLICY_ALLOW) ? "allow" : "deny", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</policy>", XML_TAG);


	if (filter_list->dtemplate != NULL) {
		xml_list = xml_list_add(xml_list, "<default>", XML_TAG);
		ptr = string_to_xml(filter_list->dtemplate);
		xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
		xfree(ptr);
		xml_list = xml_list_add(xml_list, "</default>", XML_TAG);

	}

	for (i = 0; i < 2; i++) {
		switch (i) {
		case 0:
			fl = filter_list->allow;
			break;
		case 1:
			fl = filter_list->deny;
			break;
		}

		for (; fl; fl = fl->next) {
			xml_list = xml_list_add(xml_list, (i == 0) ? "<allow>" : "<deny>", XML_TAG);

			xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
			xml_list = xml_list_add(xml_list, (fl->enabled == TRUE) ? "true" : "false", XML_VALUE);
			xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

			if (fl->comment != NULL) {
				xml_list = xml_list_add(xml_list, "<comment>", XML_TAG);
				ptr = string_to_xml(fl->comment);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</comment>", XML_TAG);
			}
			if (fl->profiles != NULL) {
				xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);
				ptr = array_merge(fl->profiles, ',');
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);
			}
			if (fl->template != NULL) {
				xml_list = xml_list_add(xml_list, "<template>", XML_TAG);
				ptr = string_to_xml(fl->template);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</template>", XML_TAG);

			}
			if (fl->host != NULL) {
				xml_list = xml_list_add(xml_list, "<host>", XML_TAG);
				ptr = string_to_xml(fl->host);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</host>", XML_TAG);

			}
			if (fl->file != NULL) {
				xml_list = xml_list_add(xml_list, "<file>", XML_TAG);
				ptr = string_to_xml(fl->file);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</file>", XML_TAG);
			}

			xml_list = xml_list_add(xml_list, (i == 0) ? "</allow>" : "</deny>", XML_TAG);
		}
	}

	xml_list = xml_list_add(xml_list, "</filter>", XML_TAG);

	pthread_rwlock_unlock(&filter_list->lock);

	return xml_list;
}


/*
free memory used by FILTER_LIST linked list
*/
void filter_free(FILTER_LIST * filter_list)
{
	filter_ll_free(filter_list->allow);
	filter_ll_free(filter_list->deny);
	filter_ll_free(filter_list->empty);

	FREE_AND_NULL(filter_list->dtemplate);
	pthread_rwlock_destroy(&filter_list->lock);

	xfree(filter_list);
}

void filter_ll_free(struct FILTER_LIST_LIST *fl)
{
	struct FILTER_LIST_LIST *tmp;

	while (fl != NULL) {
		tmp = fl->next;

		if (fl->he != NULL)
			reg_free(fl->he);
		if (fl->fe != NULL)
			reg_free(fl->fe);
		if (fl->profiles != NULL)
			array_free(fl->profiles);
		FREE_AND_NULL(fl->comment);
		FREE_AND_NULL(fl->host);
		FREE_AND_NULL(fl->file);
		FREE_AND_NULL(fl->template);

		xfree(fl);
		fl = tmp;
	}
}

/*
determine whether or not the host and/or file is allowed to be retrieved
based on the rules given in the filter_list linked list
*/
struct FILTER_LIST_LIST *filter_check(FILTER_LIST * filter_list, CONNECTION * connection)
{
	int action = FALSE, result = TRUE, ret, i;
	struct FILTER_LIST_LIST *current, *match = NULL;

	pthread_rwlock_rdlock(&filter_list->lock);

	if (filter_list->enabled == FALSE || connection->bypass & FEATURE_FILTER)
		return NULL;

	for (i = 0; i < 2; i++) {
		if (i == 0) {
			if (filter_list->policy == POLICY_ALLOW) {
				current = filter_list->deny;
				action = FALSE;
				result = TRUE;
			} else {
				current = filter_list->allow;
				action = TRUE;
				result = FALSE;
			}
		} else if (result == action) {
			if (filter_list->policy == POLICY_ALLOW) {
				current = filter_list->allow;
				action = TRUE;
			} else {
				current = filter_list->deny;
				action = FALSE;
			}
		} else
			break;

		for (; current != NULL; current = current->next) {
			if (current->enabled == FALSE)
				continue;

			if (!profile_find(connection->profiles, current->profiles))
				continue;

			if (current->he != NULL) {
				ret = reg_exec(current->he, connection->header->host);
				if (ret)
					continue;
			}

			if (current->fe != NULL) {
				ret = reg_exec(current->fe, connection->header->file);
				if (ret)
					continue;
			}

			match = current;
			result = action;
			break;
		}
	}

	return (!result) ? (match != NULL) ? match : filter_list->empty : NULL;
}

void filter_ll_insert(struct FILTER_LIST_LIST *x, char *profiles, char *comment, char *host, char *file, char *template)
{
	if (profiles != NULL) {
		if (x->profiles != NULL)
			array_free(x->profiles);

		if (strcmp(profiles, ""))
			x->profiles = string_break(profiles, ',');
		else
			x->profiles = NULL;
	}
	if (comment != NULL) {
		FREE_AND_NULL(x->comment);

		if (strcmp(comment, ""))
			x->comment = xstrdup(comment);
	}
	if (host != NULL) {
		if (x->he != NULL)
			reg_free(x->he);
		FREE_AND_NULL(x->host);

		if (strcmp(host, "")) {
			x->he = reg_compile(host, REGFLAGS);
			x->host = xstrdup(host);
		} else
			x->he = NULL;
	}

	if (file != NULL) {
		if (x->fe != NULL)
			reg_free(x->fe);
		FREE_AND_NULL(x->file);

		if (strcmp(file, "")) {
			x->fe = reg_compile(file, REGFLAGS);
			x->file = xstrdup(file);
		} else
			x->fe = NULL;
	}

	if (template != NULL) {
		FREE_AND_NULL(x->template);

		if (strcmp(template, ""))
			x->template = xstrdup(template);
	}
}

struct FILTER_LIST_LIST *filter_ll_delete(struct FILTER_LIST_LIST *x)
{
	struct FILTER_LIST_LIST *start = x;

	while (start->prev != NULL)
		start = start->prev;

	if (x->next != NULL)
		x->next->prev = x->prev;
	if (x->prev != NULL)
		x->prev->next = x->next;
	else
		start = start->next;

	if (x->he != NULL)
		reg_free(x->he);
	if (x->fe != NULL)
		reg_free(x->fe);
	if (x->profiles != NULL)
		array_free(x->profiles);
	FREE_AND_NULL(x->comment);
	FREE_AND_NULL(x->host);
	FREE_AND_NULL(x->file);
	FREE_AND_NULL(x->template);

	xfree(x);

	return start;
}

struct FILTER_LIST_LIST *filter_ll_new(struct FILTER_LIST_LIST *x)
{
	if (x == NULL) {
		x = xmalloc(sizeof(struct FILTER_LIST_LIST));
		x->prev = NULL;
	} else {
		while (x->next != NULL)
			x = x->next;
		x->next = xmalloc(sizeof(struct FILTER_LIST_LIST));
		x->next->prev = x;
		x = x->next;
	}
	x->next = NULL;
	x->enabled = TRUE;
	x->comment = NULL;
	x->profiles = NULL;
	x->he = NULL;
	x->fe = NULL;
	x->host = NULL;
	x->file = NULL;
	x->template = NULL;

	return x;
}
