/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include "proto.h"

extern GLOBAL *global;

KEYWORD_LIST *keyword_load(KEYWORD_LIST * keyword_list, XML_LIST * xml_list)
{
	KEYWORD_LIST *tmp_list = keyword_list;
	struct KEYWORD_LIST_LIST *keywords = NULL;

	if (tmp_list == NULL) {
		tmp_list = xmalloc(sizeof(KEYWORD_LIST));
		tmp_list->keyword_list = NULL;
		tmp_list->threshold = 0;
		tmp_list->template = NULL;
		tmp_list->id = 0;
		tmp_list->enabled = TRUE;

		keyword_list = tmp_list;

		pthread_rwlock_init(&tmp_list->lock, NULL);
	} else
		keywords = tmp_list->keyword_list;

	while ((xml_list = xml_section(xml_list, "<keywords>"))) {
		XML_LIST_LOOP(xml_list, "<keywords>") {
			XML_LIST_CMP(xml_list, "<item>") {
				keywords = keyword_list_new(keywords);
				keywords->id = tmp_list->id++;

				if (tmp_list->keyword_list == NULL)
					tmp_list->keyword_list = keywords;

				XML_LIST_LOOP(xml_list, "<item>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								keywords->enabled = FALSE;
							else
								keywords->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							keyword_list_insert(keywords, NULL, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							keyword_list_insert(keywords, xml_list->item, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<mime>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							keyword_list_insert(keywords, NULL, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<keyword>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							keyword_list_insert(keywords, NULL, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<score>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							keyword_list_insert(keywords, NULL, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<size>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							keyword_list_insert(keywords, NULL, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}
			XML_LIST_CMP(xml_list, "<enabled>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "false"))
						tmp_list->enabled = FALSE;
					else
						tmp_list->enabled = TRUE;
				}
			}
			XML_LIST_CMP(xml_list, "<threshold>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE)
					tmp_list->threshold = atoi(xml_list->item);
			}
			XML_LIST_CMP(xml_list, "<template>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					FREE_AND_NULL(tmp_list->template);
					tmp_list->template = xstrdup(xml_list->item);
				}
			}
		}
	}

	return tmp_list;
}

XML_LIST *keyword_xml(KEYWORD_LIST * keyword_list, XML_LIST * xml_list)
{
	char *ptr, buf[128];
	struct KEYWORD_LIST_LIST *kl;

	pthread_rwlock_rdlock(&keyword_list->lock);

	xml_list = xml_list_add(xml_list, "<keywords>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
	xml_list = xml_list_add(xml_list, (keyword_list->enabled == TRUE) ? "true" : "false", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<threshold>", XML_TAG);
	snprintf(buf, sizeof(buf), "%d", keyword_list->threshold);
	xml_list = xml_list_add(xml_list, buf, XML_VALUE);
	xml_list = xml_list_add(xml_list, "</threshold>", XML_TAG);

	if (keyword_list->template != NULL) {
		xml_list = xml_list_add(xml_list, "<template>", XML_TAG);
		ptr = string_to_xml(keyword_list->template);
		xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
		xfree(ptr);
		xml_list = xml_list_add(xml_list, "</template>", XML_TAG);
	}

	for (kl = keyword_list->keyword_list; kl; kl = kl->next) {
		xml_list = xml_list_add(xml_list, "<item>", XML_TAG);

		xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
		xml_list = xml_list_add(xml_list, (kl->enabled == TRUE) ? "true" : "false", XML_VALUE);
		xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

		if (kl->comment != NULL) {
			xml_list = xml_list_add(xml_list, "<comment>", XML_TAG);
			ptr = string_to_xml(kl->comment);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</comment>", XML_TAG);
		}
		if (kl->profiles != NULL) {
			xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);
			ptr = array_merge(kl->profiles, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);
		}
		if (kl->mime != NULL) {
			xml_list = xml_list_add(xml_list, "<mime>", XML_TAG);
			ptr = string_to_xml(kl->mime);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</mime>", XML_TAG);
		}

		if (kl->keyword != NULL) {
			xml_list = xml_list_add(xml_list, "<keyword>", XML_TAG);
			ptr = string_to_xml(kl->keyword);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</keyword>", XML_TAG);
		}

		if (kl->size != 0) {
			xml_list = xml_list_add(xml_list, "<size>", XML_TAG);
                        snprintf(buf, sizeof(buf), "%d", kl->size);
			xml_list = xml_list_add(xml_list, buf, XML_VALUE);
			xml_list = xml_list_add(xml_list, "</size>", XML_TAG);
		}

		xml_list = xml_list_add(xml_list, "<score>", XML_TAG);
		snprintf(buf, sizeof(buf), "%d", kl->score);
		xml_list = xml_list_add(xml_list, buf, XML_VALUE);
		xml_list = xml_list_add(xml_list, "</score>", XML_TAG);

		xml_list = xml_list_add(xml_list, "</item>", XML_TAG);
	}

	xml_list = xml_list_add(xml_list, "</keywords>", XML_TAG);

	pthread_rwlock_unlock(&keyword_list->lock);

	return xml_list;
}

struct KEYWORD_LIST_LIST *keyword_list_new(struct KEYWORD_LIST_LIST *x)
{
	if (x == NULL) {
		x = xmalloc(sizeof(struct KEYWORD_LIST_LIST));
		x->prev = NULL;
	} else {
		while (x->next != NULL)
			x = x->next;
		x->next = xmalloc(sizeof(struct KEYWORD_LIST_LIST));
		x->next->prev = x;
		x = x->next;
	}

	x->enabled = TRUE;
	x->comment = NULL;
	x->profiles = NULL;
	x->score = 0;
	x->mime = NULL;
	x->keyword = NULL;
	x->me = NULL;
	x->ke = NULL;
	x->size = 0;
	x->next = NULL;

	return x;
}

void keyword_list_insert(struct KEYWORD_LIST_LIST *x, char *profiles, char *comment, char *mime, char *keyword, char *score, char *size)
{
	if (profiles != NULL) {
		if (x->profiles != NULL)
			array_free(x->profiles);

		if (strcmp(profiles, ""))
			x->profiles = string_break(profiles, ',');
		else
			x->profiles = NULL;
	}
	if (comment != NULL) {
		FREE_AND_NULL(x->comment);

		if (strcmp(comment, ""))
			x->comment = xstrdup(comment);
	}
	if (mime != NULL) {
		if (x->me != NULL)
			reg_free(x->me);
		FREE_AND_NULL(x->mime);

		if (strcmp(mime, "")) {
			x->me = reg_compile(mime, REGFLAGS);
			x->mime = xstrdup(mime);
		} else
			x->me = NULL;
	}
	if (keyword != NULL) {
		if (x->ke != NULL)
			reg_free(x->ke);
		FREE_AND_NULL(x->keyword);

		if (strcmp(keyword, "")) {
			x->ke = reg_compile(keyword, REGFLAGS);
			x->keyword = xstrdup(keyword);
		} else
			x->ke = NULL;
	}
	if (score != NULL) {
		x->score = 0;

		if (strcmp(score, ""))
			x->score = atoi(score);
	}
	if (size != NULL) {
		x->size = 0;

		if (strcmp(size, ""))
	                x->size = string_to_size(size);
	}
}

struct KEYWORD_LIST_LIST *keyword_list_delete(struct KEYWORD_LIST_LIST *x)
{
	struct KEYWORD_LIST_LIST *start = x;

	while (start->prev != NULL)
		start = start->prev;

	if (x->next != NULL)
		x->next->prev = x->prev;
	if (x->prev != NULL)
		x->prev->next = x->next;
	else
		start = start->next;

	if (x->me != NULL)
		reg_free(x->me);
	if (x->ke != NULL)
		reg_free(x->ke);
	if (x->profiles != NULL)
		array_free(x->profiles);
	FREE_AND_NULL(x->comment);
	FREE_AND_NULL(x->mime);
	FREE_AND_NULL(x->keyword);

	xfree(x);

	return start;
}

void keyword_list_free(struct KEYWORD_LIST_LIST *kl)
{
	struct KEYWORD_LIST_LIST *tmp;

	while (kl != NULL) {
		tmp = kl->next;

		if (kl->me != NULL)
			reg_free(kl->me);
		if (kl->ke != NULL)
			reg_free(kl->ke);
		if (kl->profiles != NULL)
			array_free(kl->profiles);
		FREE_AND_NULL(kl->comment);
		FREE_AND_NULL(kl->mime);
		FREE_AND_NULL(kl->keyword);

		xfree(kl);

		kl = tmp;
	}
}

void keyword_free(KEYWORD_LIST * keyword_list)
{
	if (keyword_list == NULL)
		return;

	keyword_list_free(keyword_list->keyword_list);

	FREE_AND_NULL(keyword_list->template);

	pthread_rwlock_destroy(&keyword_list->lock);

	xfree(keyword_list);
}

int keyword_check(KEYWORD_LIST * keyword_list, CONNECTION * connection, FILEBUF * filebuf, int action)
{
	int ret, x = FALSE, maxbuffer;
	struct KEYWORD_LIST_LIST *kl;

	if (!keyword_list || connection->bypass & FEATURE_KEYWORDS)
		return FALSE;

	pthread_rwlock_rdlock(&keyword_list->lock);

	if (keyword_list->enabled == FALSE) {
		pthread_rwlock_unlock(&keyword_list->lock);

		return FALSE;
	}

	if (filebuf != NULL)
		filebuf_add(filebuf, "", 1);

        maxbuffer = atomic_read_rwlock(&global->general->lock, &global->general->maxbuffer);

	for (kl = keyword_list->keyword_list; kl; kl = kl->next) {
		if (kl->enabled == FALSE)
			continue;

		if (!profile_find(connection->profiles, kl->profiles))
			continue;

		if (kl->me != NULL && connection->rheader->content_type != NULL) {
			ret = reg_exec(kl->me, connection->rheader->content_type);
			if (ret)
				continue;
		} else if (kl->me != NULL) {
			ret = reg_exec(kl->me, "");
			if (ret)
				continue;
		}

                if (kl->size != 0) {
                    ret = (kl->size == -1)? maxbuffer : kl->size;
                        
                    if (action == FALSE) {
                        if (connection->rheader->content_length > ret)
                            continue;
                    } else {
                        if (connection->rheader->content_length == -1 || connection->rheader->content_length > ret)
                            continue;
                    }
                }

		if (kl->ke != NULL && filebuf != NULL) {
			ret = reg_exec(kl->ke, filebuf->data);
			if (ret)
				continue;
		}

		if (action == FALSE) {
			/* just check if any entires apply to current page */
			x = TRUE;

			break;
		}

		x += kl->score;
	}

	if (filebuf != NULL)
		filebuf_resize(filebuf, filebuf->size - 1);

	pthread_rwlock_unlock(&keyword_list->lock);

	return x;
}
