/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include "proto.h"

LIMITS_LIST *limits_load(LIMITS_LIST * limits_list, XML_LIST * xml_list)
{
	LIMITS_LIST *tmp_list = limits_list;
	struct LIMITS_LIST_LIST *ll = NULL;
	struct tm from, to;
	int totime, fromtime;

	if (tmp_list == NULL) {
		tmp_list = xmalloc(sizeof(LIMITS_LIST));
		tmp_list->limits_list = NULL;
		tmp_list->enabled = TRUE;
		tmp_list->id = 0;

		limits_list = tmp_list;

		pthread_rwlock_init(&tmp_list->lock, NULL);
	} else
		ll = limits_list->limits_list;

	while ((xml_list = xml_section(xml_list, "<limits>"))) {
		XML_LIST_LOOP(xml_list, "<limits>") {
			XML_LIST_CMP(xml_list, "<item>") {
				ll = limits_ll_new(ll);
				ll->id = tmp_list->id++;

				if (limits_list->limits_list == NULL)
					limits_list->limits_list = ll;

				memset(&from, 0, sizeof(struct tm));
				memset(&to, 0, sizeof(struct tm));
				fromtime = totime = FALSE;

				XML_LIST_LOOP(xml_list, "<item>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								ll->enabled = FALSE;
							else
								ll->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<action>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "allow"))
								ll->action = POLICY_ALLOW;
							else
								ll->action = POLICY_DENY;
						}
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							limits_ll_insert(ll, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							limits_ll_insert(ll, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<template>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							limits_ll_insert(ll, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<fromhour>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							fromtime = TRUE;
							from.tm_hour = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<frommin>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							fromtime = TRUE;
							from.tm_min = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<frommday>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							fromtime = TRUE;
							from.tm_mday = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<frommon>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							fromtime = TRUE;
							from.tm_mon = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<fromwday>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							fromtime = TRUE;
							from.tm_wday = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<tohour>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							totime = TRUE;
							to.tm_hour = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<tomin>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							totime = TRUE;
							to.tm_min = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<tomday>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							totime = TRUE;
							to.tm_mday = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<tomon>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							totime = TRUE;
							to.tm_mon = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<towday>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							totime = TRUE;
							to.tm_wday = strtol(xml_list->item, NULL, 10);
						}
					}
					XML_LIST_CMP(xml_list, "<maxbytes>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							limits_ll_insert(ll, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<maxrequests>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							limits_ll_insert(ll, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<flags>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							limits_ll_insert(ll, NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}

				if (fromtime == TRUE)
					limits_ll_insert(ll, NULL, NULL, NULL, &from, NULL, NULL, NULL, NULL);
				if (totime == TRUE)
					limits_ll_insert(ll, NULL, NULL, NULL, NULL, &to, NULL, NULL, NULL);
			}
			XML_LIST_CMP(xml_list, "<enabled>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "false"))
						limits_list->enabled = FALSE;
					else
						limits_list->enabled = TRUE;
				}
			}
		}
	}


	return limits_list;
}

XML_LIST *limits_xml(LIMITS_LIST * limits_list, XML_LIST * xml_list)
{
	char *ptr, buf[128];
	struct LIMITS_LIST_LIST *ll;

	pthread_rwlock_rdlock(&limits_list->lock);

	xml_list = xml_list_add(xml_list, "<limits>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
	xml_list = xml_list_add(xml_list, (limits_list->enabled == TRUE) ? "true" : "false", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

	for (ll = limits_list->limits_list; ll; ll = ll->next) {
		xml_list = xml_list_add(xml_list, "<item>", XML_TAG);

		if (ll->comment != NULL) {
			xml_list = xml_list_add(xml_list, "<comment>", XML_TAG);
			ptr = string_to_xml(ll->comment);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</comment>", XML_TAG);
		}

		if (ll->profiles != NULL) {
			xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);
			ptr = array_merge(ll->profiles, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);
		}

		xml_list = xml_list_add(xml_list, "<action>", XML_TAG);
		xml_list = xml_list_add(xml_list, (ll->action == POLICY_ALLOW) ? "allow" : "deny", XML_VALUE);
		xml_list = xml_list_add(xml_list, "</action>", XML_TAG);

		if (ll->template != NULL) {
			xml_list = xml_list_add(xml_list, "<template>", XML_TAG);
			ptr = string_to_xml(ll->template);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</template>", XML_TAG);
		}

		if (ll->from != NULL) {
			if (ll->from->tm_hour != 0) {
				xml_list = xml_list_add(xml_list, "<fromhour>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->from->tm_hour);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</fromhour>", XML_TAG);
			}

			if (ll->from->tm_min != 0) {
				xml_list = xml_list_add(xml_list, "<frommin>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->from->tm_min);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</frommin>", XML_TAG);
			}

			if (ll->from->tm_mday != 0) {
				xml_list = xml_list_add(xml_list, "<frommday>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->from->tm_mday);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</frommday>", XML_TAG);
			}

			if (ll->from->tm_mon != 0) {
				xml_list = xml_list_add(xml_list, "<frommon>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->from->tm_mon);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</frommon>", XML_TAG);
			}

			if (ll->from->tm_wday != 0) {
				xml_list = xml_list_add(xml_list, "<fromwday>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->from->tm_wday);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</fromwday>", XML_TAG);
			}
		}

		if (ll->to != NULL) {
			if (ll->to->tm_hour != 0) {
				xml_list = xml_list_add(xml_list, "<tohour>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->to->tm_hour);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</tohour>", XML_TAG);
			}

			if (ll->to->tm_min != 0) {
				xml_list = xml_list_add(xml_list, "<tomin>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->to->tm_min);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</tomin>", XML_TAG);
			}

			if (ll->to->tm_mday != 0) {
				xml_list = xml_list_add(xml_list, "<tomday>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->to->tm_mday);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</tomday>", XML_TAG);
			}

			if (ll->to->tm_mon != 0) {
				xml_list = xml_list_add(xml_list, "<tomon>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->to->tm_mon);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</tomon>", XML_TAG);
			}

			if (ll->to->tm_wday != 0) {
				xml_list = xml_list_add(xml_list, "<towday>", XML_TAG);
				snprintf(buf, sizeof(buf), "%d", ll->to->tm_wday);
				xml_list = xml_list_add(xml_list, buf, XML_VALUE);
				xml_list = xml_list_add(xml_list, "</towday>", XML_TAG);
			}
		}

		if (ll->maxbytes != 0) {
			xml_list = xml_list_add(xml_list, "<maxbytes>", XML_TAG);
			snprintf(buf, sizeof(buf), "%u", ll->maxbytes);
			xml_list = xml_list_add(xml_list, buf, XML_VALUE);
			xml_list = xml_list_add(xml_list, "</maxbytes>", XML_TAG);
		}

		if (ll->maxrequests != 0) {
			xml_list = xml_list_add(xml_list, "<maxrequests>", XML_TAG);
			snprintf(buf, sizeof(buf), "%u", ll->maxrequests);
			xml_list = xml_list_add(xml_list, buf, XML_VALUE);
			xml_list = xml_list_add(xml_list, "</maxrequests>", XML_TAG);
		}

		if (ll->flags != 0) {
			xml_list = xml_list_add(xml_list, "<flags>", XML_TAG);
			snprintf(buf, sizeof(buf), "%s,%s,%s", (ll->flags & LIMIT_CACHE) ? "limitcache" : "", (ll->flags & LIMIT_ABSOLUTETIME) ? "absolutetime" : "", (ll->flags & LIMIT_ALLRANGETIME) ? "allrangetime" : "");
			xml_list = xml_list_add(xml_list, buf, XML_VALUE);
			xml_list = xml_list_add(xml_list, "</flags>", XML_TAG);
		}

		xml_list = xml_list_add(xml_list, "</item>", XML_TAG);
	}

	xml_list = xml_list_add(xml_list, "</limits>", XML_TAG);

	pthread_rwlock_unlock(&limits_list->lock);

	return xml_list;
}

void limits_ll_insert(struct LIMITS_LIST_LIST *x, char *profiles, char *comment, char *template, struct tm *from, struct tm *to, char *maxbytes, char *maxrequests, char *flags)
{
	int i;
	char **args;

	if (profiles != NULL) {
		if (x->profiles != NULL)
			array_free(x->profiles);

		if (strcmp(profiles, ""))
			x->profiles = string_break(profiles, ',');
		else
			x->profiles = NULL;
	}

	if (comment != NULL) {
		FREE_AND_NULL(x->comment);

		if (strcmp(comment, ""))
			x->comment = xstrdup(comment);
	}

	if (template != NULL) {
		FREE_AND_NULL(x->template);

		if (strcmp(template, ""))
			x->template = xstrdup(template);
	}

	if (from != NULL) {
		FREE_AND_NULL(x->from);

		x->from = xmalloc(sizeof(struct tm));
		*x->from = *from;
	}

	if (to != NULL) {
		FREE_AND_NULL(x->to);

		x->to = xmalloc(sizeof(struct tm));
		*x->to = *to;
	}

	if (maxbytes != NULL) {
		x->maxbytes = 0;

		if (strcmp(maxbytes, ""))
			x->maxbytes = strtoll(maxbytes, NULL, 10);
	}

	if (maxrequests != NULL) {
		x->maxrequests = 0;

		if (strcmp(maxrequests, ""))
			x->maxrequests = strtoll(maxrequests, NULL, 10);
	}

	if (flags != NULL) {
		x->flags = LIMIT_ABSOLUTETIME;;

		if (strcmp(flags, "")) {
			args = string_break(flags, ',');
			for (i = 0; args[i]; i++) {
				if (!strcasecmp(args[i], "limitcache"))
					x->flags |= LIMIT_CACHE;
				else if (!strcasecmp(args[i], "absolutetime")) {
					x->flags &= ~LIMIT_ALLRANGETIME;
					x->flags |= LIMIT_ABSOLUTETIME;
				} else if (!strcasecmp(args[i], "allrangetime")) {
					x->flags &= ~LIMIT_ABSOLUTETIME;
					x->flags |= LIMIT_ALLRANGETIME;
				}

				xfree(args[i]);
			}
			xfree(args);
		}
	}	
		
}

struct LIMITS_LIST_LIST *limits_ll_new(struct LIMITS_LIST_LIST *x)
{
	if (x == NULL) {
		x = xmalloc(sizeof(struct LIMITS_LIST_LIST));
		x->prev = NULL;
	} else {
		while (x->next != NULL)
			x = x->next;
		x->next = xmalloc(sizeof(struct LIMITS_LIST_LIST));
		x->next->prev = x;
		x = x->next;
	}

	x->next = NULL;
	x->enabled = TRUE;
	x->profiles = NULL;
	x->action = POLICY_ALLOW;
	x->comment = NULL;
	x->template = NULL;
	x->from = NULL;
	x->to = NULL;
	x->maxrequests = 0;
	x->maxbytes = 0;
	x->n_requests = 0;
	x->n_bytes = 0;
	x->flags = LIMIT_ABSOLUTETIME;

	return x;
}

struct LIMITS_LIST_LIST *limits_ll_delete(struct LIMITS_LIST_LIST *x)
{
	struct LIMITS_LIST_LIST *start = x;

	while (start->prev != NULL)
		start = start->prev;

	if (x->next != NULL)
		x->next->prev = x->prev;
	if (x->prev != NULL)
		x->prev->next = x->next;
	else
		start = x->next;

	if (x->profiles != NULL)
		array_free(x->profiles);
	FREE_AND_NULL(x->comment);
	FREE_AND_NULL(x->template);
	FREE_AND_NULL(x->to);
	FREE_AND_NULL(x->from);

	xfree(x);

	return start;
}

void limits_ll_free(struct LIMITS_LIST_LIST *x)
{
	struct LIMITS_LIST_LIST *tmp;

	while (x != NULL) {
		tmp = x->next;

		if (x->profiles != NULL)
			array_free(x->profiles);
		FREE_AND_NULL(x->comment);
		FREE_AND_NULL(x->template);
		FREE_AND_NULL(x->to);
		FREE_AND_NULL(x->from);

		xfree(x);

		x = tmp;
	}
}

void limits_free(LIMITS_LIST * limits_list)
{
	if (!limits_list)
		return;

	limits_ll_free(limits_list->limits_list);

	pthread_rwlock_destroy(&limits_list->lock);

	xfree(limits_list);
}

struct LIMITS_LIST_LIST *limits_check(LIMITS_LIST * limits_list, CONNECTION * connection)
{
	struct LIMITS_LIST_LIST *ll, *match = NULL;
	time_t curtime;
	struct tm lt;

	/* caller must always unlock this */
	pthread_rwlock_rdlock(&limits_list->lock);

	if (!limits_list || (connection->bypass & FEATURE_LIMITS))
		return NULL;

	curtime = time(NULL);
	localtime_r(&curtime, &lt);

	if (limits_list->enabled == FALSE)
		return NULL;

	for (ll = limits_list->limits_list; ll; ll = ll->next) {
		if (ll->enabled == FALSE)
			continue;

		if (!profile_find(connection->profiles, ll->profiles))
			continue;

		if (ll->from != NULL && ll->to != NULL) {
			if (ll->flags & LIMIT_ABSOLUTETIME) {
				if (!in_absolutetimerange(&lt, ll->from, ll->to))
					continue;
			} else {
				if (!in_alltimerange(&lt, ll->from, ll->to))
					continue;
			}
		}

		match = ll;
	}

	return match;
}

void limits_update(LIMITS_LIST * limits_list, CONNECTION * connection)
{
	struct LIMITS_LIST_LIST *ll;
	time_t curtime;
	struct tm lt;

	if (!limits_list || (connection->bypass & FEATURE_LIMITS))
		return;

	curtime = time(NULL);
	localtime_r(&curtime, &lt);

	pthread_rwlock_rdlock(&limits_list->lock);

	if (limits_list->enabled == FALSE) {
		pthread_rwlock_unlock(&limits_list->lock);

		return;
	}

	for (ll = limits_list->limits_list; ll; ll = ll->next) {
		if (ll->enabled == FALSE)
			continue;

		if (!profile_find(connection->profiles, ll->profiles))
			continue;

		if (ll->from != NULL && ll->to != NULL) {
			if (ll->flags & LIMIT_ABSOLUTETIME) {
				if (!in_absolutetimerange(&lt, ll->from, ll->to))
					continue;
			} else {
				if (!in_alltimerange(&lt, ll->from, ll->to))
					continue;
			}
		}

		if (ll->maxbytes != 0 && (connection->cachemap == NULL || (connection->flags & CONNECTION_CACHING) || (ll->flags & LIMIT_CACHE))) ll->n_bytes += connection->transferred;
		if (ll->maxrequests != 0) ll->n_requests++;
	}

	pthread_rwlock_unlock(&limits_list->lock);
}


void limits_reset(LIMITS_LIST * limits_list)
{
	struct LIMITS_LIST_LIST *ll;
	time_t curtime;
	struct tm lt;

	if (!limits_list)
		return;

	curtime = time(NULL);
	localtime_r(&curtime, &lt);

	pthread_rwlock_rdlock(&limits_list->lock);

	if (limits_list->enabled == FALSE) {
		pthread_rwlock_unlock(&limits_list->lock);

		return;
	}

	for (ll = limits_list->limits_list; ll; ll = ll->next) {
		if (ll->enabled == FALSE)
			continue;

		if (ll->n_bytes == 0 && ll->n_requests == 0) continue;

		if (ll->from != NULL && ll->to != NULL) {
			if (ll->flags & LIMIT_ABSOLUTETIME) {
				if (in_absolutetimerange(&lt, ll->from, ll->to))
					continue;
			} else {
				if (in_alltimerange(&lt, ll->from, ll->to))
					continue;
			}
		} else continue;

		/* ok, we are outside the time range this entry applies to, we can reset the
		   byte and request counter */
		ll->n_bytes = ll->n_requests = 0;
	}

	pthread_rwlock_unlock(&limits_list->lock);
}


