/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "proto.h"

/*
load <mime> section from XML file into MIME_LIST-type struct
*/
MIME_LIST *mime_load(MIME_LIST * mime_list, XML_LIST * xml_list)
{
	MIME_LIST *tmp_list = mime_list;
	struct MIME_LIST_LIST *allow = NULL, *deny = NULL;

	if (tmp_list == NULL) {
		tmp_list = xmalloc(sizeof(MIME_LIST));
		tmp_list->dtemplate = NULL;
		tmp_list->allow = NULL;
		tmp_list->deny = NULL;
		tmp_list->policy = POLICY_ALLOW;
		tmp_list->empty = mime_ll_new(NULL);
		tmp_list->id = 0;
		tmp_list->enabled = TRUE;

		mime_list = tmp_list;

		pthread_rwlock_init(&tmp_list->lock, NULL);
	} else {
		allow = tmp_list->allow;
		deny = tmp_list->deny;
	}

	while ((xml_list = xml_section(xml_list, "<mime>"))) {
		XML_LIST_LOOP(xml_list, "<mime>") {
			XML_LIST_CMP(xml_list, "<allow>") {
				allow = mime_ll_new(allow);
				allow->id = mime_list->id++;

				if (tmp_list->allow == NULL)
					tmp_list->allow = allow;
				XML_LIST_LOOP(xml_list, "<allow>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								allow->enabled = FALSE;
							else
								allow->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(allow, NULL, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(allow, xml_list->item, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<template>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(allow, NULL, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<host>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(allow, NULL, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<mime>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(allow, NULL, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<file>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(allow, NULL, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}
			XML_LIST_CMP(xml_list, "<deny>") {
				deny = mime_ll_new(deny);
				deny->id = mime_list->id++;

				if (tmp_list->deny == NULL)
					tmp_list->deny = deny;
				XML_LIST_LOOP(xml_list, "<deny>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								deny->enabled = FALSE;
							else
								deny->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(deny, NULL, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(deny, xml_list->item, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<template>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(deny, NULL, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<host>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(deny, NULL, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<mime>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(deny, NULL, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<file>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							mime_ll_insert(deny, NULL, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}
			XML_LIST_CMP(xml_list, "<policy>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "allow"))
						tmp_list->policy = POLICY_ALLOW;
					else if (!strcasecmp(xml_list->item, "deny"))
						tmp_list->policy = POLICY_DENY;
				}
			}
			XML_LIST_CMP(xml_list, "<default>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					FREE_AND_NULL(tmp_list->dtemplate);
					tmp_list->dtemplate = xstrdup(xml_list->item);
				}
			}
			XML_LIST_CMP(xml_list, "<enabled>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "false"))
						tmp_list->enabled = FALSE;
					else
						tmp_list->enabled = TRUE;
				}
			}
		}
	}

	return mime_list;
}

XML_LIST *mime_xml(MIME_LIST * mime_list, XML_LIST * xml_list)
{
	int i;
	char *ptr;
	struct MIME_LIST_LIST *ml = NULL;

	if (mime_list == NULL)
		return xml_list;

	pthread_rwlock_rdlock(&mime_list->lock);

	xml_list = xml_list_add(xml_list, "<mime>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
	xml_list = xml_list_add(xml_list, (mime_list->enabled == TRUE) ? "true" : "false", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<policy>", XML_TAG);
	xml_list = xml_list_add(xml_list, (mime_list->policy == POLICY_ALLOW) ? "allow" : "deny", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</policy>", XML_TAG);

	if (mime_list->dtemplate != NULL) {
		xml_list = xml_list_add(xml_list, "<default>", XML_TAG);
		ptr = string_to_xml(mime_list->dtemplate);
		xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
		xfree(ptr);
		xml_list = xml_list_add(xml_list, "</default>", XML_TAG);
	}

	for (i = 0; i < 2; i++) {
		switch (i) {
		case 0:
			ml = mime_list->allow;
			break;
		case 1:
			ml = mime_list->deny;
			break;
		}

		for (; ml; ml = ml->next) {
			xml_list = xml_list_add(xml_list, (i == 0) ? "<allow>" : "<deny>", XML_TAG);

			xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
			xml_list = xml_list_add(xml_list, (ml->enabled == TRUE) ? "true" : "false", XML_VALUE);
			xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

			if (ml->comment != NULL) {
				xml_list = xml_list_add(xml_list, "<comment>", XML_TAG);
				ptr = string_to_xml(ml->comment);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</comment>", XML_TAG);
			}

			if (ml->profiles != NULL) {
				xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);
				ptr = array_merge(ml->profiles, ',');
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);
			}

			if (ml->mime != NULL) {
				xml_list = xml_list_add(xml_list, "<mime>", XML_TAG);
				ptr = string_to_xml(ml->mime);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</mime>", XML_TAG);
			}

			if (ml->host != NULL) {
				xml_list = xml_list_add(xml_list, "<host>", XML_TAG);
				ptr = string_to_xml(ml->host);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</host>", XML_TAG);
			}

			if (ml->file != NULL) {
				xml_list = xml_list_add(xml_list, "<file>", XML_TAG);
				ptr = string_to_xml(ml->file);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</file>", XML_TAG);
			}

			if (ml->template != NULL) {
				xml_list = xml_list_add(xml_list, "<template>", XML_TAG);
				ptr = string_to_xml(ml->template);
				xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
				xfree(ptr);
				xml_list = xml_list_add(xml_list, "</template>", XML_TAG);
			}

			xml_list = xml_list_add(xml_list, (i == 0) ? "</allow>" : "</deny>", XML_TAG);
		}
	}

	xml_list = xml_list_add(xml_list, "</mime>", XML_TAG);

	pthread_rwlock_unlock(&mime_list->lock);

	return xml_list;
}

void mime_ll_insert(struct MIME_LIST_LIST *x, char *profiles, char *comment, char *template, char *host, char *file, char *mime)
{
	if (profiles != NULL) {
		if (x->profiles != NULL)
			array_free(x->profiles);

		if (strcmp(profiles, ""))
			x->profiles = string_break(profiles, ',');
		else
			x->profiles = NULL;
	}
	if (comment != NULL) {
		FREE_AND_NULL(x->comment);

		if (strcmp(comment, ""))
			x->comment = xstrdup(comment);
	}
	if (template != NULL) {
		FREE_AND_NULL(x->template);

		if (strcmp(template, ""))
			x->template = xstrdup(template);
	}
	if (host != NULL) {
		if (x->he != NULL)
			reg_free(x->he);
		FREE_AND_NULL(x->host);

		if (strcmp(host, "")) {
			x->host = xstrdup(host);
			x->he = reg_compile(host, REGFLAGS);
		} else
			x->he = NULL;
	}
	if (mime != NULL) {
		if (x->me != NULL)
			reg_free(x->me);
		FREE_AND_NULL(x->mime);

		if (strcmp(mime, "")) {
			x->mime = xstrdup(mime);
			x->me = reg_compile(mime, REGFLAGS);
		} else
			x->me = NULL;
	}
	if (file != NULL) {
		if (x->fe != NULL)
			reg_free(x->fe);

		if (strcmp(file, "")) {
			x->file = xstrdup(file);
			x->fe = reg_compile(file, REGFLAGS);
		} else
			x->fe = NULL;
	}
}

struct MIME_LIST_LIST *mime_ll_delete(struct MIME_LIST_LIST *x)
{
	struct MIME_LIST_LIST *start = x;

	while (start->prev != NULL)
		start = start->prev;

	if (x->next != NULL)
		x->next->prev = x->prev;
	if (x->prev != NULL)
		x->prev->next = x->next;
	else
		start = start->next;

	if (x->he != NULL)
		reg_free(x->he);
	if (x->fe != NULL)
		reg_free(x->fe);
	if (x->me != NULL)
		reg_free(x->me);
	if (x->profiles != NULL)
		array_free(x->profiles);
	FREE_AND_NULL(x->comment);
	FREE_AND_NULL(x->template);
	FREE_AND_NULL(x->host);
	FREE_AND_NULL(x->file);
	FREE_AND_NULL(x->mime);

	xfree(x);

	return start;
}

struct MIME_LIST_LIST *mime_ll_new(struct MIME_LIST_LIST *x)
{
	if (x == NULL) {
		x = xmalloc(sizeof(struct MIME_LIST_LIST));
		x->prev = NULL;
	} else {
		while (x->next != NULL)
			x = x->next;
		x->next = xmalloc(sizeof(struct MIME_LIST_LIST));
		x->next->prev = x;
		x = x->next;
	}

	x->enabled = TRUE;
	x->profiles = NULL;
	x->comment = NULL;
	x->template = NULL;
	x->he = NULL;
	x->fe = NULL;
	x->me = NULL;
	x->host = NULL;
	x->file = NULL;
	x->mime = NULL;
	x->next = NULL;

	return x;
}

/*
free memory used by a MIME_LIST-type struct
*/
void mime_free(MIME_LIST * mime_list)
{
	if (!mime_list)
		return;

	mime_ll_free(mime_list->allow);
	mime_ll_free(mime_list->deny);
	mime_ll_free(mime_list->empty);

	FREE_AND_NULL(mime_list->dtemplate);

	pthread_rwlock_destroy(&mime_list->lock);

	xfree(mime_list);
}

void mime_ll_free(struct MIME_LIST_LIST *ml)
{
	struct MIME_LIST_LIST *tmp;

	while (ml != NULL) {
		tmp = ml->next;

		if (ml->he != NULL)
			reg_free(ml->he);
		if (ml->fe != NULL)
			reg_free(ml->fe);
		if (ml->me != NULL)
			reg_free(ml->me);
		if (ml->profiles != NULL)
			array_free(ml->profiles);
		FREE_AND_NULL(ml->comment);
		FREE_AND_NULL(ml->host);
		FREE_AND_NULL(ml->file);
		FREE_AND_NULL(ml->mime);
		FREE_AND_NULL(ml->template);

		xfree(ml);
		ml = tmp;
	}
}


/*
check if a specific mime-type is allowed, and return pointer to list node
if not
*/
struct MIME_LIST_LIST *mime_check(MIME_LIST * mime_list, CONNECTION * connection)
{
	int action = FALSE, result = TRUE, ret, i;
	struct MIME_LIST_LIST *current, *match = NULL;

	pthread_rwlock_rdlock(&mime_list->lock);

	if (mime_list->enabled == FALSE || connection->bypass & FEATURE_MIME)
		return NULL;

	for (i = 0; i < 2; i++) {
		if (i == 0) {
			if (mime_list->policy == POLICY_ALLOW) {
				current = mime_list->deny;
				action = FALSE;
				result = TRUE;
			} else {
				current = mime_list->allow;
				action = TRUE;
				result = FALSE;
			}
		} else if (result == action) {
			if (mime_list->policy == POLICY_ALLOW) {
				current = mime_list->allow;
				action = TRUE;
			} else {
				current = mime_list->deny;
				action = FALSE;
			}
		} else
			break;

		for (; current != NULL; current = current->next) {
			if (current->enabled == FALSE)
				continue;

			if (!profile_find(connection->profiles, current->profiles))
				continue;

			if (current->he != NULL) {
				ret = reg_exec(current->he, connection->header->host);
				if (ret)
					continue;
			}

			if (current->fe != NULL) {
				ret = reg_exec(current->fe, connection->header->file);
				if (ret)
					continue;
			}

			if (current->me != NULL && connection->rheader->content_type != NULL) {
				ret = reg_exec(current->me, connection->rheader->content_type);
				if (ret)
					continue;
			} else if (current->me != NULL)
				continue;

			match = current;
			result = action;
			break;
		}
	}

	return (!result) ? (match != NULL) ? match : mime_list->empty : NULL;
}
