/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include "proto.h"

extern GLOBAL *global;

PROFILES_LIST *profiles_load(PROFILES_LIST * profiles_list, XML_LIST * xml_list)
{
	PROFILES_LIST *tmp_list = profiles_list;
	struct PROFILES_LIST_LIST *profiles = NULL;

	if (tmp_list == NULL) {
		tmp_list = xmalloc(sizeof(PROFILES_LIST));
		tmp_list->profiles = NULL;
		tmp_list->id = 0;

		profiles_list = tmp_list;

		pthread_rwlock_init(&tmp_list->lock, NULL);
	} else
		profiles = tmp_list->profiles;

	while ((xml_list = xml_section(xml_list, "<profiles>"))) {
		XML_LIST_LOOP(xml_list, "<profiles>") {
			XML_LIST_CMP(xml_list, "<item>") {
				profiles = profiles_list_new(profiles);
				profiles->id = profiles_list->id++;

				if (tmp_list->profiles == NULL)
					tmp_list->profiles = profiles;
				XML_LIST_LOOP(xml_list, "<item>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								profiles->enabled = FALSE;
							else
								profiles->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<host>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<file>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<addprofiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<removeprofiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<urlcommand>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<portrange>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							profiles_list_insert(profiles, NULL, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}
		}
	}

	return profiles_list;
}

XML_LIST *profiles_xml(PROFILES_LIST * profiles_list, XML_LIST * xml_list)
{
	char *ptr;
	struct PROFILES_LIST_LIST *pl;

	if (profiles_list == NULL)
		return xml_list;

	pthread_rwlock_rdlock(&profiles_list->lock);

	xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);

	for (pl = profiles_list->profiles; pl; pl = pl->next) {
		xml_list = xml_list_add(xml_list, "<item>", XML_TAG);

		xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
		xml_list = xml_list_add(xml_list, (pl->enabled == TRUE) ? "true" : "false", XML_VALUE);
		xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

		if (pl->comment != NULL) {
			xml_list = xml_list_add(xml_list, "<comment>", XML_TAG);
			ptr = string_to_xml(pl->comment);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</comment>", XML_TAG);
		}
		if (pl->profiles != NULL) {
			xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);
			ptr = array_merge(pl->profiles, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);
		}
		if (pl->host != NULL) {
			xml_list = xml_list_add(xml_list, "<host>", XML_TAG);
			ptr = string_to_xml(pl->host);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</host>", XML_TAG);
		}
		if (pl->file != NULL) {
			xml_list = xml_list_add(xml_list, "<file>", XML_TAG);
			ptr = string_to_xml(pl->file);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</file>", XML_TAG);
		}

		if (pl->addprofiles != NULL) {
			xml_list = xml_list_add(xml_list, "<addprofiles>", XML_TAG);
			ptr = array_merge(pl->addprofiles, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</addprofiles>", XML_TAG);
		}

		if (pl->removeprofiles != NULL) {
			xml_list = xml_list_add(xml_list, "<removeprofiles>", XML_TAG);
			ptr = array_merge(pl->removeprofiles, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</removeprofiles>", XML_TAG);
		}

		if (pl->urlcommand != NULL) {
			xml_list = xml_list_add(xml_list, "<urlcommand>", XML_TAG);
			ptr = array_merge(pl->urlcommand, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</urlcommand>", XML_TAG);
		}

		if (pl->portrange != NULL) {
			xml_list = xml_list_add(xml_list, "<portrange>", XML_TAG);
			ptr = string_to_xml(pl->portrange);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</portrange>", XML_TAG);
		}

		xml_list = xml_list_add(xml_list, "</item>", XML_TAG);
	}

	xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);

	pthread_rwlock_unlock(&profiles_list->lock);

	return xml_list;
}
void profiles_list_insert(struct PROFILES_LIST_LIST *x, char *profiles, char *comment, char *host, char *file, char *addprofiles, char *removeprofiles, char *urlcommand, char *portrange)
{
	if (profiles != NULL) {
		if (x->profiles != NULL)
			array_free(x->profiles);

		if (strcmp(profiles, ""))
			x->profiles = string_break(profiles, ',');
		else
			x->profiles = NULL;
	}

	if (comment != NULL) {
		FREE_AND_NULL(x->comment);

		if (strcmp(comment, ""))
			x->comment = xstrdup(comment);
	}

	if (host != NULL) {
		if (x->he != NULL)
			reg_free(x->he);
		FREE_AND_NULL(x->host);

		if (strcmp(host, "")) {
			x->he = reg_compile(host, REGFLAGS);
			x->host = xstrdup(host);
		} else
			x->he = NULL;

	}
	if (file != NULL) {
		if (x->fe != NULL)
			reg_free(x->fe);
		FREE_AND_NULL(x->file);

		if (strcmp(file, "")) {
			x->fe = reg_compile(file, REGFLAGS);
			x->file = xstrdup(file);
		} else
			x->fe = NULL;
	}

	if (addprofiles != NULL) {
		if (x->addprofiles != NULL)
			array_free(x->addprofiles);

		if (strcmp(addprofiles, ""))
			x->addprofiles = string_break(addprofiles, ',');
		else
			x->addprofiles = NULL;
	}

	if (removeprofiles != NULL) {
		if (x->removeprofiles != NULL)
			array_free(x->removeprofiles);

		if (strcmp(removeprofiles, ""))
			x->removeprofiles = string_break(removeprofiles, ',');
		else
			x->removeprofiles = NULL;
	}

       if (urlcommand != NULL) {
                if (x->urlcommand != NULL)
                        array_free(x->urlcommand);

                if (strcmp(urlcommand, ""))
                        x->urlcommand = string_break(urlcommand, ',');
                else
                        x->urlcommand = NULL;
        }

	if (portrange != NULL) {
		FREE_AND_NULL(x->portrange);

		if (strcmp(portrange, ""))
			x->portrange = xstrdup(portrange);
	}
}

struct PROFILES_LIST_LIST *profiles_list_new(struct PROFILES_LIST_LIST *x)
{
	if (x == NULL) {
		x = xmalloc(sizeof(struct PROFILES_LIST_LIST));
		x->prev = NULL;
	} else {
		while (x->next != NULL)
			x = x->next;
		x->next = xmalloc(sizeof(struct PROFILES_LIST_LIST));
		x->next->prev = x;
		x = x->next;
	}
	x->next = NULL;
	x->enabled = TRUE;
	x->comment = NULL;
	x->profiles = NULL;
	x->he = NULL;
	x->fe = NULL;
	x->host = NULL;
	x->file = NULL;
	x->addprofiles = NULL;
	x->removeprofiles = NULL;
	x->urlcommand = NULL;
	x->portrange = NULL;

	return x;
}

struct PROFILES_LIST_LIST *profiles_list_delete(struct PROFILES_LIST_LIST *x)
{
	struct PROFILES_LIST_LIST *start = x;

	while (start->prev != NULL)
		start = start->prev;

	if (x->next != NULL)
		x->next->prev = x->prev;
	if (x->prev != NULL)
		x->prev->next = x->next;
	else
		start = start->next;

	if (x->he != NULL)
		reg_free(x->he);
	if (x->fe != NULL)
		reg_free(x->fe);
	if (x->profiles != NULL)
		array_free(x->profiles);
	FREE_AND_NULL(x->comment);
	FREE_AND_NULL(x->host);
	FREE_AND_NULL(x->file);
	FREE_AND_NULL(x->portrange);
	if (x->addprofiles != NULL)
		array_free(x->addprofiles);
	if (x->removeprofiles != NULL)
		array_free(x->removeprofiles);
	if (x->urlcommand != NULL)
		array_free(x->urlcommand);

	xfree(x);

	return start;
}

/*
free memory used by PROFILES_LIST-type structure
*/
void profiles_free(PROFILES_LIST * profiles_list)
{
	if (!profiles_list)
		return;

	profiles_list_free(profiles_list->profiles);

	pthread_rwlock_destroy(&profiles_list->lock);

	xfree(profiles_list);
}

void profiles_list_free(struct PROFILES_LIST_LIST *pl)
{
	struct PROFILES_LIST_LIST *tmp;

	while (pl != NULL) {
		tmp = pl->next;

		if (pl->he != NULL)
			reg_free(pl->he);
		if (pl->fe != NULL)
			reg_free(pl->fe);
		if (pl->profiles != NULL)
			array_free(pl->profiles);
		FREE_AND_NULL(pl->comment);
		FREE_AND_NULL(pl->host);
		FREE_AND_NULL(pl->file);
		FREE_AND_NULL(pl->portrange);
		if (pl->addprofiles != NULL)
			array_free(pl->addprofiles);
		if (pl->removeprofiles != NULL)
			array_free(pl->removeprofiles);
		if (pl->urlcommand != NULL)
			array_free(pl->urlcommand);

		xfree(pl);
		pl = tmp;
	}
}

/*
remove areas of text matching patterns in PROFILES_LIST-type struct, and possibly
replace with something else
*/
void profiles_update(PROFILES_LIST * profiles_list, CONNECTION * connection)
{
	int ret;
	char **args;
	struct PROFILES_LIST_LIST *profiles;
	struct url_command_t **url_command;

	if (!profiles_list)
		return;

	pthread_rwlock_rdlock(&profiles_list->lock);

	for (profiles = profiles_list->profiles; profiles != NULL; profiles = profiles->next) {
		if (profiles->enabled == FALSE)
			continue;

		if (!profile_find(connection->profiles, profiles->profiles))
			continue;

		if (profiles->portrange != NULL)
			if (range_check(profiles->portrange, connection->header->port) == FALSE) continue;

		if (profiles->he != NULL && connection->header->host != NULL) {
			ret = reg_exec(profiles->he, connection->header->host);
			if (ret)
				continue;
		} else if (profiles->he != NULL)
			continue;

		if (profiles->fe != NULL) {
			ret = reg_exec(profiles->fe, connection->header->file);
			if (ret)
				continue;
		}

                if (profiles->urlcommand != NULL) {
                        if (connection->url_command == NULL)
                                continue;

                        url_command = connection->url_command;

                        for (; *url_command; url_command++)
                                if (array_find(profiles->urlcommand, (*url_command)->command))
                                        break;

                        if (*url_command == NULL)
                                continue;
                }

		for (args = profiles->addprofiles; args && *args; args++)
			connection->profiles = array_add(connection->profiles, *args);

		for (args = profiles->removeprofiles; args && *args; args++)
			connection->profiles = array_delete(connection->profiles, *args);
	}

	pthread_rwlock_unlock(&profiles_list->lock);
}


int profile_find(char **userp, char **entryp)
{
	int x, y, neg, found_negated = FALSE;
	char *ptr;

	/* If the entry does not have any profiles, it is valid. */
	if (!entryp)
		return TRUE;

	for (x = 0; entryp[x]; x++) {
		neg = (*entryp[x] == '!');
		ptr =  neg ? entryp[x]+1 : entryp[x];

		for (y = 0; userp && userp[y]; y++) {
			if (!strcmp(userp[y], ptr))
				return neg ? FALSE : TRUE;
		}

		/* remember if an unmatched negated entry profile is found */
		if (neg)
			found_negated = TRUE;
	}

	/* No active profile matched an entry profile. But if one of the entry's
	   profiles was negated, the entry is valid.
	*/
	return found_negated ? TRUE : FALSE;
}
