/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include "proto.h"

extern GLOBAL *global;

/*
load <rewrite> section from XML file into a linked list
*/
REWRITE_LIST *rewrite_load(REWRITE_LIST * rewrite_list, XML_LIST * xml_list)
{
	REWRITE_LIST *tmp_list = rewrite_list;
	struct REWRITE_LIST_LIST *rewrite = NULL;

	if (tmp_list == NULL) {
		tmp_list = xmalloc(sizeof(REWRITE_LIST));
		tmp_list->rewrite = NULL;
		tmp_list->id = 0;
		tmp_list->enabled = TRUE;

		rewrite_list = tmp_list;

		pthread_rwlock_init(&tmp_list->lock, NULL);
	} else
		rewrite = tmp_list->rewrite;

	while ((xml_list = xml_section(xml_list, "<rewrite>"))) {
		XML_LIST_LOOP(xml_list, "<rewrite>") {
			XML_LIST_CMP(xml_list, "<item>") {
				rewrite = rewrite_list_new(rewrite);
				rewrite->id = rewrite_list->id++;

				if (tmp_list->rewrite == NULL)
					tmp_list->rewrite = rewrite;
				XML_LIST_LOOP(xml_list, "<item>") {
					XML_LIST_CMP(xml_list, "<enabled>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE) {
							if (!strcasecmp(xml_list->item, "false"))
								rewrite->enabled = FALSE;
							else
								rewrite->enabled = TRUE;
						}
					}
					XML_LIST_CMP(xml_list, "<comment>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							rewrite_list_insert(rewrite, NULL, xml_list->item, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<profiles>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							rewrite_list_insert(rewrite, xml_list->item, NULL, NULL, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<pattern>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							rewrite_list_insert(rewrite, NULL, NULL, xml_list->item, NULL, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<replace>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							rewrite_list_insert(rewrite, NULL, NULL, NULL, xml_list->item, NULL, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<which>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							rewrite_list_insert(rewrite, NULL, NULL, NULL, NULL, xml_list->item, NULL, NULL);
					}
					XML_LIST_CMP(xml_list, "<mime>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							rewrite_list_insert(rewrite, NULL, NULL, NULL, NULL, NULL, xml_list->item, NULL);
					}
					XML_LIST_CMP(xml_list, "<size>") {
						xml_list = xml_list->next;
						if (xml_list->type == XML_VALUE)
							rewrite_list_insert(rewrite, NULL, NULL, NULL, NULL, NULL, NULL, xml_list->item);
					}
				}
			}
			XML_LIST_CMP(xml_list, "<enabled>") {
				xml_list = xml_list->next;
				if (xml_list->type == XML_VALUE) {
					if (!strcasecmp(xml_list->item, "false"))
						tmp_list->enabled = FALSE;
					else
						tmp_list->enabled = TRUE;
				}
			}
		}
	}

	return rewrite_list;
}

XML_LIST *rewrite_xml(REWRITE_LIST * rewrite_list, XML_LIST * xml_list)
{
	char *ptr, buf[128];
	struct REWRITE_LIST_LIST *rl;

	if (rewrite_list == NULL)
		return xml_list;

	pthread_rwlock_rdlock(&rewrite_list->lock);

	xml_list = xml_list_add(xml_list, "<rewrite>", XML_TAG);

	xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
	xml_list = xml_list_add(xml_list, (rewrite_list->enabled == TRUE) ? "true" : "false", XML_VALUE);
	xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

	for (rl = rewrite_list->rewrite; rl; rl = rl->next) {
		xml_list = xml_list_add(xml_list, "<item>", XML_TAG);

		xml_list = xml_list_add(xml_list, "<enabled>", XML_TAG);
		xml_list = xml_list_add(xml_list, (rl->enabled == TRUE) ? "true" : "false", XML_VALUE);
		xml_list = xml_list_add(xml_list, "</enabled>", XML_TAG);

		if (rl->comment != NULL) {
			xml_list = xml_list_add(xml_list, "<comment>", XML_TAG);
			ptr = string_to_xml(rl->comment);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</comment>", XML_TAG);
		}
		if (rl->profiles != NULL) {
			xml_list = xml_list_add(xml_list, "<profiles>", XML_TAG);
			ptr = array_merge(rl->profiles, ',');
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</profiles>", XML_TAG);
		}
		if (rl->mime != NULL) {
			xml_list = xml_list_add(xml_list, "<mime>", XML_TAG);
			ptr = string_to_xml(rl->mime);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</mime>", XML_TAG);
		}
		if (rl->pattern != NULL) {
			xml_list = xml_list_add(xml_list, "<pattern>", XML_TAG);
			ptr = string_to_xml(rl->pattern);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</pattern>", XML_TAG);
		}
		if (rl->replace != NULL) {
			xml_list = xml_list_add(xml_list, "<replace>", XML_TAG);
			ptr = string_to_xml(rl->replace);
			xml_list = xml_list_add(xml_list, ptr, XML_VALUE);
			xfree(ptr);
			xml_list = xml_list_add(xml_list, "</replace>", XML_TAG);
		}

		if (rl->size != 0) {
			xml_list = xml_list_add(xml_list, "<size>", XML_TAG);
                        snprintf(buf, sizeof(buf), "%d", rl->size);
			xml_list = xml_list_add(xml_list, buf, XML_VALUE);
			xml_list = xml_list_add(xml_list, "</size>", XML_TAG);
		}

		xml_list = xml_list_add(xml_list, "<which>", XML_TAG);
		snprintf(buf, sizeof(buf), "%s,%s,%s,%s", (rl->which & REWRITE_SERVER) ? "server" : "", (rl->which & REWRITE_CLIENT) ? "client" : "", (rl->which & REWRITE_BODY) ? "body" : "", (rl->which & REWRITE_POST) ? "post" : "");
		xml_list = xml_list_add(xml_list, buf, XML_VALUE);
		xml_list = xml_list_add(xml_list, "</which>", XML_TAG);

		xml_list = xml_list_add(xml_list, "</item>", XML_TAG);
	}

	xml_list = xml_list_add(xml_list, "</rewrite>", XML_TAG);

	pthread_rwlock_unlock(&rewrite_list->lock);

	return xml_list;
}
void rewrite_list_insert(struct REWRITE_LIST_LIST *x, char *profiles, char *comment, char *pattern, char *replace, char *which, char *mime, char *size)
{
	int i;
	char **args;

	if (profiles != NULL) {
		if (x->profiles != NULL)
			array_free(x->profiles);

		if (strcmp(profiles, ""))
			x->profiles = string_break(profiles, ',');
		else
			x->profiles = NULL;
	}
	if (comment != NULL) {
		FREE_AND_NULL(x->comment);

		if (strcmp(comment, ""))
			x->comment = xstrdup(comment);
	}

	if (pattern != NULL) {
		if (x->pe != NULL)
			reg_sub_free(x->pe);
		FREE_AND_NULL(x->pattern);

		if (strcmp(pattern, "")) {
			x->pe = reg_sub_compile(pattern, PCREFLAGS);
			x->pattern = xstrdup(pattern);
		} else
			x->pe = NULL;
	}

	if (replace != NULL) {
		FREE_AND_NULL(x->replace);

		if (strcmp(replace, ""))
			x->replace = xstrdup(replace);
	}
	if (which != NULL) {
		x->which = 0;

		if (strcmp(which, "")) {
			args = string_break(which, ',');
			for (i = 0; args[i]; i++) {
				if (!strcasecmp(args[i], "server"))
					x->which |= REWRITE_SERVER;
				else if (!strcasecmp(args[i], "client"))
					x->which |= REWRITE_CLIENT;
				else if (!strcasecmp(args[i], "body"))
					x->which |= REWRITE_BODY;
				else if (!strcasecmp(args[i], "post"))
					x->which |= REWRITE_POST;

				xfree(args[i]);
			}

			xfree(args);
		}
	}
	if (mime != NULL) {
		if (x->me != NULL)
			reg_free(x->me);
		FREE_AND_NULL(x->mime);

		if (strcmp(mime, "")) {
			x->me = reg_compile(mime, REGFLAGS);
			x->mime = xstrdup(mime);
		} else
			x->me = NULL;
	}
	if (size != NULL) {
		x->size = 0;

		if (strcmp(size, ""))
	                x->size = string_to_size(size);
	}
}

struct REWRITE_LIST_LIST *rewrite_list_new(struct REWRITE_LIST_LIST *x)
{
	if (x == NULL) {
		x = xmalloc(sizeof(struct REWRITE_LIST_LIST));
		x->prev = NULL;
	} else {
		while (x->next != NULL)
			x = x->next;
		x->next = xmalloc(sizeof(struct REWRITE_LIST_LIST));
		x->next->prev = x;
		x = x->next;
	}
	x->next = NULL;
	x->enabled = TRUE;
	x->comment = NULL;
	x->profiles = NULL;
	x->pe = NULL;
	x->me = NULL;
	x->mime = NULL;
	x->pattern = NULL;
	x->replace = NULL;
	x->which = REWRITE_BODY;
	x->size = 0;

	return x;
}

struct REWRITE_LIST_LIST *rewrite_list_delete(struct REWRITE_LIST_LIST *x)
{
	struct REWRITE_LIST_LIST *start = x;

	while (start->prev != NULL)
		start = start->prev;

	if (x->next != NULL)
		x->next->prev = x->prev;
	if (x->prev != NULL)
		x->prev->next = x->next;
	else
		start = start->next;

	if (x->pe != NULL)
		reg_sub_free(x->pe);
	if (x->me != NULL)
		reg_free(x->me);
	if (x->profiles != NULL)
		array_free(x->profiles);
	FREE_AND_NULL(x->comment);
	FREE_AND_NULL(x->pattern);
	FREE_AND_NULL(x->replace);
	FREE_AND_NULL(x->mime);

	xfree(x);

	return start;
}

/*
free memory used by REWRITE_LIST-type structure
*/
void rewrite_free(REWRITE_LIST * rewrite_list)
{
	if (!rewrite_list)
		return;

	rewrite_list_free(rewrite_list->rewrite);

	pthread_rwlock_destroy(&rewrite_list->lock);

	xfree(rewrite_list);
}

void rewrite_list_free(struct REWRITE_LIST_LIST *rl)
{
	struct REWRITE_LIST_LIST *tmp;

	while (rl != NULL) {
		tmp = rl->next;

		if (rl->pe != NULL)
			reg_sub_free(rl->pe);
		if (rl->me != NULL)
			reg_free(rl->me);
		if (rl->profiles != NULL)
			array_free(rl->profiles);
		FREE_AND_NULL(rl->comment);
		FREE_AND_NULL(rl->pattern);
		FREE_AND_NULL(rl->replace);
		FREE_AND_NULL(rl->mime);

		xfree(rl);
		rl = tmp;
	}
}

/*
remove areas of text matching patterns in REWRITE_LIST-type struct, and possibly
replace with something else
*/
int rewrite_do(REWRITE_LIST * rewrite_list, CONNECTION * connection, FILEBUF * filebuf, int flags, int action)
{
	int x = FALSE, ret, maxbuffer;
	struct REWRITE_LIST_LIST *rewrite;

	if (!rewrite_list || connection->bypass & FEATURE_REWRITE)
		return FALSE;

	pthread_rwlock_rdlock(&rewrite_list->lock);

	if (rewrite_list->enabled == FALSE) {
		pthread_rwlock_unlock(&rewrite_list->lock);

		return FALSE;
	}

        maxbuffer = atomic_read_rwlock(&global->general->lock, &global->general->maxbuffer);

	for (rewrite = rewrite_list->rewrite; rewrite != NULL; rewrite = rewrite->next) {
		if (rewrite->enabled == FALSE)
			continue;

		if (!(rewrite->which & flags))
			continue;

		if (!profile_find(connection->profiles, rewrite->profiles))
			continue;

		if (connection->header != NULL) {
			if (flags == REWRITE_BODY && rewrite->me != NULL && connection->rheader->content_type != NULL) {
				ret = reg_exec(rewrite->me, connection->rheader->content_type);
				if (ret)
					continue;
			} else if (rewrite->me != NULL) {
				ret = reg_exec(rewrite->me, "");
				if (ret)
					continue;
			}

                        if (rewrite->size != 0) {
                            ret = (rewrite->size == -1)? maxbuffer : rewrite->size;
                                
                            if (action == FALSE) {
                                if (connection->rheader->content_length > ret)
                                    continue;
                            } else {
                                if (connection->rheader->content_length == -1 || connection->rheader->content_length > ret)
                                    continue;
                            }
                        }
		}

		if (rewrite->pe != NULL) {
			x = TRUE;

			if (action == FALSE)
				break;

			reg_replace(filebuf, rewrite->pe, rewrite->replace);
		}
	}

	pthread_rwlock_unlock(&rewrite_list->lock);

	return x;
}
