/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include "proto.h"

extern char configfile[];

extern THREADLIST threads[];
extern PROFILES_LIST *profiles_list;
extern FILTER_LIST *filter_list;
extern HEADER_LIST *header_list;
extern MIME_LIST *mime_list;
extern REDIRECT_LIST *redirect_list;
extern KEYWORD_LIST *keyword_list;
extern ACCESS_LIST *access_list;
extern COOKIE_LIST *cookie_list;
extern REWRITE_LIST *rewrite_list;
extern FORWARD_LIST *forward_list;
extern LIMITS_LIST *limits_list;
extern TEMPLATES *templates;
extern EXTERNAL *external;
extern NETWORK *network;
extern GLOBAL *global;

struct pages_t {
	char *file;
	void (*handler) (FILEBUF *, struct cgi_args_t *, CONNECTION *);
};

struct calendarmap_t {
	char *name;
	char *html;
};

struct calendarmap_t calendar_months[] = {
	{"January", "<option value=\"January\" %s>January\n"},
	{"February", "<option value=\"February\" %s>February\n"},
	{"March", "<option value=\"March\" %s>March\n"},
	{"April", "<option value=\"April\" %s>April\n"},
	{"May", "<option value=\"May\" %s>May\n"},
	{"June", "<option value=\"June\" %s>June\n"},
	{"July", "<option value=\"July\" %s>July\n"},
	{"August", "<option value=\"August\" %s>August\n"},
	{"September", "<option value=\"September\" %s>September\n"},
	{"October", "<option value=\"October\" %s>October\n"},
	{"November", "<option value=\"November\" %s>November\n"},
	{"December", "<option value=\"December\" %s>December\n"},
	{NULL, NULL}
};

struct calendarmap_t calendar_weekdays[] = {
	{"Sunday", "<option value=\"Sunday\" %s>Sunday\n"},
	{"Monday", "<option value=\"Monday\" %s>Monday\n"},
	{"Tuesday", "<option value=\"Tuesday\" %s>Tuesday\n"},
	{"Wednesday", "<option value=\"Wednesday\" %s>Wednesday\n"},
	{"Thursday", "<option value=\"Thursday\" %s>Thursday\n"},
	{"Friday", "<option value=\"Friday\" %s>Friday\n"},
	{"Saturday", "<option value=\"Saturday\" %s>Saturday\n"},
	{NULL, NULL}
};

struct pages_t pages[] = {
	{"/connections", interface_page_connections},
	{"/dnscache", interface_page_dnscache},
	{"/headers", interface_page_headers},
	{"/config", interface_page_config},
	{"/save", interface_page_save},
	{"/load", interface_page_load},
	{"/log", interface_page_log},
	{"/cache", interface_page_cache},
	{"/pool", interface_page_pool},
	{"/prefetch", interface_page_prefetch},
	{NULL, NULL}
};

struct pages_t config_pages[] = {
	{"global", interface_page_global},
	{"profiles", interface_page_config_profiles},
	{"cache", interface_page_config_cache},
	{"filter", interface_page_config_filter},
	{"access", interface_page_config_access},
	{"rewrite", interface_page_config_rewrite},
	{"mime", interface_page_config_mime},
	{"cookies", interface_page_config_cookies},
	{"external", interface_page_config_external},
	{"templates", interface_page_config_templates},
	{"network", interface_page_config_network},
	{"header", interface_page_config_header},
	{"redirect", interface_page_config_redirect},
	{"forward", interface_page_config_forward},
	{"keywords", interface_page_config_keywords},
	{"limits", interface_page_config_limits},
	{NULL, NULL}
};

struct pages_t action_pages[] = {
	{"profiles", interface_page_config_profiles_action},
	{"cache", interface_page_config_cache_action},
	{"filter", interface_page_config_filter_action},
	{"external", interface_page_config_external_action},
	{"rewrite", interface_page_config_rewrite_action},
	{"access", interface_page_config_access_action},
	{"mime", interface_page_config_mime_action},
	{"templates", interface_page_config_templates_action},
	{"network", interface_page_config_network_action},
	{"header", interface_page_config_header_action},
	{"cookies", interface_page_config_cookies_action},
	{"redirect", interface_page_config_redirect_action},
	{"forward", interface_page_config_forward_action},
	{"keywords", interface_page_config_keywords_action},
	{"limits", interface_page_config_limits_action},
	{NULL, NULL}
};

struct pages_t dialog_pages[] = {
	{"profiles", interface_page_config_profiles_dialog},
	{"cache", interface_page_config_cache_dialog},
	{"filter", interface_page_config_filter_dialog},
	{"external", interface_page_config_external_dialog},
	{"rewrite", interface_page_config_rewrite_dialog},
	{"access", interface_page_config_access_dialog},
	{"mime", interface_page_config_mime_dialog},
	{"templates", interface_page_config_templates_dialog},
	{"network", interface_page_config_network_dialog},
	{"header", interface_page_config_header_dialog},
	{"cookies", interface_page_config_cookies_dialog},
	{"redirect", interface_page_config_redirect_dialog},
	{"forward", interface_page_config_forward_dialog},
	{"keywords", interface_page_config_keywords_dialog},
	{"limits", interface_page_config_limits_dialog},
	{NULL, NULL}
};

struct pages_t global_pages[] = {
	{"General", interface_page_global_general},
	{"FTP", interface_page_global_ftp},
	{"DNSBL", interface_page_global_dnsbl},
	{"Prefetch", interface_page_global_prefetch},
	{NULL, NULL}
};

struct pages_t global_action_pages[] = {
	{"General", interface_page_global_general_action},
	{"FTP", interface_page_global_ftp_action},
	{"DNSBL", interface_page_global_dnsbl_action},
	{"Prefetch", interface_page_global_prefetch_action},
	{NULL, NULL}
};

struct cgi_args_t *cgi_parse_request(char *file)
{
	int pos = 0, astart = 0, aequal = -1;
	struct cgi_args_t *args = NULL, *start = NULL;

	putlog(MMLOG_DEBUG, "cgi arguments: %s", file);

	for (; file[pos]; pos++) {
		if (file[pos] == '=')
			aequal = pos;
		else if (file[pos] == '&' && aequal != -1) {
			if (args == NULL) {
				args = xmalloc(sizeof(struct cgi_args_t));
				start = args;
			} else {
				args->next = xmalloc(sizeof(struct cgi_args_t));
				args = args->next;
			}

			args->name = url_decode(&file[astart], aequal - astart);
			args->value = url_decode(&file[aequal + 1], pos - aequal - 1);

			args->next = NULL;

			aequal = -1;
			astart = pos + 1;
		}
	}

	if (aequal != -1) {
		if (args == NULL) {
			args = xmalloc(sizeof(struct cgi_args_t));
			start = args;
		} else {
			args->next = xmalloc(sizeof(struct cgi_args_t));
			args = args->next;
		}

		args->next = NULL;

		args->name = url_decode(&file[astart], aequal - astart);
		args->value = url_decode(&file[aequal + 1], pos - aequal - 1);
	}

	return start;
}

void cgi_args_free(struct cgi_args_t *args)
{
	struct cgi_args_t *tmp;

	while (args != NULL) {
		tmp = args->next;

		xfree(args->name);
		xfree(args->value);

		xfree(args);
		args = tmp;
	}
}

int interface_check_referer(CONNECTION * connection)
{
	char buf[128];

	if (connection->header->referer != NULL) {
		snprintf(buf, sizeof(buf), "http://%s", INTERFACEURL);
		if (!strncasecmp(connection->header->referer, buf, strlen(buf))) {
			if (connection->header->referer[strlen(buf)] == ':' || connection->header->referer[strlen(buf)] == '/')
				return TRUE;
		}
	}

	return FALSE;
}

void interface_handle_request(CONNECTION * connection)
{
	int i;
	char *ptr = NULL, *file;
	HEADER *header;
	FILEBUF *filebuf;
	struct cgi_args_t *args = NULL;

	if (!strncasecmp(&connection->header->file[(connection->header->type == HTTP_REQUEST) ? strlen(INTERFACEURL) + 1 : 0], "/template/", 9)) {
		template_send(templates, &connection->header->file[(connection->header->type == HTTP_REQUEST) ? strlen(INTERFACEURL) + 11 : 10], connection, 200);

		return;
	}

	if (!(connection->access & ACCESS_CONFIG)) {
		template_send(templates, "noaccess", connection, 403);

		return;
	}

	header = header_new();

	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");

	if (connection->header->type == HTTP_REQUEST) {
		/* we wouldn't be here if connection->header->file wasn't this big */
		file = connection->header->file + strlen(INTERFACEURL) + 1;

		if (*file != '/') {
			/* send redirect to correct relative links */

			header->code = 302;
			header->content_length = 0;
			header->location = xmalloc(strlen(INTERFACEURL) + 2);
			sprintf(header->location, "%s/", INTERFACEURL);

			header_send(header, connection, CLIENT, HEADER_RESP);

			http_header_free(header);

			return;
		}
	} else
		file = connection->header->file;


	if (interface_check_referer(connection) == TRUE) {
		/* only parse cgi options if Referer is valid, this prevents
		   malicious websites from changing proxy settings by having inline
		   images from carefully crafted URL's */
		if (strcasecmp(connection->header->method, "POST")) {
			/* retrieve cgi arguments from GET request */
			ptr = strchr(file, '?');
			if (ptr != NULL)
				args = cgi_parse_request(&ptr[1]);
		} else if (connection->header->content_length != -1) {
			/* retrieve form data from POST request */
			filebuf = http_transfer_filebuf(connection, CLIENT, -1);
			filebuf_add(filebuf, "", 1);

			args = cgi_parse_request(filebuf->data);

			filebuf_free(filebuf);
		}
	}

	filebuf = filebuf_new();

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Middleman %s web interface</title></head>\n", VERSION);
	filebuf_addf(filebuf, "<body bgcolor=\"%s\" text=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_BG, INTERFACE_TEXT, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\"><tr><td align=\"center\"><font size=\"6\">Middleman Web interface</font></td></tr></table>\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "</td></tr><tr><td><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\"><tr align=\"center\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<td><a href=\"connections\">Active connections</a></td>\n");
	filebuf_addf(filebuf, "<td><a href=\"config\">Config</a></td>\n");
	filebuf_addf(filebuf, "<td><a href=\"dnscache\">DNS cache</a></td>\n");
	filebuf_addf(filebuf, "<td><a href=\"headers\">Show headers</a></td></tr>\n");
	filebuf_addf(filebuf, "<tr align=\"center\">\n");
	filebuf_addf(filebuf, "<td><a href=\"log\">View log entries</a></td>");
	filebuf_addf(filebuf, "<td><a href=\"cache\">View cache entries</a></td>");
	filebuf_addf(filebuf, "<td><a href=\"save\">Save settings</a></td>\n");
	filebuf_addf(filebuf, "<td><a href=\"load\">Load settings</a></td>\n");
	filebuf_addf(filebuf, "</tr>\n");
	filebuf_addf(filebuf, "<tr align=\"center\">\n");
	filebuf_addf(filebuf, "<td colspan=\"2\"><a href=\"pool\">Connection pool</a></td>\n");
	filebuf_addf(filebuf, "<td colspan=\"2\"><a href=\"prefetch\">Prefetch queue</a></td>\n");
	filebuf_addf(filebuf, "</tr>\n");

	filebuf_addf(filebuf, "</table></td></tr>\n<tr><td><br></td></tr><tr><td>\n");

	for (i = 0; pages[i].file != NULL; i++) {
		if (!strncasecmp(file, pages[i].file, (ptr != NULL) ? ptr - file : ~0))
			pages[i].handler(filebuf, args, connection);
	}

	filebuf_addf(filebuf, "</td></tr></table></body></html>\n");

	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);

	if (strcasecmp(connection->header->method, "HEAD"))
		net_filebuf_send(filebuf, connection, CLIENT);

	if (args != NULL)
		cgi_args_free(args);
	http_header_free(header);
	filebuf_free(filebuf);
}

void interface_page_connections(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;

	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	filebuf_addf(filebuf, "<tr><td>PID</td><td>IP</td><td>Requests</td><td>Method</td><td>Host</td><td>Port</td><td>File</td></tr>\n");

	for (i = 0; i < MAXTHREADS; i++) {
		pthread_mutex_lock(&threads[i].lock);
		if (threads[i].flags == THREAD_BUSY)
			filebuf_addf(filebuf, "<tr><td>%d</td><td>%s</td><td>%d</td><td>%s</td><td>%s</td><td>%d</td><td>%s</td></tr>\n", threads[i].pid, threads[i].ip, threads[i].requests, (threads[i].method != NULL) ? threads[i].method : "NONE", (threads[i].host != NULL) ? threads[i].host : "NONE", threads[i].port, (threads[i].file != NULL) ? threads[i].file : "NONE");

		pthread_mutex_unlock(&threads[i].lock);
	}

	filebuf_addf(filebuf, "</table>\n");
}

void interface_page_dnscache(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;
	time_t curtime = time(NULL);
	struct HASH_LIST *hash_list;

	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	filebuf_addf(filebuf, "<tr><td>Hostname</td><td>IP address</td><td>Age</td></tr>\n");

	pthread_mutex_lock(&global->dns_cache_lock);
	for (i = 0; i < global->dns_cache->size; i++) {
		hash_list = global->dns_cache->hash_list[i];

		for (; hash_list; hash_list = hash_list->next)
			filebuf_addf(filebuf, "<tr><td>%s</td><td>%s</td><td>%d</td></tr>\n", hash_list->ref, (char *) hash_list->data, curtime - hash_list->age);
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_mutex_unlock(&global->dns_cache_lock);
}

void interface_page_headers(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	struct HTTP_HEADER_LIST *header_list;


	filebuf_addf(filebuf, "<tr><td><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><th colspan=\"2\">Unfiltered</th></tr>\n");
	filebuf_addf(filebuf, "<tr><td>Type</td><td>Value</td></tr>\n");

	for (header_list = connection->header->header; header_list; header_list = header_list->next)
		filebuf_addf(filebuf, "<tr><td>%s</td><td>%s</td></tr>\n", header_list->type, header_list->value);

	filebuf_addf(filebuf, "</table></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td><br></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><th colspan=\"2\">Filtered</th></tr>\n");
	filebuf_addf(filebuf, "<tr><td>Type</td><td>Value</td></tr>\n");

	for (header_list = connection->header->header_filtered; header_list; header_list = header_list->next)
		filebuf_addf(filebuf, "<tr><td>%s</td><td>%s</td></tr>\n", header_list->type, header_list->value);

	filebuf_addf(filebuf, "</table></td></tr>\n");
}

void interface_page_save(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int ret;
	char *filename = NULL;

	for (; args; args = args->next)
		if (!strcasecmp(args->name, "filename"))
			filename = args->value;

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");

	if (filename == NULL) {
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<form action=\"save\" method=\"POST\">\n");
		filebuf_addf(filebuf, "<tr><td>Filename</td><td><input type=\"text\" name=\"filename\" value=\"%s\" size=\"80\"></td></tr>\n", (*configfile) ? configfile : "");
		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
		filebuf_addf(filebuf, "</form></table>\n");

		filebuf_addf(filebuf, "</td></tr></table>\n");

		return;
	}

	ret = config_save(filename);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
	if (ret == TRUE)
		filebuf_addf(filebuf, "File saved\n");
	else
		filebuf_addf(filebuf, "Saved failed\n");

	filebuf_addf(filebuf, "</td></tr></table></td></tr></table>\n");
}

void interface_page_load(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int ret, overwrite = FALSE;
	char *filename = NULL;

	for (; args; args = args->next) {
		if (!strcasecmp(args->name, "filename"))
			filename = args->value;
		else if (!strcasecmp(args->name, "overwrite")) {
			if (!strcasecmp(args->value, "yes"))
				overwrite = TRUE;
			else
				overwrite = FALSE;
		}
	}

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");

	if (filename == NULL) {
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<form action=\"load\" method=\"POST\">\n");
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Filename</td><td><input type=\"text\" name=\"filename\" value=\"%s\" size=\"80\"></td></tr>\n", (*configfile) ? configfile : "");
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Overwrite</td><td align=\"center\">Yes: <input type=\"radio\" name=\"overwrite\" value=\"yes\" checked> No: <input type=\"radio\" name=\"overwrite\" value=\"no\"></td></tr>\n");
		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
		filebuf_addf(filebuf, "</form></table>\n");

		filebuf_addf(filebuf, "</td></tr></table>\n");

		return;
	}

	ret = config_load(overwrite, filename);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td align=\"center\">\n");

	if (ret == TRUE)
		filebuf_addf(filebuf, "File loaded\n");
	else
		filebuf_addf(filebuf, "Load failed\n");

	filebuf_addf(filebuf, "</td></tr></table></td></tr></table>\n");
}

void interface_page_log(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int size = -1, clear = FALSE;
	char *ptr, *pattern = NULL;
	struct LOGBUFFER_LIST *ll;
	struct cgi_args_t *a;
	regex_t *pe = NULL;

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "pattern"))
			pattern = a->value;
		if (!strcasecmp(a->name, "clear"))
			clear = TRUE;
		if (!strcasecmp(a->name, "size"))
			size = atoi(a->value);
	}

	if (clear == TRUE)
		logbuffer_clear(global->logbuffer);
	if (size != -1)
		logbuffer_resize(global->logbuffer, size);

	pthread_rwlock_rdlock(&global->logbuffer->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"log\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\">Regular expression match</td></tr>\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"text\" name=\"pattern\" value=\"%s\" size=\"80\"></td></tr>\n", (pattern != NULL) ? pattern : "");
	filebuf_addf(filebuf, "<tr><td align=\"center\">Log buffer size: <input type=\"text\" name=\"size\" value=\"%d\" size=\"10\"></td>\n", global->logbuffer->size);
	filebuf_addf(filebuf, "<td align=\"center\">Clear: <input type=\"checkbox\" name=\"clear\"></td></tr>");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr><tr><td><br></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	/* unlock to prevent deadlock incase the regexp pattern can't be compiled and it generates a log message */
	pthread_rwlock_unlock(&global->logbuffer->lock);

	if (pattern != NULL && strcmp(pattern, ""))
		pe = reg_compile(pattern, REGFLAGS);

	pthread_rwlock_rdlock(&global->logbuffer->lock);

	for (ll = global->logbuffer->head; ll; ll = ll->next) {
		if (pe != NULL && reg_exec(pe, ll->msg))
			continue;

		ptr = string_to_html(ll->msg, HTML_NEWLINES);
		filebuf_addf(filebuf, "<tr><td>%s</td></tr>\n", ptr);
		xfree(ptr);
	}

	if (pe != NULL)
		reg_free(pe);

	filebuf_addf(filebuf, "</table></td></tr></table>\n");

	pthread_rwlock_unlock(&global->logbuffer->lock);
}

void interface_page_prefetch(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	struct PREFETCH_QUEUE *queue;
	char *ptr, *qadd = NULL;
	URL *url;
	struct cgi_args_t *a;

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "queue"))
			qadd = a->value;
	}

	if (qadd != NULL) {
		url = url_parse(qadd);
		if (url != NULL)
			prefetch_queue_add(global->prefetch, url, 1);
	}

	pthread_mutex_lock(&global->prefetch->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"prefetch\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\">Add URL to prefetch queue</td></tr>\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"text\" name=\"queue\" size=\"80\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr><tr><td><br></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	for (queue = global->prefetch->queue; queue; queue = queue->next) {
		ptr = url_create(queue->url);
		filebuf_addf(filebuf, "<tr><td>%s</td></tr>\n", ptr);
		xfree(ptr);
	}

	filebuf_addf(filebuf, "</table></td></tr></table>\n");

	pthread_mutex_unlock(&global->prefetch->lock);
}

void interface_page_cache(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i, delete = FALSE, show = FALSE;
	char *ptr, *pattern = NULL, *ptr2, proto[16];
	size_t minsize = 0, maxsize = ~0;
	int allsizes = FALSE;
	int mbday = -1, maday = -1;
	int mbyear = -1, mayear = -1;
	int mbhour = -1, mahour = -1;
	int mbminute = -1, maminute = -1;
	int abday = -1, aaday = -1;
	int abyear = -1, aayear = -1;
	int abhour = -1, aahour = -1;
	int abminute = -1, aaminute = -1;
	int mbcurtime = FALSE, abcurtime = FALSE;
	int maepoch = FALSE, aaepoch = FALSE;
	char *mbmonth = NULL, *mamonth = NULL;
	char *abmonth = NULL, *aamonth = NULL;
	regex_t *pe = NULL;
	struct cgi_args_t *a;
	CACHEMAP *cachemap;
	struct CACHEMAPLIST_T *cl, *tmp;
	struct cache_store *cstore;
	time_t curtime, mbdate, madate, abdate, aadate;
	struct tm lt;

	curtime = time(NULL);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "delete"))
			delete = TRUE;
		else if (!strcasecmp(a->name, "show"))
			show = TRUE;
		else if (!strcasecmp(a->name, "pattern"))
			pattern = a->value;
		else if (!strcasecmp(a->name, "mbday"))
			mbday = atoi(a->value);
		else if (!strcasecmp(a->name, "maday"))
			maday = atoi(a->value);
		else if (!strcasecmp(a->name, "mbyear"))
			mbyear = atoi(a->value);
		else if (!strcasecmp(a->name, "mayear"))
			mayear = atoi(a->value);
		else if (!strcasecmp(a->name, "mbhour"))
			mbhour = atoi(a->value);
		else if (!strcasecmp(a->name, "mahour"))
			mahour = atoi(a->value);
		else if (!strcasecmp(a->name, "mbminute"))
			mbminute = atoi(a->value);
		else if (!strcasecmp(a->name, "maminute"))
			maminute = atoi(a->value);
		else if (!strcasecmp(a->name, "abday"))
			abday = atoi(a->value);
		else if (!strcasecmp(a->name, "aaday"))
			aaday = atoi(a->value);
		else if (!strcasecmp(a->name, "abyear"))
			abyear = atoi(a->value);
		else if (!strcasecmp(a->name, "aayear"))
			aayear = atoi(a->value);
		else if (!strcasecmp(a->name, "abhour"))
			abhour = atoi(a->value);
		else if (!strcasecmp(a->name, "aahour"))
			aahour = atoi(a->value);
		else if (!strcasecmp(a->name, "abminute"))
			abminute = atoi(a->value);
		else if (!strcasecmp(a->name, "aaminute"))
			aaminute = atoi(a->value);
		else if (!strcasecmp(a->name, "mbmonth"))
			mbmonth = a->value;
		else if (!strcasecmp(a->name, "mamonth"))
			mamonth = a->value;
		else if (!strcasecmp(a->name, "abmonth"))
			abmonth = a->value;
		else if (!strcasecmp(a->name, "aamonth"))
			aamonth = a->value;
		else if (!strcasecmp(a->name, "mbcurtime"))
			mbcurtime = TRUE;
		else if (!strcasecmp(a->name, "maepoch"))
			maepoch = TRUE;
		else if (!strcasecmp(a->name, "abcurtime"))
			abcurtime = TRUE;
		else if (!strcasecmp(a->name, "aaepoch"))
			aaepoch = TRUE;
		else if (!strcasecmp(a->name, "minsize"))
			minsize = string_to_size(a->value);
		else if (!strcasecmp(a->name, "maxsize"))
			maxsize = string_to_size(a->value);
		else if (!strcasecmp(a->name, "allsizes"))
			allsizes = TRUE;
	}

	localtime_r(&curtime, &lt);

	if (allsizes == TRUE) {
		minsize = 0;
		maxsize = ~0;
	}

	if (mbday == -1 || mbcurtime == TRUE)
		mbday = lt.tm_mday;
	if (maday == -1 || maepoch == TRUE)
		maday = 1;
	if (abday == -1 || abcurtime == TRUE)
		abday = lt.tm_mday;
	if (aaday == -1 || aaepoch == TRUE)
		aaday = 1;

	if (mbyear == -1 || mbcurtime == TRUE)
		mbyear = lt.tm_year + 1900;
	if (mayear == -1 || maepoch == TRUE)
		mayear = 1970;
	if (abyear == -1 || abcurtime == TRUE)
		abyear = lt.tm_year + 1900;
	if (aayear == -1 || aaepoch == TRUE)
		aayear = 1970;

	if (mbminute == -1 || mbcurtime == TRUE)
		mbminute = lt.tm_min;
	if (maminute == -1 || maepoch == TRUE)
		maminute = 0;
	if (abminute == -1 || abcurtime == TRUE)
		abminute = lt.tm_min;
	if (aaminute == -1 || aaepoch == TRUE)
		aaminute = 0;

	if (mbhour == -1 || mbcurtime == TRUE)
		mbhour = lt.tm_hour;
	if (mahour == -1 || maepoch == TRUE)
		mahour = 0;
	if (abhour == -1 || abcurtime == TRUE)
		abhour = lt.tm_hour;
	if (aahour == -1 || aaepoch == TRUE)
		aahour = 0;

	/* hope it's safe to assume lt.tm_mon <= 12 ... */
	if (mbmonth == NULL || mbcurtime == TRUE)
		mbmonth = calendar_months[lt.tm_mon].name;
	if (mamonth == NULL || maepoch == TRUE)
		mamonth = calendar_months[0].name;
	if (abmonth == NULL || mbcurtime == TRUE)
		abmonth = calendar_months[lt.tm_mon].name;
	if (aamonth == NULL || aaepoch == TRUE)
		aamonth = calendar_months[0].name;

	memset(&lt, 0, sizeof(lt));

	lt.tm_year = mbyear - 1900;
	lt.tm_mday = mbday;
	for (i = 0; calendar_months[i].name && strcasecmp(calendar_months[i].name, mbmonth); i++);
	lt.tm_mon = i;
	lt.tm_hour = mbhour;
	lt.tm_min = mbminute;
	mbdate = mktime(&lt);

	lt.tm_year = mayear - 1900;
	lt.tm_mday = maday;
	for (i = 0; calendar_months[i].name && strcasecmp(calendar_months[i].name, mamonth); i++);
	lt.tm_mon = i;
	lt.tm_hour = mahour;
	lt.tm_min = maminute;
	madate = mktime(&lt);

	lt.tm_year = abyear - 1900;
	lt.tm_mday = abday;
	for (i = 0; calendar_months[i].name && strcasecmp(calendar_months[i].name, abmonth); i++);
	lt.tm_mon = i;
	lt.tm_hour = abhour;
	lt.tm_min = abminute;
	abdate = mktime(&lt);

	lt.tm_year = aayear - 1900;
	lt.tm_mday = aaday;
	for (i = 0; calendar_months[i].name && strcasecmp(calendar_months[i].name, aamonth); i++);
	lt.tm_mon = i;
	lt.tm_hour = aahour;
	lt.tm_min = aaminute;
	aadate = mktime(&lt);

	pthread_mutex_lock(&global->cache->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><th align=\"center\" colspan=\"2\">Cache information</th></tr>\n");
	filebuf_addf(filebuf, "<tr><td>Statistics</td><td>%d hits / %d misses / %d%% hit ratio</td></tr>\n", global->cache->hit, global->cache->miss, (global->cache->hit == 0) ? 0 : (int) (((float) global->cache->hit / (float)(global->cache->miss + global->cache->hit)) *100.0));
	filebuf_addf(filebuf, "<tr><td>Memory cache</td><td>%d entries / %u bytes / %u%% used</td></tr>\n", global->cache->mementries, global->cache->memsize, (global->cache->memsize == 0) ? 0 : (int) (((float) global->cache->memsize / (float)global->cache->maxmemsize) *100.0));
	for (cstore = global->cache->stores; cstore; cstore = cstore->next) {
		if (cstore->path == NULL)
			continue;
		filebuf_addf(filebuf, "<tr><td>Disk cache %s</td><td>%d entries / %u bytes / %u%% used</td></tr>\n", cstore->path, cstore->diskentries, cstore->disksize, (cstore->disksize == 0) ? 0 : (int) (((float) cstore->disksize / (float) cstore->maxdisksize) * 100.0));
	}
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr> <tr><td align=\"center\">\n");

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"cache\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"show\" value=\"yes\"\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\">Regular expression match</td></tr>\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"text\" name=\"pattern\" value=\"%s\" size=\"80\"></td></tr>\n", (pattern != NULL) ? pattern : "");

	filebuf_addf(filebuf, "<tr><td>Modified after</td>");
	filebuf_addf(filebuf, "<td>Date: <select name=\"mamonth\">\n");
	for (i = 0; calendar_months[i].name; i++)
		filebuf_addf(filebuf, calendar_months[i].html, (!strcasecmp(mamonth, calendar_months[i].name)) ? "selected" : "");
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<select name=\"maday\">\n");
	for (i = 1; i <= 31; i++)
		filebuf_addf(filebuf, "<option value=\"%d\" %s>%d\n", i, (i == maday) ? "selected" : "", i);
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<input type=\"text\" name=\"mayear\" value=\"%d\" size=\"4\">\n", mayear);
	filebuf_addf(filebuf, "Time: <input type=\"text\" name=\"mahour\" value=\"%.2d\" size=\"2\">:\n", mahour);
	filebuf_addf(filebuf, "<input type=\"text\" name=\"maminute\" value=\"%.2d\" size=\"2\">\n", maminute);
	filebuf_addf(filebuf, " Epoch time: <input type=\"checkbox\" name=\"maepoch\" %s>", (maepoch == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "</tr></td>\n");

	filebuf_addf(filebuf, "<tr><td>Modified before</td>");
	filebuf_addf(filebuf, "<td>Date: <select name=\"mbmonth\">\n");
	for (i = 0; calendar_months[i].name; i++)
		filebuf_addf(filebuf, calendar_months[i].html, (!strcasecmp(mbmonth, calendar_months[i].name)) ? "selected" : "");
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<select name=\"mbday\">\n");
	for (i = 1; i <= 31; i++)
		filebuf_addf(filebuf, "<option value=\"%d\" %s>%d\n", i, (i == mbday) ? "selected" : "", i);
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<input type=\"text\" name=\"mbyear\" value=\"%d\" size=\"4\">\n", mbyear);
	filebuf_addf(filebuf, "Time: <input type=\"text\" name=\"mbhour\" value=\"%.2d\" size=\"2\">:\n", mbhour);
	filebuf_addf(filebuf, "<input type=\"text\" name=\"mbminute\" value=\"%.2d\" size=\"2\">\n", mbminute);
	filebuf_addf(filebuf, " Current time: <input type=\"checkbox\" name=\"mbcurtime\" %s>", (mbcurtime == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "</tr></td>\n");

	filebuf_addf(filebuf, "<tr><td>Accessed after</td>");
	filebuf_addf(filebuf, "<td>Date: <select name=\"aamonth\">\n");
	for (i = 0; calendar_months[i].name; i++)
		filebuf_addf(filebuf, calendar_months[i].html, (!strcasecmp(aamonth, calendar_months[i].name)) ? "selected" : "");
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<select name=\"aaday\">\n");
	for (i = 1; i <= 31; i++)
		filebuf_addf(filebuf, "<option value=\"%d\" %s>%d\n", i, (i == aaday) ? "selected" : "", i);
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<input type=\"text\" name=\"aayear\" value=\"%d\" size=\"4\">\n", aayear);
	filebuf_addf(filebuf, "Time: <input type=\"text\" name=\"aahour\" value=\"%.2d\" size=\"2\">:\n", aahour);
	filebuf_addf(filebuf, "<input type=\"text\" name=\"aaminute\" value=\"%.2d\" size=\"2\">\n", aaminute);
	filebuf_addf(filebuf, " Epoch time: <input type=\"checkbox\" name=\"aaepoch\" %s>", (aaepoch == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "</tr></td>\n");

	filebuf_addf(filebuf, "<tr><td>Accessed before</td>");
	filebuf_addf(filebuf, "<td>Date: <select name=\"abmonth\">\n");
	for (i = 0; calendar_months[i].name; i++)
		filebuf_addf(filebuf, calendar_months[i].html, (!strcasecmp(abmonth, calendar_months[i].name)) ? "selected" : "");
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<select name=\"abday\">\n");
	for (i = 1; i <= 31; i++)
		filebuf_addf(filebuf, "<option value=\"%d\" %s>%d\n", i, (i == abday) ? "selected" : "", i);
	filebuf_addf(filebuf, "</select>\n");
	filebuf_addf(filebuf, "<input type=\"text\" name=\"abyear\" value=\"%d\" size=\"4\">\n", abyear);
	filebuf_addf(filebuf, "Time: <input type=\"text\" name=\"abhour\" value=\"%.2d\" size=\"2\">:\n", abhour);
	filebuf_addf(filebuf, "<input type=\"text\" name=\"abminute\" value=\"%.2d\" size=\"2\">\n", abminute);
	filebuf_addf(filebuf, " Current time: <input type=\"checkbox\" name=\"abcurtime\" %s>", (abcurtime == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "</tr></td>\n");

	filebuf_addf(filebuf, "<tr><td>File size</td>");
	filebuf_addf(filebuf, "<td>All sizes: <input type=\"checkbox\" name=\"allsizes\" %s> \n", (allsizes == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "<input type=\"textbox\" name=\"minsize\" value=\"%u\" size=\"10\">", minsize);
	filebuf_addf(filebuf, " - <input type=\"textbox\" name=\"maxsize\" value=\"%u\" size=\"10\"> bytes", maxsize);
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\">Delete matches: <input type=\"checkbox\" name=\"delete\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr><tr><td><br></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	if (show == FALSE)
		goto out;

	filebuf_addf(filebuf, "<tr align=\"center\"><td><b>URL</b></td><td><b>Size</b></td><td><b>Mem</b></td><td><b>Disk</b></td></tr>\n");

	if (pattern != NULL && strcmp(pattern, ""))
		pe = reg_compile(pattern, REGFLAGS);

	for (i = 0; i < CACHEMAP_HASH_SIZE; i++) {
		for (cl = global->cache->cachemap[i]; cl; cl = tmp) {
			tmp = cl->next;
			cachemap = cl->cachemap;

			if (cachemap->mtime < madate || cachemap->mtime > mbdate)
				continue;
			if (cachemap->atime < aadate || cachemap->atime > abdate)
				continue;
			if (cachemap->size < minsize || cachemap->size > maxsize)
				continue;

			if (cachemap->refcount || (pe != NULL && reg_exec(pe, cachemap->key)))
				continue;

			if (delete == TRUE) {
				/* cache_invalidate_unlocked assumes caller has
				   cache item open */
				cachemap->refcount++;
				cache_invalidate_unlocked(global->cache, cachemap);
			} else {
				ptr2 = strstr(cachemap->key, "://");
				if (ptr2 == NULL)
					continue;
				if (ptr2 - cachemap->key + 1 > sizeof(proto))
					continue;

				s_strncpy(proto, cachemap->key, ptr2 - cachemap->key + 1);
				ptr2 += 3;

				ptr = string_to_html(cachemap->key, HTML_NEWLINES);

				filebuf_addf(filebuf, "<tr><td><a href=\"%s://cache..%s\">%s</a></td><td align=\"right\">%u</td><td align=\"center\">%s</td><td align=\"center\">%s</td></tr>\n", proto, ptr2, ptr, cachemap->size, (cachemap->flags & CACHE_MEM) ? "yes" : "no", (cachemap->flags & CACHE_DISK) ? "yes" : "no");
				xfree(ptr);
			}

		}
	}

	if (pe != NULL)
		reg_free(pe);

      out:

	pthread_mutex_unlock(&global->cache->lock);

	filebuf_addf(filebuf, "</table></td></tr></table>\n");
}


void interface_page_pool(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	time_t curtime;
	struct pool_t *pl;

	curtime = time(NULL);

	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	filebuf_addf(filebuf, "<tr><td>Protocol</td><td>Host</td><td>Port</td><td>Username</td><td>Age</td></tr>\n");

	pthread_mutex_lock(&global->pool->lock);

	for (pl = global->pool->pool_list; pl; pl = pl->next)
		filebuf_addf(filebuf, "<tr><td>%s</td><td>%s</td><td>%d</td><td>%s</td><td>%u</td></tr>\n", pl->proto, pl->host, pl->port, (pl->username != NULL) ? pl->username : "(none)", curtime - pl->age);

	pthread_mutex_unlock(&global->pool->lock);

	filebuf_addf(filebuf, "</table>");
}

void interface_page_config(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;
	char *section = NULL, *action = NULL, *dialog = NULL;
	struct cgi_args_t *a;

	for (a = args; a != NULL; a = a->next) {
		if (!strcasecmp(a->name, "section"))
			section = a->value;
		else if (!strcasecmp(a->name, "action"))
			action = a->value;
		else if (!strcasecmp(a->name, "dialog"))
			dialog = a->value;
		if (!strcasecmp(a->name, "filter")) {
			pthread_rwlock_wrlock(&filter_list->lock);
			if (!strcasecmp(a->value, "off"))
				filter_list->enabled = FALSE;
			else
				filter_list->enabled = TRUE;
			pthread_rwlock_unlock(&filter_list->lock);
		} else if (!strcasecmp(a->name, "header")) {
			pthread_rwlock_wrlock(&header_list->lock);
			if (!strcasecmp(a->value, "off"))
				header_list->enabled = FALSE;
			else
				header_list->enabled = TRUE;
			pthread_rwlock_unlock(&header_list->lock);
		} else if (!strcasecmp(a->name, "mime")) {
			pthread_rwlock_wrlock(&mime_list->lock);
			if (!strcasecmp(a->value, "off"))
				mime_list->enabled = FALSE;
			else
				mime_list->enabled = TRUE;
			pthread_rwlock_unlock(&mime_list->lock);
		} else if (!strcasecmp(a->name, "cookie")) {
			pthread_rwlock_wrlock(&cookie_list->lock);
			if (!strcasecmp(a->value, "off"))
				cookie_list->enabled = FALSE;
			else
				cookie_list->enabled = TRUE;
			pthread_rwlock_unlock(&cookie_list->lock);
		} else if (!strcasecmp(a->name, "redirect")) {
			pthread_rwlock_wrlock(&redirect_list->lock);
			if (!strcasecmp(a->value, "off"))
				redirect_list->enabled = FALSE;
			else
				redirect_list->enabled = TRUE;
			pthread_rwlock_unlock(&redirect_list->lock);
		} else if (!strcasecmp(a->name, "rewrite")) {
			pthread_rwlock_wrlock(&rewrite_list->lock);
			if (!strcasecmp(a->value, "off"))
				rewrite_list->enabled = FALSE;
			else
				rewrite_list->enabled = TRUE;
			pthread_rwlock_unlock(&rewrite_list->lock);
		} else if (!strcasecmp(a->name, "keywords")) {
			pthread_rwlock_wrlock(&keyword_list->lock);
			if (!strcasecmp(a->value, "off"))
				keyword_list->enabled = FALSE;
			else
				keyword_list->enabled = TRUE;
			pthread_rwlock_unlock(&keyword_list->lock);
		} else if (!strcasecmp(a->name, "external")) {
			pthread_rwlock_wrlock(&external->lock);
			if (!strcasecmp(a->value, "off"))
				external->enabled = FALSE;
			else
				external->enabled = TRUE;
			pthread_rwlock_unlock(&external->lock);
		} else if (!strcasecmp(a->name, "forward")) {
			pthread_rwlock_wrlock(&forward_list->lock);
			if (!strcasecmp(a->value, "off"))
				forward_list->enabled = FALSE;
			else
				forward_list->enabled = TRUE;
			pthread_rwlock_unlock(&forward_list->lock);
		} else if (!strcasecmp(a->name, "cache")) {
			pthread_mutex_lock(&global->cache->lock);
			if (!strcasecmp(a->value, "off"))
				global->cache->enabled = FALSE;
			else
				global->cache->enabled = TRUE;
			pthread_mutex_unlock(&global->cache->lock);
		} else if (!strcasecmp(a->name, "dnsbl")) {
			pthread_rwlock_wrlock(&global->dnsbl->lock);
			if (!strcasecmp(a->value, "on"))
				global->dnsbl->enabled = TRUE;
			else
				global->dnsbl->enabled = FALSE;
			pthread_rwlock_unlock(&global->dnsbl->lock);
		} else if (!strcasecmp(a->name, "limits")) {
			pthread_rwlock_wrlock(&limits_list->lock);
			if (!strcasecmp(a->value, "on"))
				limits_list->enabled = TRUE;
			else
				limits_list->enabled = FALSE;
			pthread_rwlock_unlock(&limits_list->lock);
		} else if (!strcasecmp(a->name, "prefetch")) {
			pthread_mutex_lock(&global->prefetch->lock);
			if (!strcasecmp(a->value, "on"))
				global->prefetch->enabled = TRUE;
			else
				global->prefetch->enabled = FALSE;
			pthread_mutex_unlock(&global->prefetch->lock);
		}
	}

	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\"><tr><td align=\"center\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<select name=\"section\">\n");
	filebuf_addf(filebuf, "<option value=\"\">Select a section\n");
	filebuf_addf(filebuf, "<option value=\"global\" %s>global\n", (section != NULL && !strcasecmp(section, "global")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"network\" %s>network\n", (section != NULL && !strcasecmp(section, "network")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"profiles\" %s>profiles\n", (section != NULL && !strcasecmp(section, "profiles")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"cache\" %s>cache\n", (section != NULL && !strcasecmp(section, "cache")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"access\" %s>access\n", (section != NULL && !strcasecmp(section, "access")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"filter\" %s>filter\n", (section != NULL && !strcasecmp(section, "filter")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"mime\" %s>mime\n", (section != NULL && !strcasecmp(section, "mime")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"redirect\" %s>redirect\n", (section != NULL && !strcasecmp(section, "redirect")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"forward\" %s>forward\n", (section != NULL && !strcasecmp(section, "forward")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"header\" %s>header\n", (section != NULL && !strcasecmp(section, "header")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"rewrite\" %s>rewrite\n", (section != NULL && !strcasecmp(section, "rewrite")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"keywords\" %s>keywords\n", (section != NULL && !strcasecmp(section, "keywords")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"templates\" %s>templates\n", (section != NULL && !strcasecmp(section, "templates")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"cookies\" %s>cookies\n", (section != NULL && !strcasecmp(section, "cookies")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"external\" %s>external\n", (section != NULL && !strcasecmp(section, "external")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"limits\" %s>limits\n", (section != NULL && !strcasecmp(section, "limits")) ? "selected" : "");
	filebuf_addf(filebuf, "</select></td>\n");
	filebuf_addf(filebuf, "<td align=\"center\"><input type=\"submit\" value=\"Submit\"></form></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td><br></td></tr><tr><td align=\"center\">\n");

	if (dialog != NULL) {
		for (i = 0; dialog_pages[i].file; i++) {
			if (!strcasecmp(section, dialog_pages[i].file))
				dialog_pages[i].handler(filebuf, args, connection);
		}
	} else if (section != NULL && strcmp(section, "")) {
		if (action != NULL) {
			for (i = 0; action_pages[i].file; i++) {
				if (!strcasecmp(section, action_pages[i].file))
					action_pages[i].handler(filebuf, args, connection);
			}
		}
		for (i = 0; config_pages[i].file; i++) {
			if (!strcasecmp(section, config_pages[i].file))
				config_pages[i].handler(filebuf, args, connection);
		}
	} else {
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\">Feature</td><td align=\"center\">Enabled</td><td align=\"center\">Disabled</td></tr>\n");

		pthread_mutex_lock(&global->cache->lock);
		filebuf_addf(filebuf, "<tr><td>Caching</td><td><input type=\"radio\" name=\"cache\" value=\"on\" %s></td><td><input type=\"radio\" name=\"cache\" value=\"off\" %s></td></tr>\n", (global->cache->enabled == TRUE) ? "checked" : "", (global->cache->enabled == FALSE) ? "checked" : "");
		pthread_mutex_unlock(&global->cache->lock);

		pthread_mutex_lock(&global->prefetch->lock);
		filebuf_addf(filebuf, "<tr><td>Prefetching</td><td><input type=\"radio\" name=\"prefetch\" value=\"on\" %s></td><td><input type=\"radio\" name=\"prefetch\" value=\"off\" %s></td></tr>\n", (global->prefetch->enabled == TRUE) ? "checked" : "", (global->prefetch->enabled == FALSE) ? "checked" : "");
		pthread_mutex_unlock(&global->prefetch->lock);

		pthread_rwlock_rdlock(&global->dnsbl->lock);
		filebuf_addf(filebuf, "<tr><td>DNS blacklist</td><td><input type=\"radio\" name=\"dnsbl\" value=\"on\" %s></td><td><input type=\"radio\" name=\"dnsbl\" value=\"off\" %s></td></tr>\n", (global->dnsbl->enabled == TRUE) ? "checked" : "", (global->dnsbl->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&global->dnsbl->lock);

		pthread_rwlock_rdlock(&filter_list->lock);
		filebuf_addf(filebuf, "<tr><td>URL filtering</td><td><input type=\"radio\" name=\"filter\" value=\"on\" %s></td><td><input type=\"radio\" name=\"filter\" value=\"off\" %s></td></tr>\n", (filter_list->enabled == TRUE) ? "checked" : "", (filter_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&filter_list->lock);

		pthread_rwlock_rdlock(&header_list->lock);
		filebuf_addf(filebuf, "<tr><td>Header filtering</td><td><input type=\"radio\" name=\"header\" value=\"on\" %s></td><td><input type=\"radio\" name=\"header\" value=\"off\" %s></td></tr>\n", (header_list->enabled == TRUE) ? "checked" : "", (header_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&header_list->lock);

		pthread_rwlock_rdlock(&mime_list->lock);
		filebuf_addf(filebuf, "<tr><td>MIME filtering</td><td><input type=\"radio\" name=\"mime\" value=\"on\" %s></td><td><input type=\"radio\" name=\"mime\" value=\"off\" %s></td></tr>\n", (mime_list->enabled == TRUE) ? "checked" : "", (mime_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&mime_list->lock);

		pthread_rwlock_rdlock(&cookie_list->lock);
		filebuf_addf(filebuf, "<tr><td>Cookie filtering</td><td><input type=\"radio\" name=\"cookie\" value=\"on\" %s></td><td><input type=\"radio\" name=\"cookie\" value=\"off\" %s></td></tr>\n", (cookie_list->enabled == TRUE) ? "checked" : "", (cookie_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&cookie_list->lock);

		pthread_rwlock_rdlock(&redirect_list->lock);
		filebuf_addf(filebuf, "<tr><td>URL redirecting</td><td><input type=\"radio\" name=\"redirect\" value=\"on\" %s></td><td><input type=\"radio\" name=\"redirect\" value=\"off\" %s></td></tr>\n", (redirect_list->enabled == TRUE) ? "checked" : "", (redirect_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&redirect_list->lock);

		pthread_rwlock_rdlock(&forward_list->lock);
		filebuf_addf(filebuf, "<tr><td>Forwarding</td><td><input type=\"radio\" name=\"forward\" value=\"on\" %s></td><td><input type=\"radio\" name=\"forward\" value=\"off\" %s></td></tr>\n", (forward_list->enabled == TRUE) ? "checked" : "", (forward_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&forward_list->lock);

		pthread_rwlock_rdlock(&rewrite_list->lock);
		filebuf_addf(filebuf, "<tr><td>Document rewriting</td><td><input type=\"radio\" name=\"rewrite\" value=\"on\" %s></td><td><input type=\"radio\" name=\"rewrite\" value=\"off\" %s></td></tr>\n", (rewrite_list->enabled == TRUE) ? "checked" : "", (rewrite_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&rewrite_list->lock);

		pthread_rwlock_rdlock(&keyword_list->lock);
		filebuf_addf(filebuf, "<tr><td>Keyword filtering</td><td><input type=\"radio\" name=\"keywords\" value=\"on\" %s></td><td><input type=\"radio\" name=\"keywords\" value=\"off\" %s></td></tr>\n", (keyword_list->enabled == TRUE) ? "checked" : "", (keyword_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&keyword_list->lock);

		pthread_rwlock_rdlock(&external->lock);
		filebuf_addf(filebuf, "<tr><td>External parsers</td><td><input type=\"radio\" name=\"external\" value=\"on\" %s></td><td><input type=\"radio\" name=\"external\" value=\"off\" %s></td></tr>\n", (external->enabled == TRUE) ? "checked" : "", (external->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&external->lock);

		pthread_rwlock_rdlock(&limits_list->lock);
		filebuf_addf(filebuf, "<tr><td>Limits</td><td><input type=\"radio\" name=\"limits\" value=\"on\" %s></td><td><input type=\"radio\" name=\"limits\" value=\"off\" %s></td></tr>\n", (limits_list->enabled == TRUE) ? "checked" : "", (limits_list->enabled == FALSE) ? "checked" : "");
		pthread_rwlock_unlock(&limits_list->lock);

		filebuf_addf(filebuf, "<tr><td colspan=\"3\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
		filebuf_addf(filebuf, "</form></table></td></tr></table>\n");
	}

	filebuf_addf(filebuf, "</td></tr></table>\n");
}

void interface_page_config_filter(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;
	char *ptr;

	struct FILTER_LIST_LIST *fl = NULL;
	if (filter_list == NULL)
		return;

	pthread_rwlock_rdlock(&filter_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"filter\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Policy</td><td><select name=\"policy\"><option value=\"allow\" %s>allow<option value=\"deny\" %s>deny</select></td></tr>", (filter_list->policy == POLICY_ALLOW) ? "selected" : "", (filter_list->policy == POLICY_DENY) ? "selected" : "");
	filebuf_addf(filebuf, "<tr><td>Default template</td><td><input type=\"text\" name=\"dtemplate\" value=\"%s\" size=\"20\"></td></tr>\n", (filter_list->dtemplate != NULL) ? filter_list->dtemplate : "");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=filter&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr> <tr><td><br></td></tr>\n");

	for (i = 0; i < 2; i++) {
		filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"30%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		switch (i) {
		case 0:
			fl = filter_list->allow;
			if (fl != NULL)
				filebuf_addf(filebuf, "<tr><th>ALLOW</th></tr>\n");
			break;
		case 1:
			fl = filter_list->deny;
			if (fl != NULL)
				filebuf_addf(filebuf, "<tr><th>DENY</th></tr>\n");
			break;
		}

		filebuf_addf(filebuf, "</table></td></tr>\n");

		if (fl != NULL)
			filebuf_addf(filebuf, "<tr><td><br></td></tr>\n");

		for (; fl; fl = fl->next) {
			filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
			filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (fl->enabled == TRUE) ? "yes" : "no");

			if (fl->comment != NULL) {
				ptr = string_to_html(fl->comment, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (fl->profiles != NULL) {
				ptr = array_merge(fl->profiles, ',');
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (fl->host != NULL) {
				ptr = string_to_html(fl->host, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Host</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (fl->file != NULL) {
				ptr = string_to_html(fl->file, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">File</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (fl->template != NULL)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Template</td><td>%s</td></tr>\n", fl->template);

			filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
			filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=filter&amp;dialog=show&amp;list=%s&amp;id=%d\">Edit</a> <a href=\"config?section=filter&amp;action=delete&amp;list=%s&amp;id=%d\">Delete</a></td>\n", (i == 0) ? "allow" : "deny", fl->id, (i == 0) ? "allow" : "deny", fl->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=filter&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=filter&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=down\">Down</a></td>\n", (i == 0) ? "allow" : "deny", fl->id, (i == 0) ? "allow" : "deny", fl->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=filter&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=filter&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", (i == 0) ? "allow" : "deny", fl->id, (i == 0) ? "allow" : "deny", fl->id);
			filebuf_addf(filebuf, "</table></td></tr>\n");
			filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
		}
	}

	filebuf_addf(filebuf, "</table>\n");
	pthread_rwlock_unlock(&filter_list->lock);
}

void interface_page_config_filter_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *ptr = NULL;
	struct FILTER_LIST_LIST *fl = NULL;
	struct cgi_args_t *a = args;

	if (filter_list == NULL)
		return;

	pthread_rwlock_rdlock(&filter_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (id != -1 && list != NULL) {
		if (!strcasecmp(list, "allow"))
			for (fl = filter_list->allow; fl && fl->id != id; fl = fl->next);
		else if (!strcasecmp(list, "deny"))
			for (fl = filter_list->deny; fl && fl->id != id; fl = fl->next);
	}

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"filter\">\n");

	if (list != NULL)
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"list\" value=\"%s\">\n", list);
	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (fl == NULL || fl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (fl != NULL && fl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (fl != NULL && fl->comment != NULL)
		ptr = string_to_html(fl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->profiles != NULL)
		ptr = array_merge(fl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Action</td><td><select name=\"%s\">\n", (list != NULL) ? "move" : "list");
	filebuf_addf(filebuf, "<option value=\"allow\" %s>allow\n", (list != NULL && !strcasecmp(list, "allow")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"deny\" %s>deny\n", (list != NULL && !strcasecmp(list, "deny")) ? "selected" : "");
	filebuf_addf(filebuf, "</select></tr>\n");
	if (fl != NULL && fl->host != NULL)
		ptr = string_to_html(fl->host, 0);
	filebuf_addf(filebuf, "<tr><td>Host</td><td><input type=\"text\" name=\"host\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->file != NULL)
		ptr = string_to_html(fl->file, 0);
	filebuf_addf(filebuf, "<tr><td>File</td><td><input type=\"text\" name=\"file\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->template != NULL)
		ptr = string_to_html(fl->template, 0);
	filebuf_addf(filebuf, "<tr><td>Template</td><td><input type=\"text\" name=\"template\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");
	pthread_rwlock_unlock(&filter_list->lock);
}


void interface_page_config_filter_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *action = NULL;
	struct FILTER_LIST_LIST *fl = NULL;
	struct cgi_args_t *a;

	if (filter_list == NULL)
		return;

	pthread_rwlock_wrlock(&filter_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (action != NULL && !strcasecmp(action, "global")) {
		for (a = args; a; a = a->next) {
			if (!strcasecmp(a->name, "policy")) {
				if (!strcasecmp(a->value, "allow"))
					filter_list->policy = POLICY_ALLOW;
				else if (!strcasecmp(a->value, "deny"))
					filter_list->policy = POLICY_DENY;
			} else if (!strcasecmp(a->name, "dtemplate")) {
				FREE_AND_NULL(filter_list->dtemplate);
				if (strcmp(a->name, ""))
					filter_list->dtemplate = xstrdup(a->value);
			}
		}
	}

	if (action == NULL || list == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		if (!strcasecmp(list, "allow"))
			fl = filter_list->allow;
		else if (!strcasecmp(list, "deny"))
			fl = filter_list->deny;
		for (; fl; fl = fl->next) {
			if (fl->id == id) {
				if (!strcasecmp(list, "allow"))
					filter_list->allow = filter_ll_delete(fl);
				else if (!strcasecmp(list, "deny"))
					filter_list->deny = filter_ll_delete(fl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			if (!strcasecmp(list, "allow")) {
				fl = filter_ll_new(filter_list->allow);
				if (filter_list->allow == NULL)
					filter_list->allow = fl;
			} else if (!strcasecmp(list, "deny")) {
				fl = filter_ll_new(filter_list->deny);
				if (filter_list->deny == NULL)
					filter_list->deny = fl;
			} else
				goto finish;
			fl->id = filter_list->id++;
		} else {
			if (!strcasecmp(list, "allow")) {
				for (fl = filter_list->allow; fl; fl = fl->next)
					if (fl->id == id)
						break;
			} else if (!strcasecmp(list, "deny")) {
				for (fl = filter_list->deny; fl; fl = fl->next)
					if (fl->id == id)
						break;
			}

			if (fl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						fl->enabled = FALSE;
					else
						fl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					filter_ll_insert(fl, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					filter_ll_insert(fl, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "host"))
					filter_ll_insert(fl, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "file"))
					filter_ll_insert(fl, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "template"))
					filter_ll_insert(fl, NULL, NULL, NULL, NULL, a->value);
				else if (!strcasecmp(a->name, "move")) {
					if (!strcasecmp(a->value, "allow") && strcasecmp(list, a->value)) {
						MOVENODE(struct FILTER_LIST_LIST *, fl, filter_list->deny, filter_list->allow);
					} else if (!strcasecmp(a->value, "deny") && strcasecmp(list, a->value)) {
						MOVENODE(struct FILTER_LIST_LIST *, fl, filter_list->allow, filter_list->deny);
					}
				}
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct FILTER_LIST_LIST *, (!strcasecmp(list, "allow")) ? filter_list->allow : filter_list->deny, fl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct FILTER_LIST_LIST *, (!strcasecmp(list, "allow")) ? filter_list->allow : filter_list->deny, fl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct FILTER_LIST_LIST *, (!strcasecmp(list, "allow")) ? filter_list->allow : filter_list->deny, fl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct FILTER_LIST_LIST *, (!strcasecmp(list, "allow")) ? filter_list->allow : filter_list->deny, fl, BOTTOM);
					}
				}
			}
		}
	}

      finish:
	pthread_rwlock_unlock(&filter_list->lock);
}

void interface_page_config_header(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connectin)
{
	int i;
	char *ptr;
	struct HEADER_LIST_LIST *hl = NULL;

	if (header_list == NULL)
		return;

	pthread_rwlock_rdlock(&header_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"header\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Policy</td><td><select name=\"policy\"><option value=\"allow\" %s>allow<option value=\"deny\" %s>deny</select></td></tr>", (header_list->policy == POLICY_ALLOW) ? "selected" : "", (header_list->policy == POLICY_DENY) ? "selected" : "");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=header&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr> <tr><td><br></td></tr>\n");

	for (i = 0; i < 3; i++) {
		filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"30%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		switch (i) {
		case 0:
			hl = header_list->allow;
			if (hl != NULL)
				filebuf_addf(filebuf, "<tr><th>ALLOW</th></tr>\n");
			break;
		case 1:
			hl = header_list->deny;
			if (hl != NULL)
				filebuf_addf(filebuf, "<tr><th>DENY</th></tr>\n");
			break;
		case 2:
			hl = header_list->insert;
			if (hl != NULL)
				filebuf_addf(filebuf, "<tr><th>INSERT</th></tr>\n");
		}

		filebuf_addf(filebuf, "</table></td></tr><tr>\n");

		if (hl != NULL)
			filebuf_addf(filebuf, "<tr><td><br></td></tr>\n");

		for (; hl; hl = hl->next) {
			filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
			filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (hl->enabled == TRUE) ? "yes" : "no");

			if (hl->comment != NULL) {
				ptr = string_to_html(hl->comment, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (hl->profiles != NULL) {
				ptr = array_merge(hl->profiles, ',');
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (hl->type != NULL) {
				ptr = string_to_html(hl->type, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Type</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (hl->value != NULL) {
				ptr = string_to_html(hl->value, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Value</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (hl->which != 0)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Applies to</td><td>%s,%s</td></tr>\n", (hl->which & HEADER_SERVER) ? "server header" : "", (hl->which & HEADER_CLIENT) ? "client header" : "");

			filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
			filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=header&amp;dialog=show&amp;list=%s&amp;id=%d\">Edit</a> <a href=\"config?section=header&amp;action=delete&amp;list=%s&amp;id=%d\">Delete</a></td>\n", (i == 0) ? "allow" : (i == 1) ? "deny" : "insert", hl->id, (i == 0) ? "allow" : (i == 1) ? "deny" : "insert", hl->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=header&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=header&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=down\">Down</a></td>\n", (i == 0) ? "allow" : (i == 1) ? "deny" : "insert", hl->id, (i == 0) ? "allow" : (i == 1) ? "deny" : "insert", hl->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=header&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=header&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", (i == 0) ? "allow" : (i == 1) ? "deny" : "insert", hl->id, (i == 0) ? "allow" : (i == 1) ? "deny" : "insert", hl->id);
			filebuf_addf(filebuf, "</table></td></tr>\n");
			filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
		}
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&header_list->lock);
}

void interface_page_config_header_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *ptr = NULL;
	struct HEADER_LIST_LIST *hl = NULL;
	struct cgi_args_t *a = args;

	if (header_list == NULL)
		return;

	pthread_rwlock_rdlock(&header_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (id != -1 && list != NULL) {
		if (!strcasecmp(list, "allow"))
			for (hl = header_list->allow; hl && hl->id != id; hl = hl->next);
		else if (!strcasecmp(list, "deny"))
			for (hl = header_list->deny; hl && hl->id != id; hl = hl->next);
		else if (!strcasecmp(list, "insert"))
			for (hl = header_list->insert; hl && hl->id != id; hl = hl->next);
	}

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"header\">\n");

	if (list != NULL)
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"list\" value=\"%s\">\n", list);
	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (hl == NULL || hl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (hl != NULL && hl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (hl != NULL && hl->comment != NULL)
		ptr = string_to_html(hl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (hl != NULL && hl->profiles != NULL)
		ptr = array_merge(hl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Action</td><td><select name=\"%s\">\n", (list != NULL) ? "move" : "list");
	filebuf_addf(filebuf, "<option value=\"allow\" %s>allow\n", (list != NULL && !strcasecmp(list, "allow")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"deny\" %s>deny\n", (list != NULL && !strcasecmp(list, "deny")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"insert\" %s>insert\n", (list != NULL && !strcasecmp(list, "insert")) ? "selected" : "");
	filebuf_addf(filebuf, "</select></tr>\n");

	if (hl != NULL && hl->type != NULL)
		ptr = string_to_html(hl->type, 0);
	filebuf_addf(filebuf, "<tr><td>Type</td><td><input type=\"text\" name=\"type\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (hl != NULL && hl->value != NULL)
		ptr = string_to_html(hl->value, 0);
	filebuf_addf(filebuf, "<tr><td>Value</td><td><input type=\"text\" name=\"value\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Applies to</td><td align=\"center\">");
	filebuf_addf(filebuf, "Server header: <input type=\"checkbox\" name=\"server\" %s>   ", (hl != NULL && hl->which & HEADER_SERVER) ? "checked" : "");
	filebuf_addf(filebuf, "Client header: <input type=\"checkbox\" name=\"client\" %s>   ", (hl != NULL && hl->which & HEADER_CLIENT) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&header_list->lock);
}


void interface_page_config_header_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *action = NULL;
	struct HEADER_LIST_LIST *hl = NULL;
	struct cgi_args_t *a;

	if (header_list == NULL)
		return;

	pthread_rwlock_wrlock(&header_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (action != NULL && !strcasecmp(action, "global")) {
		for (a = args; a; a = a->next) {
			if (!strcasecmp(a->name, "policy")) {
				if (!strcasecmp(a->value, "allow"))
					header_list->policy = POLICY_ALLOW;
				else if (!strcasecmp(a->value, "deny"))
					header_list->policy = POLICY_DENY;
			}
		}
	}

	if (action == NULL || list == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		if (!strcasecmp(list, "allow"))
			hl = header_list->allow;
		else if (!strcasecmp(list, "deny"))
			hl = header_list->deny;
		else if (!strcasecmp(list, "insert"))
			hl = header_list->insert;
		for (; hl; hl = hl->next) {
			if (hl->id == id) {
				if (!strcasecmp(list, "allow"))
					header_list->allow = header_ll_delete(hl);
				else if (!strcasecmp(list, "deny"))
					header_list->deny = header_ll_delete(hl);
				else if (!strcasecmp(list, "insert"))
					header_list->insert = header_ll_delete(hl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			if (!strcasecmp(list, "allow")) {
				hl = header_ll_new(header_list->allow);
				if (header_list->allow == NULL)
					header_list->allow = hl;
			} else if (!strcasecmp(list, "deny")) {
				hl = header_ll_new(header_list->deny);
				if (header_list->deny == NULL)
					header_list->deny = hl;
			} else if (!strcasecmp(list, "insert")) {
				hl = header_ll_new(header_list->insert);
				if (header_list->insert == NULL)
					header_list->insert = hl;
			} else
				goto finish;
			hl->id = header_list->id++;
		} else {
			if (!strcasecmp(list, "allow")) {
				for (hl = header_list->allow; hl; hl = hl->next)
					if (hl->id == id)
						break;
			} else if (!strcasecmp(list, "deny")) {
				for (hl = header_list->deny; hl; hl = hl->next)
					if (hl->id == id)
						break;
			} else if (!strcasecmp(list, "insert")) {
				for (hl = header_list->insert; hl; hl = hl->next)
					if (hl->id == id)
						break;
			}

			if (hl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			hl->which = 0;
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						hl->enabled = FALSE;
					else
						hl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					header_ll_insert(hl, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					header_ll_insert(hl, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "type"))
					header_ll_insert(hl, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "value"))
					header_ll_insert(hl, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "server"))
					hl->which |= HEADER_SERVER;
				else if (!strcasecmp(a->name, "client"))
					hl->which |= HEADER_CLIENT;
				else if (!strcasecmp(a->name, "move")) {
					if (!strcasecmp(a->value, "allow") && strcasecmp(list, a->value)) {
						if (!strcasecmp(list, "deny")) {
							MOVENODE(struct HEADER_LIST_LIST *, hl, header_list->deny, header_list->allow);
						} else if (!strcasecmp(list, "insert")) {
							MOVENODE(struct HEADER_LIST_LIST *, hl, header_list->insert, header_list->allow);
						}
					} else if (!strcasecmp(a->value, "deny") && strcasecmp(list, a->value)) {
						if (!strcasecmp(list, "allow")) {
							MOVENODE(struct HEADER_LIST_LIST *, hl, header_list->allow, header_list->deny);
						} else if (!strcasecmp(list, "insert")) {
							MOVENODE(struct HEADER_LIST_LIST *, hl, header_list->insert, header_list->deny);
						}
					} else if (!strcasecmp(a->value, "insert") && strcasecmp(list, a->value)) {
						if (!strcasecmp(list, "allow")) {
							MOVENODE(struct HEADER_LIST_LIST *, hl, header_list->allow, header_list->insert);
						} else if (!strcasecmp(list, "deny")) {
							MOVENODE(struct HEADER_LIST_LIST *, hl, header_list->deny, header_list->insert);
						}
					}
				}
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct HEADER_LIST_LIST *, (!strcasecmp(list, "allow")) ? header_list->allow : (!strcasecmp(list, "deny")) ? header_list->deny : header_list->insert, hl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct HEADER_LIST_LIST *, (!strcasecmp(list, "allow")) ? header_list->allow : (!strcasecmp(list, "deny")) ? header_list->deny : header_list->insert, hl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct HEADER_LIST_LIST *, (!strcasecmp(list, "allow")) ? header_list->allow : (!strcasecmp(list, "deny")) ? header_list->deny : header_list->insert, hl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct HEADER_LIST_LIST *, (!strcasecmp(list, "allow")) ? header_list->allow : (!strcasecmp(list, "deny")) ? header_list->deny : header_list->insert, hl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&header_list->lock);
}

void interface_page_config_access(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;
	char *ptr;
	struct ACCESS_LIST_LIST *al = NULL;

	if (access_list == NULL)
		return;

	pthread_rwlock_rdlock(&access_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"access\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Policy</td><td><select name=\"policy\"><option value=\"allow\" %s>allow<option value=\"deny\" %s>deny</select></td></tr>", (access_list->policy == POLICY_ALLOW) ? "selected" : "", (access_list->policy == POLICY_DENY) ? "selected" : "");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=access&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr> <tr><td><br></td></tr>\n");

	for (i = 0; i < 2; i++) {
		filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"30%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		switch (i) {
		case 0:
			al = access_list->allow;
			if (al != NULL)
				filebuf_addf(filebuf, "<tr><th>ALLOW</th></tr>\n");
			break;
		case 1:
			al = access_list->deny;
			if (al != NULL)
				filebuf_addf(filebuf, "<tr><th>DENY</th></tr>\n");
			break;
		}

		filebuf_addf(filebuf, "</table></td></tr>\n");
		if (al != NULL)
			filebuf_addf(filebuf, "<tr><td><br></td></tr>\n");
		for (; al; al = al->next) {
			filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
			filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (al->enabled == TRUE) ? "yes" : "no");
			if (al->comment != NULL) {
				ptr = string_to_html(al->comment, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}
			if (al->profiles != NULL) {
				ptr = array_merge(al->profiles, ',');
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}
			if (al->ip != NULL) {
				ptr = string_to_html(al->ip, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">IP address</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}
#ifdef HAVE_PAM
			filebuf_addf(filebuf, "<tr><td widht=\"10%%\">PAM authentication</td><td>%s</td></tr>\n", (al->pamauth == TRUE) ? "Yes" : "No");
#endif				/* HAVE_PAM */
			if (al->username != NULL) {
				ptr = string_to_html(al->username, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Username</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}
			if (al->password != NULL) {
				ptr = string_to_html(al->password, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Password</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (al->bypass != 0)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Bypass</td><td>%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s</td></tr>\n", (al->bypass & FEATURE_FILTER) ? "filter" : "", (al->bypass & FEATURE_HEADER) ? "header" : "", (al->bypass & FEATURE_MIME) ? "mime" : "", (al->bypass & FEATURE_REDIRECT) ? "redirect" : "", (al->bypass & FEATURE_COOKIES) ? "cookies" : "", (al->bypass & FEATURE_REWRITE) ? "rewrite" : "", (al->bypass & FEATURE_EXTERNAL) ? "external" : "", (al->bypass & FEATURE_FORWARD) ? "forward" : "", (al->bypass & FEATURE_KEYWORDS) ? "keywords" : "", (al->bypass & FEATURE_DNSBL) ? "dnsbl" : "", (al->bypass & FEATURE_LIMITS) ? "limits" : "");
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Access</td><td>%s,%s,%s,%s,%s,%s,%s</td.</tr>\n", (al->access & ACCESS_CONFIG) ? "Web interface" : "", (al->access & ACCESS_PROXY) ? "Proxy requests" : "", (al->access & ACCESS_CONNECT) ? "CONNECT requests" : "", (al->access & ACCESS_HTTP) ? "HTTP requests" : "", (al->access & ACCESS_TRANSPARENT) ? "Transparent proxying" : "", (al->access & ACCESS_BYPASS) ? "Allow bypassing" : "", (al->access & ACCESS_URLCOMMAND) ? "URL commands" : "");
			filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
			filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=access&amp;dialog=show&amp;list=%s&amp;id=%d\">Edit</a> <a href=\"config?section=access&amp;action=delete&amp;list=%s&amp;id=%d\">Delete</a></td>\n", (i == 0) ? "allow" : "deny", al->id, (i == 0) ? "allow" : "deny", al->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=access&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=access&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=down\">Down</a></td>\n", (i == 0) ? "allow" : "deny", al->id, (i == 0) ? "allow" : "deny", al->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=access&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=access&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", (i == 0) ? "allow" : "deny", al->id, (i == 0) ? "allow" : "deny", al->id);
			filebuf_addf(filebuf, "</table></td></tr>\n");
			filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
		}
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&access_list->lock);
}

void interface_page_config_access_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *ptr = NULL;
	struct ACCESS_LIST_LIST *al = NULL;
	struct cgi_args_t *a = args;

	if (access_list == NULL)
		return;

	pthread_rwlock_rdlock(&access_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (id != -1 && list != NULL) {
		if (!strcasecmp(list, "allow"))
			for (al = access_list->allow; al && al->id != id; al = al->next);
		else if (!strcasecmp(list, "deny"))
			for (al = access_list->deny; al && al->id != id; al = al->next);
	}

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"access\">\n");

	if (list != NULL)
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"list\" value=\"%s\">\n", list);
	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (al == NULL || al->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (al != NULL && al->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (al != NULL && al->comment != NULL)
		ptr = string_to_html(al->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (al != NULL && al->profiles != NULL)
		ptr = array_merge(al->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Action</td><td><select name=\"%s\">\n", (list != NULL) ? "move" : "list");
	filebuf_addf(filebuf, "<option value=\"allow\" %s>allow\n", (list != NULL && !strcasecmp(list, "allow")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"deny\" %s>deny\n", (list != NULL && !strcasecmp(list, "deny")) ? "selected" : "");
	filebuf_addf(filebuf, "</select></tr>\n");

	if (al != NULL && al->ip != NULL)
		ptr = string_to_html(al->ip, 0);
	filebuf_addf(filebuf, "<tr><td>IP Address</td><td><input type=\"text\" name=\"ip\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

#ifdef HAVE_PAM
	filebuf_addf(filebuf, "<tr><td>PAM authentication</td><td>Yes: <input type=\"radio\" name=\"pamauth\" value=\"yes\" %s> No: <input type=\"radio\" name=\"pamauth\" value=\"no\" %s></td></tr>\n", (al != NULL && al->pamauth == TRUE) ? "checked" : "", (al == NULL || al->pamauth == FALSE) ? "checked" : "");
#endif				/* HAVE_PAM */

	if (al != NULL && al->username != NULL)
		ptr = string_to_html(al->username, 0);
	filebuf_addf(filebuf, "<tr><td>Username</td><td><input type=\"text\" name=\"username\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (al != NULL && al->password != NULL)
		ptr = string_to_html(al->password, 0);
	filebuf_addf(filebuf, "<tr><td>Password</td><td><input type=\"text\" name=\"password\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Access</td><td align=\"center\">");
	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\">\n");
	filebuf_addf(filebuf, "<tr><td>Web interface: </td><td><input type=\"checkbox\" name=\"access_config\" %s></td></tr>", (al != NULL && (al->access & ACCESS_CONFIG)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Proxy requests: </td><td><input type=\"checkbox\" name=\"access_proxy\" %s></td></tr>", (al != NULL && (al->access & ACCESS_PROXY)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>CONNECT requests: </td><td><input type=\"checkbox\" name=\"access_connect\" %s></td></tr>", (al != NULL && (al->access & ACCESS_CONNECT)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>HTTP requests: </td><td><input type=\"checkbox\" name=\"access_http\" %s></td></tr>", (al != NULL && (al->access & ACCESS_HTTP)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Transparent proxying: </td><td><input type=\"checkbox\" name=\"access_transparent\" %s></td></tr>", (al != NULL && (al->access & ACCESS_TRANSPARENT)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Allow bypassing: </td><td><input type=\"checkbox\" name=\"access_bypass\" %s></td></tr>", (al != NULL && (al->access & ACCESS_BYPASS)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>URL commands: </td><td><input type=\"checkbox\" name=\"access_urlcommand\" %s></td></tr>", (al != NULL && (al->access & ACCESS_URLCOMMAND)) ? "checked" : "");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Bypass</td><td align=\"center\">");
	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\">\n");
	filebuf_addf(filebuf, "<tr><td>URL filtering: </td><td><input type=\"checkbox\" name=\"bfilter\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_FILTER)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Header filtering: </td><td><input type=\"checkbox\" name=\"bheader\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_HEADER)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>MIME filtering: </td><td><input type=\"checkbox\" name=\"bmime\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_MIME)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>URL redirecting: </td><td><input type=\"checkbox\" name=\"bredirect\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_REDIRECT)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Cookie filtering: </td><td><input type=\"checkbox\" name=\"bcookies\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_COOKIES)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Document rewriting: </td><td><input type=\"checkbox\" name=\"brewrite\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_REWRITE)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>External parsers: </td><td><input type=\"checkbox\" name=\"bexternal\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_EXTERNAL)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Forwarding: </td><td><input type=\"checkbox\" name=\"bforward\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_FORWARD)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Keyword filtering: </td><td><input type=\"checkbox\" name=\"bkeywords\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_KEYWORDS)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>DNS blacklist: </td><td><input type=\"checkbox\" name=\"bdnsbl\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_DNSBL)) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Limits: </td><td><input type=\"checkbox\" name=\"limits\" %s></td></tr>", (al != NULL && (al->bypass & FEATURE_LIMITS)) ? "checked" : "");

	filebuf_addf(filebuf, "</table></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&access_list->lock);
}


void interface_page_config_access_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *action = NULL;
	struct ACCESS_LIST_LIST *al = NULL;
	struct cgi_args_t *a;

	if (access_list == NULL)
		return;

	pthread_rwlock_wrlock(&access_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (action != NULL && !strcasecmp(action, "global")) {
		for (a = args; a; a = a->next) {
			if (!strcasecmp(a->name, "policy")) {
				if (!strcasecmp(a->value, "allow"))
					access_list->policy = POLICY_ALLOW;
				else if (!strcasecmp(a->value, "deny"))
					access_list->policy = POLICY_DENY;
			}
		}
	}

	if (action == NULL || list == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		if (!strcasecmp(list, "allow"))
			al = access_list->allow;
		else if (!strcasecmp(list, "deny"))
			al = access_list->deny;
		for (; al; al = al->next) {
			if (al->id == id) {
				if (!strcasecmp(list, "allow"))
					access_list->allow = access_ll_delete(al);
				else if (!strcasecmp(list, "deny"))
					access_list->deny = access_ll_delete(al);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			if (!strcasecmp(list, "allow")) {
				al = access_ll_new(access_list->allow);
				if (access_list->allow == NULL)
					access_list->allow = al;
			} else if (!strcasecmp(list, "deny")) {
				al = access_ll_new(access_list->deny);
				if (access_list->deny == NULL)
					access_list->deny = al;
			} else
				goto finish;
			al->id = access_list->id++;
		} else {
			if (!strcasecmp(list, "allow")) {
				for (al = access_list->allow; al; al = al->next)
					if (al->id == id)
						break;
			} else if (!strcasecmp(list, "deny")) {
				for (al = access_list->deny; al; al = al->next)
					if (al->id == id)
						break;
			}

			if (al == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			al->access = al->bypass = 0;
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						al->enabled = FALSE;
					else
						al->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					access_ll_insert(al, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					access_ll_insert(al, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "ip"))
					access_ll_insert(al, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "pamauth")) {
					if (!strcasecmp(a->value, "yes"))
						al->pamauth = TRUE;
					else
						al->pamauth = FALSE;
				} else if (!strcasecmp(a->name, "username"))
					access_ll_insert(al, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "password"))
					access_ll_insert(al, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
				else if (!strcasecmp(a->name, "access_config"))
					al->access |= ACCESS_CONFIG;
				else if (!strcasecmp(a->name, "access_proxy"))
					al->access |= ACCESS_PROXY;
				else if (!strcasecmp(a->name, "access_connect"))
					al->access |= ACCESS_CONNECT;
				else if (!strcasecmp(a->name, "access_http"))
					al->access |= ACCESS_HTTP;
				else if (!strcasecmp(a->name, "access_transparent"))
					al->access |= ACCESS_TRANSPARENT;
				else if (!strcasecmp(a->name, "access_bypass"))
					al->access |= ACCESS_BYPASS;
				else if (!strcasecmp(a->name, "access_urlcommand"))
					al->access |= ACCESS_URLCOMMAND;
				else if (!strcasecmp(a->name, "bfilter"))
					al->bypass |= FEATURE_FILTER;
				else if (!strcasecmp(a->name, "bheader"))
					al->bypass |= FEATURE_HEADER;
				else if (!strcasecmp(a->name, "bmime"))
					al->bypass |= FEATURE_MIME;
				else if (!strcasecmp(a->name, "bredirect"))
					al->bypass |= FEATURE_REDIRECT;
				else if (!strcasecmp(a->name, "bcookies"))
					al->bypass |= FEATURE_COOKIES;
				else if (!strcasecmp(a->name, "brewrite"))
					al->bypass |= FEATURE_REWRITE;
				else if (!strcasecmp(a->name, "bexternal"))
					al->bypass |= FEATURE_EXTERNAL;
				else if (!strcasecmp(a->name, "bforward"))
					al->bypass |= FEATURE_FORWARD;
				else if (!strcasecmp(a->name, "bkeywords"))
					al->bypass |= FEATURE_KEYWORDS;
				else if (!strcasecmp(a->name, "bdnsbl"))
					al->bypass |= FEATURE_DNSBL;
				else if (!strcasecmp(a->name, "limits"))
					al->bypass |= FEATURE_LIMITS;
				else if (!strcasecmp(a->name, "move")) {
					if (!strcasecmp(a->value, "allow") && strcasecmp(list, a->value)) {
						MOVENODE(struct ACCESS_LIST_LIST *, al, access_list->deny, access_list->allow);
					} else if (!strcasecmp(a->value, "deny") && strcasecmp(list, a->value)) {
						MOVENODE(struct ACCESS_LIST_LIST *, al, access_list->allow, access_list->deny);
					}
				}
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct ACCESS_LIST_LIST *, (!strcasecmp(list, "allow")) ? access_list->allow : access_list->deny, al, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct ACCESS_LIST_LIST *, (!strcasecmp(list, "allow")) ? access_list->allow : access_list->deny, al, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct ACCESS_LIST_LIST *, (!strcasecmp(list, "allow")) ? access_list->allow : access_list->deny, al, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct ACCESS_LIST_LIST *, (!strcasecmp(list, "allow")) ? access_list->allow : access_list->deny, al, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&access_list->lock);
}

void interface_page_config_rewrite(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct REWRITE_LIST_LIST *rl = NULL;

	if (rewrite_list == NULL)
		return;

	pthread_rwlock_rdlock(&rewrite_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=rewrite&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (rl = rewrite_list->rewrite; rl; rl = rl->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (rl->enabled == TRUE) ? "yes" : "no");

		if (rl->comment != NULL) {
			ptr = string_to_html(rl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->profiles != NULL) {
			ptr = array_merge(rl->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}
		if (rl->mime != NULL) {
			ptr = string_to_html(rl->mime, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Mimetype</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->size != 0) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Size</td><td>%d</td></tr>\n", rl->size);
		}

		if (rl->pattern != NULL) {
			ptr = string_to_html(rl->pattern, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Pattern</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->replace != NULL) {
			ptr = string_to_html(rl->replace, HTML_NEWLINES | HTML_SPACES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Replacement</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->which != 0)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Applies to</td><td>%s,%s,%s,%s</td></tr>\n", (rl->which & REWRITE_SERVER) ? "server header" : "", (rl->which & REWRITE_CLIENT) ? "client header" : "", (rl->which & REWRITE_BODY) ? "body" : "", (rl->which & REWRITE_POST) ? "post" : "");

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=rewrite&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=rewrite&amp;action=delete&amp;id=%d\">Delete</a></td>\n", rl->id, rl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=rewrite&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=rewrite&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", rl->id, rl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=rewrite&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=rewrite&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", rl->id, rl->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&rewrite_list->lock);
}

void interface_page_config_rewrite_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL;
	struct REWRITE_LIST_LIST *rl = NULL;
	struct cgi_args_t *a = args;

	if (rewrite_list == NULL)
		return;

	pthread_rwlock_rdlock(&rewrite_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (rl = rewrite_list->rewrite; rl && rl->id != id; rl = rl->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"rewrite\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (rl == NULL || rl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (rl != NULL && rl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (rl != NULL && rl->comment != NULL)
		ptr = string_to_html(rl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (rl != NULL && rl->profiles != NULL)
		ptr = array_merge(rl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (rl != NULL && rl->mime != NULL)
		ptr = string_to_html(rl->mime, 0);
	filebuf_addf(filebuf, "<tr><td>Mimetype</td><td><input type=\"text\" name=\"mime\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Size</td><td><input type=\"text\" name=\"size\" value=\"%d\" size=\"20\"></td></tr>\n", (rl != NULL) ? rl->size : 0);
	if (rl != NULL && rl->pattern != NULL)
		ptr = string_to_html(rl->pattern, 0);
	filebuf_addf(filebuf, "<tr><td>Pattern</td><td><input type=\"text\" name=\"pattern\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (rl != NULL && rl->replace != NULL)
		ptr = string_to_html(rl->replace, 0);
	filebuf_addf(filebuf, "<tr><td>Replace</td><td><textarea name=\"replace\" cols=\"80\" rows=\"8\" wrap=\"virtual\">%s</textarea></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Applies to</td><td align=\"center\">");
	filebuf_addf(filebuf, "Server header: <input type=\"checkbox\" name=\"server\" %s>   ", (rl != NULL && rl->which & REWRITE_SERVER) ? "checked" : "");
	filebuf_addf(filebuf, "Client header: <input type=\"checkbox\" name=\"client\" %s>   ", (rl != NULL && rl->which & REWRITE_CLIENT) ? "checked" : "");
	filebuf_addf(filebuf, "Body : <input type=\"checkbox\" name=\"body\" %s>   ", (rl == NULL || rl->which & REWRITE_BODY) ? "checked" : "");
	filebuf_addf(filebuf, "POST data : <input type=\"checkbox\" name=\"post\" %s>   ", (rl != NULL && rl->which & REWRITE_POST) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&rewrite_list->lock);
}


void interface_page_config_rewrite_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct REWRITE_LIST_LIST *rl = NULL;
	struct cgi_args_t *a;

	if (rewrite_list == NULL)
		return;

	pthread_rwlock_wrlock(&rewrite_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		for (rl = rewrite_list->rewrite; rl; rl = rl->next) {
			if (rl->id == id) {
				rewrite_list->rewrite = rewrite_list_delete(rl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			rl = rewrite_list_new(rewrite_list->rewrite);
			if (rewrite_list->rewrite == NULL)
				rewrite_list->rewrite = rl;
			rl->id = rewrite_list->id++;
		} else {
			for (rl = rewrite_list->rewrite; rl; rl = rl->next)
				if (rl->id == id)
					break;
			if (rl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			rl->which = 0;
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						rl->enabled = FALSE;
					else
						rl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					rewrite_list_insert(rl, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					rewrite_list_insert(rl, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "pattern"))
					rewrite_list_insert(rl, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "replace"))
					rewrite_list_insert(rl, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "mime"))
					rewrite_list_insert(rl, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "size"))
					rewrite_list_insert(rl, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
				else if (!strcasecmp(a->name, "server"))
					rl->which |= REWRITE_SERVER;
				else if (!strcasecmp(a->name, "client"))
					rl->which |= REWRITE_CLIENT;
				else if (!strcasecmp(a->name, "body"))
					rl->which |= REWRITE_BODY;
				else if (!strcasecmp(a->name, "post"))
					rl->which |= REWRITE_POST;
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct REWRITE_LIST_LIST *, rewrite_list->rewrite, rl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct REWRITE_LIST_LIST *, rewrite_list->rewrite, rl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct REWRITE_LIST_LIST *, rewrite_list->rewrite, rl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct REWRITE_LIST_LIST *, rewrite_list->rewrite, rl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&rewrite_list->lock);
}

void interface_page_config_cache(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct cache_store *sl = NULL;

	pthread_mutex_lock(&global->cache->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"cache\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Violate RFC</td><td align=\"center\">Yes: <input type=\"radio\" name=\"violaterfc\" value=\"yes\" %s> No: <input type=\"radio\" name=\"violaterfc\" value=\"no\" %s></td></tr>\n", (global->cache->violaterfc == TRUE) ? "checked" : "", (global->cache->violaterfc == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Memory cache size</td><td><input type=\"text\" name=\"maxmemsize\" value=\"%u\" size=\"20\"></td>\n", global->cache->maxmemsize);
	filebuf_addf(filebuf, "<tr><td>Memory free extra</td><td><input type=\"text\" name=\"memextra\" value=\"%u\" size=\"20\"></td>\n", global->cache->memextra);
	filebuf_addf(filebuf, "<tr><td>Minimum  age</td><td><input type=\"text\" name=\"minage\" value=\"%d\" size=\"20\"></td>\n", global->cache->minage);
	filebuf_addf(filebuf, "<tr><td>Maximum age</td><td><input type=\"text\" name=\"maxage\" value=\"%d\" size=\"20\"></td>\n", global->cache->maxage);
	filebuf_addf(filebuf, "<tr><td>Revalidate age</td><td><input type=\"text\" name=\"validate\" value=\"%d\" size=\"20\"></td>\n", global->cache->validate);
	filebuf_addf(filebuf, "<tr><td>Last-Modified time factor</td><td><input type=\"text\" name=\"lmtime\" value=\"%d\" size=\"20\"></td>\n", global->cache->lmfactor);
	filebuf_addf(filebuf, "<tr><td>Minimum file size</td><td><input type=\"text\" name=\"minsize\" value=\"%u\" size=\"20\"></td>\n", global->cache->minsize);
	filebuf_addf(filebuf, "<tr><td>Maximum file size</td><td><input type=\"text\" name=\"maxsize\" value=\"%u\" size=\"20\"></td>\n", global->cache->maxsize);
	filebuf_addf(filebuf, "<tr><td>Maximum wait size</td><td><input type=\"text\" name=\"maxwaitsize\" value=\"%u\" size=\"20\"></td>\n", global->cache->maxwaitsize);
	filebuf_addf(filebuf, "<tr><td>Prefetch window</td><td><input type=\"text\" name=\"prefetchwindow\" value=\"%d\" size=\"20\"></td>\n", global->cache->prefetchwindow);
	filebuf_addf(filebuf, "<tr><td>ICP port</td><td><input type=\"text\" name=\"icpport\" value=\"%d\" size=\"20\"></td>\n", global->cache->icpport);
	filebuf_addf(filebuf, "<tr><td>ICP timeout</td><td><input type=\"text\" name=\"icptimeout\" value=\"%d\" size=\"20\"></td>\n", global->cache->icptimeout);
	filebuf_addf(filebuf, "<tr><td>Store balance method</td><td align=\"center\">Fill size: <input type=\"radio\" name=\"sbalancemethod\" value=\"fillsize\" %s> Fill percent: <input type=\"radio\" name=\"sbalancemethod\" value=\"fillpercent\" %s></td></tr>\n", (global->cache->sbalancemethod == SBALANCE_FILLSIZE) ? "checked" : "", (global->cache->sbalancemethod == SBALANCE_FILLPERCENT) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=cache&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (sl = global->cache->stores; sl; sl = sl->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (sl->enabled == TRUE) ? "yes" : "no");

		if (sl->comment != NULL) {
			ptr = string_to_html(sl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (sl->profiles != NULL) {
			ptr = array_merge(sl->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}
		if (sl->path != NULL) {
			ptr = string_to_html(sl->path, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Path</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (sl->maxdisksize != 0) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Maximum disk size</td><td>%u</td></tr>\n", sl->maxdisksize);
		}

		if (sl->diskextra != 0) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Disk free extra</td><td>%u</td></tr>\n", sl->diskextra);
		}

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=cache&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=cache&amp;action=delete&amp;id=%d\">Delete</a></td>\n", sl->id, sl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=cache&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=cache&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", sl->id, sl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=cache&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=cache&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", sl->id, sl->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_mutex_unlock(&global->cache->lock);
}

void interface_page_config_cache_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL;
	struct cache_store *sl = NULL;
	struct cgi_args_t *a = args;

	pthread_mutex_lock(&global->cache->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (sl = global->cache->stores; sl && sl->id != id; sl = sl->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"cache\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (sl == NULL || sl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (sl != NULL && sl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (sl != NULL && sl->comment != NULL)
		ptr = string_to_html(sl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (sl != NULL && sl->profiles != NULL)
		ptr = array_merge(sl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (sl != NULL && sl->path != NULL)
		ptr = string_to_html(sl->path, 0);
	filebuf_addf(filebuf, "<tr><td>Path</td><td><input type=\"text\" name=\"path\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Maximum disk size</td><td><input type=\"text\" name=\"maxdisksize\" value=\"%u\" size=\"20\"></td></tr>\n", (sl != NULL) ? sl->maxdisksize : 0);
	filebuf_addf(filebuf, "<tr><td>Disk free extra</td><td><input type=\"text\" name=\"diskextra\" value=\"%u\" size=\"20\"></td></tr>\n", (sl != NULL) ? sl->diskextra : 0);

	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_mutex_unlock(&global->cache->lock);
}


void interface_page_config_cache_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct cache_store *sl = NULL;
	struct cgi_args_t *a;

	pthread_mutex_lock(&global->cache->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "global")) {
		for (a = args; a; a = a->next) {
			if (!strcasecmp(a->name, "violaterfc")) {
				if (!strcasecmp(a->value, "yes"))
					global->cache->violaterfc = TRUE;
				else
					global->cache->violaterfc = FALSE;
			} else if (!strcasecmp(a->name, "maxmemsize"))
				global_cache_insert(a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "memextra"))
				global_cache_insert(NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "maxage"))
				global_cache_insert(NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "validate"))
				global_cache_insert(NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "lmtime"))
				global_cache_insert(NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "maxsize"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "minage"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "minsize"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "maxwaitsize"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "prefetchwindow"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
			else if (!strcasecmp(a->name, "icpport"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
			else if (!strcasecmp(a->name, "icptimeout"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
			else if (!strcasecmp(a->name, "sbalancemethod"))
				global_cache_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
		}

		/* stores_update only needs to be called when changing the list of cache stores */
		pthread_mutex_unlock(&global->cache->lock);
		return;
	} else if (!strcasecmp(action, "delete")) {
		for (sl = global->cache->stores; sl; sl = sl->next) {
			if (sl->id == id) {
				global->cache->stores = stores_list_delete(sl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			sl = stores_list_new(global->cache->stores);
			if (global->cache->stores == NULL)
				global->cache->stores = sl;
			sl->id = global->cache->id++;
		} else {
			for (sl = global->cache->stores; sl; sl = sl->next)
				if (sl->id == id)
					break;
			if (sl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						sl->enabled = FALSE;
					else
						sl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "comment"))
					stores_list_insert(sl, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "profiles"))
					stores_list_insert(sl, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "path"))
					stores_list_insert(sl, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "maxdisksize"))
					stores_list_insert(sl, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "diskextra"))
					stores_list_insert(sl, NULL, NULL, NULL, NULL, a->value);

			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct cache_store *, global->cache->stores, sl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct cache_store *, global->cache->stores, sl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct cache_store *, global->cache->stores, sl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct cache_store *, global->cache->stores, sl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	stores_update(global->cache);

	pthread_mutex_unlock(&global->cache->lock);
}

void interface_page_config_profiles(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct PROFILES_LIST_LIST *pl = NULL;

	if (profiles_list == NULL)
		return;

	pthread_rwlock_rdlock(&profiles_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=profiles&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (pl = profiles_list->profiles; pl; pl = pl->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (pl->enabled == TRUE) ? "yes" : "no");

		if (pl->comment != NULL) {
			ptr = string_to_html(pl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (pl->profiles != NULL) {
			ptr = array_merge(pl->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (pl->host != NULL) {
			ptr = string_to_html(pl->host, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Host</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (pl->file != NULL) {
			ptr = string_to_html(pl->file, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">File</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (pl->portrange != NULL) {
			ptr = string_to_html(pl->portrange, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Port range</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (pl->addprofiles != NULL) {
			ptr = array_merge(pl->addprofiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Added profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (pl->removeprofiles != NULL) {
			ptr = array_merge(pl->removeprofiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Removed profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (pl->urlcommand != NULL) {
			ptr = array_merge(pl->urlcommand, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">URL Command</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=profiles&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=profiles&amp;action=delete&amp;id=%d\">Delete</a></td>\n", pl->id, pl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=profiles&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=profiles&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", pl->id, pl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=profiles&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=profiles&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", pl->id, pl->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&profiles_list->lock);
}

void interface_page_config_profiles_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL;
	struct PROFILES_LIST_LIST *pl = NULL;
	struct cgi_args_t *a = args;

	if (profiles_list == NULL)
		return;

	pthread_rwlock_rdlock(&profiles_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (pl = profiles_list->profiles; pl && pl->id != id; pl = pl->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"profiles\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (pl == NULL || pl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (pl != NULL && pl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (pl != NULL && pl->comment != NULL)
		ptr = string_to_html(pl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (pl != NULL && pl->profiles != NULL)
		ptr = array_merge(pl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (pl != NULL && pl->host != NULL)
		ptr = string_to_html(pl->host, 0);
	filebuf_addf(filebuf, "<tr><td>Host</td><td><input type=\"text\" name=\"host\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (pl != NULL && pl->file != NULL)
		ptr = string_to_html(pl->file, 0);
	filebuf_addf(filebuf, "<tr><td>File</td><td><input type=\"text\" name=\"file\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (pl != NULL && pl->portrange != NULL)
		ptr = string_to_html(pl->portrange, 0);
	filebuf_addf(filebuf, "<tr><td>Port range</td><td><input type=\"text\" name=\"portrange\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (pl != NULL && pl->addprofiles != NULL)
		ptr = array_merge(pl->addprofiles, ',');
	filebuf_addf(filebuf, "<tr><td>Added profiles</td><td><input type=\"text\" name=\"addprofiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (pl != NULL && pl->removeprofiles != NULL)
		ptr = array_merge(pl->removeprofiles, ',');
	filebuf_addf(filebuf, "<tr><td>Removed profiles</td><td><input type=\"text\" name=\"removeprofiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (pl != NULL && pl->urlcommand != NULL)
		ptr = array_merge(pl->urlcommand, ',');
	filebuf_addf(filebuf, "<tr><td>URL Command</td><td><input type=\"text\" name=\"urlcommand\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&profiles_list->lock);
}


void interface_page_config_profiles_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct PROFILES_LIST_LIST *pl = NULL;
	struct cgi_args_t *a;

	if (profiles_list == NULL)
		return;

	pthread_rwlock_wrlock(&profiles_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		for (pl = profiles_list->profiles; pl; pl = pl->next) {
			if (pl->id == id) {
				profiles_list->profiles = profiles_list_delete(pl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			pl = profiles_list_new(profiles_list->profiles);
			if (profiles_list->profiles == NULL)
				profiles_list->profiles = pl;
			pl->id = profiles_list->id++;
		} else {
			for (pl = profiles_list->profiles; pl; pl = pl->next)
				if (pl->id == id)
					break;
			if (pl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						pl->enabled = FALSE;
					else
						pl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					profiles_list_insert(pl, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					profiles_list_insert(pl, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "host"))
					profiles_list_insert(pl, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "file"))
					profiles_list_insert(pl, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "addprofiles"))
					profiles_list_insert(pl, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "removeprofiles"))
					profiles_list_insert(pl, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "urlcommand"))
					profiles_list_insert(pl, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "portrange"))
					profiles_list_insert(pl, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct PROFILES_LIST_LIST *, profiles_list->profiles, pl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct PROFILES_LIST_LIST *, profiles_list->profiles, pl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct PROFILES_LIST_LIST *, profiles_list->profiles, pl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct PROFILES_LIST_LIST *, profiles_list->profiles, pl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&profiles_list->lock);
}

void interface_page_config_limits(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct LIMITS_LIST_LIST *ll = NULL;

	if (limits_list == NULL)
		return;

	pthread_rwlock_rdlock(&limits_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=limits&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (ll = limits_list->limits_list; ll; ll = ll->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (ll->enabled == TRUE) ? "yes" : "no");

		if (ll->comment != NULL) {
			ptr = string_to_html(ll->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ll->profiles != NULL) {
			ptr = array_merge(ll->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Action</td><td>%s</td></tr>\n", (ll->action == POLICY_ALLOW) ? "allow" : "deny");

		if (ll->template != NULL) {
			ptr = string_to_html(ll->template, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Template</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ll->from != NULL && ll->to != NULL) {
			if (ll->from->tm_mon != 0 && ll->to->tm_mon != 0)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Month range</td><td>%s to %s</td></tr>\n", month_name(ll->from->tm_mon), month_name(ll->to->tm_mon));

			if (ll->from->tm_mday != 0 && ll->to->tm_mday != 0)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Day range</td><td>%d to %d</td></tr>\n", ll->from->tm_mday, ll->to->tm_mday);

			if (ll->from->tm_wday != 0 && ll->to->tm_wday != 0)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Weekday range</td><td>%s to %s</td></tr>\n", weekday_name(ll->from->tm_wday), weekday_name(ll->to->tm_wday));

			if (ll->to->tm_hour != 0)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Hour range</td><td>%d to %d</td></tr>\n", ll->from->tm_hour, ll->to->tm_hour);

			if (ll->to->tm_min != 0)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Minute range</td><td>%d to %d</td></tr>\n", ll->from->tm_min, ll->to->tm_min);
		}

		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Time match mode</td><td>%s</td></tr>\n", (ll->flags & LIMIT_ABSOLUTETIME) ? "absolute" : "all ranges");

		if (ll->maxbytes != 0) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Maximum bytes</td><td>%u</td></tr>\n", ll->maxbytes);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Current bytes</td><td>%u</td></tr>\n", ll->n_bytes);
		}

		if (ll->maxrequests != 0) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Maximum requests</td><td>%u</td></tr>\n", ll->maxrequests);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Current requests</td><td>%u</td></tr>\n", ll->n_requests);
		}

		if (ll->flags & ~(LIMIT_ABSOLUTETIME | LIMIT_ALLRANGETIME))
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Flags</td><td>%s</td></tr>\n", (ll->flags & LIMIT_CACHE) ? "limit cache transfers" : "");

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=limits&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=limits&amp;action=delete&amp;id=%d\">Delete</a></td>\n", ll->id, ll->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=limits&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=limits&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", ll->id, ll->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=limits&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=limits&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", ll->id, ll->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&limits_list->lock);
}

void interface_page_config_limits_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1, i;
	char *ptr = NULL;
	struct LIMITS_LIST_LIST *ll = NULL;
	struct cgi_args_t *a = args;

	if (limits_list == NULL)
		return;

	pthread_rwlock_rdlock(&limits_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (ll = limits_list->limits_list; ll && ll->id != id; ll = ll->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"limits\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (ll == NULL || ll->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (ll != NULL && ll->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (ll != NULL && ll->comment != NULL)
		ptr = string_to_html(ll->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (ll != NULL && ll->profiles != NULL)
		ptr = array_merge(ll->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Action</td><td><select name=\"eaction\">\n");
	filebuf_addf(filebuf, "<option value=\"allow\" %s>allow\n", (ll != NULL && ll->action == POLICY_ALLOW) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"deny\" %s>deny\n", (ll == NULL || ll->action == POLICY_DENY) ? "selected" : "");
	filebuf_addf(filebuf, "</select></tr>\n");

	if (ll != NULL && ll->template != NULL)
		ptr = string_to_html(ll->template, 0);
	filebuf_addf(filebuf, "<tr><td>Template</td><td><input type=\"text\" name=\"template\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Limit months</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"limitmonths\" value=\"yes\" %s> ", (ll != NULL && ll->from != NULL && ll->to != NULL && (ll->from->tm_mon != 0 && ll->to->tm_mon != 0)) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"limitmonths\" value=\"no\" %s> ", (ll == NULL || ll->from == NULL || ll->to == NULL || (ll->from->tm_mon == 0 || ll->to->tm_mon == 0)) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Month range</td><td>");
	filebuf_addf(filebuf, "<select name=\"frommonth\">");
	for (i = 0; calendar_months[i].name; i++) {
		filebuf_addf(filebuf, calendar_months[i].html, (ll != NULL && ll->from != NULL && ll->from->tm_mon == i + 1) ? "selected" : "");
	}
	filebuf_addf(filebuf, "</select>");
	filebuf_addf(filebuf, " to ");
	filebuf_addf(filebuf, "<select name=\"tomonth\">");
	for (i = 0; calendar_months[i].name; i++) {
		filebuf_addf(filebuf, calendar_months[i].html, (ll != NULL && ll->to != NULL && ll->to->tm_mon == i + 1) ? "selected" : "");
	}
	filebuf_addf(filebuf, "</select></td></tr>");
	filebuf_addf(filebuf, "<tr><td>Limit days</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"limitdays\" value=\"yes\" %s> ", (ll != NULL && ll->from != NULL && ll->to != NULL && (ll->from->tm_mday != 0 && ll->to->tm_mday != 0)) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"limitdays\" value=\"no\" %s> ", (ll == NULL || ll->from == NULL || ll->to == NULL || (ll->from->tm_mday == 0 || ll->to->tm_mday == 0)) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Day range</td><td>");
	filebuf_addf(filebuf, "<select name=\"fromday\">");
	for (i = 1; i <= 31; i++) {
		filebuf_addf(filebuf, "<option value=\"%d\" %s>%d\n", i, (ll != NULL && ll->from != NULL && ll->from->tm_mday == i) ? "selected" : "", i);
	}
	filebuf_addf(filebuf, "</select>");
	filebuf_addf(filebuf, " to ");
	filebuf_addf(filebuf, "<select name=\"today\">");
	for (i = 1; i <= 31; i++) {
		filebuf_addf(filebuf, "<option value=\"%d\" %s>%d\n", i, (ll != NULL && ll->to != NULL && ll->to->tm_mday == i) ? "selected" : "", i);
	}
	filebuf_addf(filebuf, "</select></td></tr>");

	filebuf_addf(filebuf, "<tr><td>Limit weekdays</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"limitweekdays\" value=\"yes\" %s> ", (ll != NULL && ll->from != NULL && ll->to != NULL && (ll->from->tm_wday != 0 && ll->to->tm_wday != 0)) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"limitweekdays\" value=\"no\" %s> ", (ll == NULL || ll->from == NULL || ll->to == NULL || (ll->from->tm_wday == 0 || ll->to->tm_wday == 0)) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Weekday range</td><td>");
	filebuf_addf(filebuf, "<select name=\"fromweekday\">");
	for (i = 0; calendar_weekdays[i].name; i++) {
		filebuf_addf(filebuf, calendar_weekdays[i].html, (ll != NULL && ll->from != NULL && ll->from->tm_wday == i + 1) ? "selected" : "");
	}
	filebuf_addf(filebuf, "</select>");
	filebuf_addf(filebuf, " to ");
	filebuf_addf(filebuf, "<select name=\"toweekday\">");
	for (i = 0; calendar_weekdays[i].name; i++) {
		filebuf_addf(filebuf, calendar_weekdays[i].html, (ll != NULL && ll->to != NULL && ll->to->tm_wday == i + 1) ? "selected" : "");
	}
	filebuf_addf(filebuf, "</select></td></tr>");

	/* again, we assume an hour range of 0 to 0 is invalid */
	filebuf_addf(filebuf, "<tr><td>Limit hours</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"limithours\" value=\"yes\" %s> ", (ll != NULL && ll->from != NULL && ll->to != NULL && (ll->from->tm_hour != 0 || ll->to->tm_hour != 0)) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"limithours\" value=\"no\" %s> ", (ll == NULL || ll->from == NULL || ll->to == NULL || (ll->from->tm_hour == 0 && ll->to->tm_hour == 0)) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Hour range</td><td>");
	filebuf_addf(filebuf, "<input type=\"text\" name=\"fromhour\" size=\"2\" value=\"%d\"> to <input type=\"input\" name=\"tohour\" size=\"2\" value=\"%d\"></td></tr>\n", (ll != NULL && ll->from != NULL) ? ll->from->tm_hour : 0, (ll != NULL && ll->to != NULL) ? ll->to->tm_hour : 0);

	filebuf_addf(filebuf, "<tr><td>Limit minutes</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"limitminutes\" value=\"yes\" %s> ", (ll != NULL && ll->from != NULL && ll->to != NULL && (ll->from->tm_min != 0 || ll->to->tm_min != 0)) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"limitminutes\" value=\"no\" %s> ", (ll == NULL || ll->from == NULL || ll->to == NULL || (ll->from->tm_min == 0 && ll->to->tm_min == 0)) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Minute range</td><td>");
	filebuf_addf(filebuf, "<input type=\"text\" name=\"fromminute\" size=\"2\" value=\"%d\"> to <input type=\"input\" name=\"tominute\" size=\"2\" value=\"%d\"></td></tr>\n", (ll != NULL && ll->from != NULL) ? ll->from->tm_min : 0, (ll != NULL && ll->to != NULL) ? ll->to->tm_min : 0);


	filebuf_addf(filebuf, "<tr><td>Time match mode</td><td align=\"center\">Absolute: <input type=\"radio\" name=\"matchmode\" value=\"absolutetime\" %s> All ranges: <input type=\"radio\" name=\"matchmode\" value=\"allrangetime\" %s></td></tr>\n", (ll == NULL || (ll->flags & LIMIT_ABSOLUTETIME)) ? "checked" : "", (ll != NULL && (ll->flags & LIMIT_ALLRANGETIME)) ? "checked" : "");

	filebuf_addf(filebuf, "<tr><td>Byte transfer limit</td><td><input type=\"text\" name=\"maxbytes\" value=\"%u\" size=\"20\">\n", (ll != NULL) ? ll->maxbytes : 0);
	filebuf_addf(filebuf, "<tr><td>Current bytes</td><td><input type=\"text\" name=\"currentbytes\" value=\"%u\" size=\"20\">\n", (ll != NULL) ? ll->n_bytes : 0);

	filebuf_addf(filebuf, "<tr><td>Request limit</td><td><input type=\"text\" name=\"maxrequests\" value=\"%u\" size=\"20\"></td></tr>\n", (ll != NULL) ? ll->maxrequests : 0);
	filebuf_addf(filebuf, "<tr><td>Current requests</td><td><input type=\"text\" name=\"currentrequests\" value=\"%u\" size=\"20\"></td></tr>\n", (ll != NULL) ? ll->n_requests : 0);

	filebuf_addf(filebuf, "<tr><td>Flags</td><td align=\"center\">Limit cache transfers: <input type=\"checkbox\" name=\"limitcache\" %s></td></tr>\n", (ll != NULL && (ll->flags & LIMIT_CACHE)) ? "checked" : "");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&limits_list->lock);
}


void interface_page_config_limits_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1, i;
	int limitmonths = FALSE, limitdays = FALSE, limitweekdays = FALSE, limithours = FALSE, limitminutes = FALSE;
	char *action = NULL;
	struct LIMITS_LIST_LIST *ll = NULL;
	struct cgi_args_t *a;
	struct tm to, from;

	if (limits_list == NULL)
		return;

	memset(&to, 0, sizeof(to));
	memset(&from, 0, sizeof(from));

	pthread_rwlock_wrlock(&limits_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		else if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		else if (!strcasecmp(a->name, "limitmonths") && !strcasecmp(a->value, "yes"))
			limitmonths = TRUE;
		else if (!strcasecmp(a->name, "limitdays") && !strcasecmp(a->value, "yes"))
			limitdays = TRUE;
		else if (!strcasecmp(a->name, "limitweekdays") && !strcasecmp(a->value, "yes"))
			limitweekdays = TRUE;
		else if (!strcasecmp(a->name, "limithours") && !strcasecmp(a->value, "yes"))
			limithours = TRUE;
		else if (!strcasecmp(a->name, "limitminutes") && !strcasecmp(a->value, "yes"))
			limitminutes = TRUE;
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		for (ll = limits_list->limits_list; ll; ll = ll->next) {
			if (ll->id == id) {
				limits_list->limits_list = limits_ll_delete(ll);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			ll = limits_ll_new(limits_list->limits_list);
			if (limits_list->limits_list == NULL)
				limits_list->limits_list = ll;
			ll->id = limits_list->id++;
		} else {
			for (ll = limits_list->limits_list; ll; ll = ll->next)
				if (ll->id == id)
					break;
			if (ll == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			to.tm_mon = from.tm_mon = 0;
			to.tm_wday = from.tm_wday = 0;
			to.tm_mday = from.tm_mday = 0;
			to.tm_min = from.tm_min = 0;
			to.tm_hour = from.tm_hour = 0;

			ll->flags = LIMIT_ABSOLUTETIME;
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						ll->enabled = FALSE;
					else
						ll->enabled = TRUE;
				} else if (!strcasecmp(a->name, "eaction")) {
					if (!strcasecmp(a->value, "allow"))
						ll->action = POLICY_ALLOW;
					else
						ll->action = POLICY_DENY;
				} else if (!strcasecmp(a->name, "profiles"))
					limits_ll_insert(ll, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					limits_ll_insert(ll, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "template"))
					limits_ll_insert(ll, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "maxbytes"))
					limits_ll_insert(ll, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "maxrequests"))
					limits_ll_insert(ll, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "limitcache"))
					ll->flags |= LIMIT_CACHE;
				else if (!strcasecmp(a->name, "matchmode")) {
					ll->flags &= ~(LIMIT_ABSOLUTETIME | LIMIT_ALLRANGETIME);
					if (!strcasecmp(a->value, "absolutetime"))
						ll->flags |= LIMIT_ABSOLUTETIME;
					else
						ll->flags |= LIMIT_ALLRANGETIME;
				} else if (!strcasecmp(a->name, "currentbytes"))
					ll->n_bytes = strtoll(a->value, NULL, 10);
				else if (!strcasecmp(a->name, "currentrequests"))
					ll->n_requests = strtoll(a->value, NULL, 10);

				if (limitmonths == TRUE) {
					if (!strcasecmp(a->name, "frommonth")) {
						for (i = 0; calendar_months[i].name && strcasecmp(calendar_months[i].name, a->value); i++);
						if (calendar_months[i].name)
							from.tm_mon = i + 1;
					} else if (!strcasecmp(a->name, "tomonth")) {
						for (i = 0; calendar_months[i].name && strcasecmp(calendar_months[i].name, a->value); i++);
						if (calendar_months[i].name)
							to.tm_mon = i + 1;
					}
				}

				if (limitdays == TRUE) {
					if (!strcasecmp(a->name, "fromday"))
						from.tm_mday = strtol(a->value, NULL, 10);
					else if (!strcasecmp(a->name, "today"))
						to.tm_mday = strtol(a->value, NULL, 10);
				}

				if (limitweekdays == TRUE) {
					if (!strcasecmp(a->name, "fromweekday")) {
						for (i = 0; calendar_weekdays[i].name && strcasecmp(calendar_weekdays[i].name, a->value); i++);
						if (calendar_weekdays[i].name)
							from.tm_wday = i + 1;
					} else if (!strcasecmp(a->name, "toweekday")) {
						for (i = 0; calendar_weekdays[i].name && strcasecmp(calendar_weekdays[i].name, a->value); i++);
						if (calendar_weekdays[i].name)
							to.tm_wday = i + 1;
					}
				}

				if (limithours == TRUE) {
					if (!strcasecmp(a->name, "fromhour"))
						from.tm_hour = strtol(a->value, NULL, 10);
					else if (!strcasecmp(a->name, "tohour"))
						to.tm_hour = strtol(a->value, NULL, 10);
				}

				if (limitminutes == TRUE) {
					if (!strcasecmp(a->name, "fromminute"))
						from.tm_min = strtol(a->value, NULL, 10);
					else if (!strcasecmp(a->name, "tominute"))
						to.tm_min = strtol(a->value, NULL, 10);
				}
			}

			limits_ll_insert(ll, NULL, NULL, NULL, &from, NULL, NULL, NULL, NULL);
			limits_ll_insert(ll, NULL, NULL, NULL, NULL, &to, NULL, NULL, NULL);
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct LIMITS_LIST_LIST *, limits_list->limits_list, ll, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct LIMITS_LIST_LIST *, limits_list->limits_list, ll, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct LIMITS_LIST_LIST *, limits_list->limits_list, ll, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct LIMITS_LIST_LIST *, limits_list->limits_list, ll, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&limits_list->lock);
}

void interface_page_config_mime(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;
	char *ptr;
	struct MIME_LIST_LIST *ml = NULL;

	if (mime_list == NULL)
		return;

	pthread_rwlock_rdlock(&mime_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"mime\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Policy</td><td><select name=\"policy\"><option value=\"allow\" %s>allow<option value=\"deny\" %s>deny</select></td></tr>", (mime_list->policy == POLICY_ALLOW) ? "selected" : "", (mime_list->policy == POLICY_DENY) ? "selected" : "");
	filebuf_addf(filebuf, "<tr><td>Default template</td><td><input type=\"text\" name=\"dtemplate\" value=\"%s\" size=\"20\"></td></tr>\n", (mime_list->dtemplate != NULL) ? mime_list->dtemplate : "");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=mime&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr> <tr><td><br></td></tr>\n");

	for (i = 0; i < 2; i++) {
		filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"30%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		switch (i) {
		case 0:
			ml = mime_list->allow;
			if (ml != NULL)
				filebuf_addf(filebuf, "<tr><th>ALLOW</th></tr>\n");
			break;
		case 1:
			ml = mime_list->deny;
			if (ml != NULL)
				filebuf_addf(filebuf, "<tr><th>DENY</th></tr>\n");
			break;
		}

		filebuf_addf(filebuf, "</table></td></tr>\n");

		if (ml != NULL)
			filebuf_addf(filebuf, "<tr><td><br></td></tr>\n");
		for (; ml; ml = ml->next) {
			filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
			filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (ml->enabled == TRUE) ? "yes" : "no");

			if (ml->comment != NULL) {
				ptr = string_to_html(ml->comment, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (ml->profiles != NULL) {
				ptr = array_merge(ml->profiles, ',');
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (ml->host != NULL) {
				ptr = string_to_html(ml->host, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Host</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (ml->file != NULL) {
				ptr = string_to_html(ml->file, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">File</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (ml->mime != NULL) {
				ptr = string_to_html(ml->mime, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Mimetype</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (ml->template != NULL)
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Template</td><td>%s</td></tr>\n", ml->template);

			filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
			filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=mime&amp;dialog=show&amp;list=%s&amp;id=%d\">Edit</a> <a href=\"config?section=mime&amp;action=delete&amp;list=%s&amp;id=%d\">Delete</a></td>\n", (i == 0) ? "allow" : "deny", ml->id, (i == 0) ? "allow" : "deny", ml->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=mime&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=mime&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=down\">Down</a></td>\n", (i == 0) ? "allow" : "deny", ml->id, (i == 0) ? "allow" : "deny", ml->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=mime&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=mime&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", (i == 0) ? "allow" : "deny", ml->id, (i == 0) ? "allow" : "deny", ml->id);
			filebuf_addf(filebuf, "</table></td></tr>\n");
			filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
		}
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&mime_list->lock);
}

void interface_page_config_mime_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *ptr = NULL;
	struct MIME_LIST_LIST *ml = NULL;
	struct cgi_args_t *a = args;

	if (mime_list == NULL)
		return;

	pthread_rwlock_rdlock(&mime_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (id != -1 && list != NULL) {
		if (!strcasecmp(list, "allow"))
			for (ml = mime_list->allow; ml && ml->id != id; ml = ml->next);
		else if (!strcasecmp(list, "deny"))
			for (ml = mime_list->deny; ml && ml->id != id; ml = ml->next);
	}

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"mime\">\n");

	if (list != NULL)
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"list\" value=\"%s\">\n", list);
	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (ml == NULL || ml->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (ml != NULL && ml->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (ml != NULL && ml->comment != NULL)
		ptr = string_to_html(ml->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (ml != NULL && ml->profiles != NULL)
		ptr = array_merge(ml->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Action</td><td><select name=\"%s\">\n", (list != NULL) ? "move" : "list");
	filebuf_addf(filebuf, "<option value=\"allow\" %s>allow\n", (list != NULL && !strcasecmp(list, "allow")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"deny\" %s>deny\n", (list != NULL && !strcasecmp(list, "deny")) ? "selected" : "");
	filebuf_addf(filebuf, "</select></tr>\n");

	if (ml != NULL && ml->host != NULL)
		ptr = string_to_html(ml->host, 0);
	filebuf_addf(filebuf, "<tr><td>Host</td><td><input type=\"text\" name=\"host\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (ml != NULL && ml->file != NULL)
		ptr = string_to_html(ml->file, 0);
	filebuf_addf(filebuf, "<tr><td>File</td><td><input type=\"text\" name=\"file\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (ml != NULL && ml->mime != NULL)
		ptr = string_to_html(ml->mime, 0);
	filebuf_addf(filebuf, "<tr><td>Mimetype</td><td><input type=\"text\" name=\"mime\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (ml != NULL && ml->template != NULL)
		ptr = string_to_html(ml->template, 0);
	filebuf_addf(filebuf, "<tr><td>Template</td><td><input type=\"text\" name=\"template\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&mime_list->lock);
}

void interface_page_config_mime_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *action = NULL;
	struct MIME_LIST_LIST *ml = NULL;
	struct cgi_args_t *a;

	if (mime_list == NULL)
		return;

	pthread_rwlock_wrlock(&mime_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (action != NULL && !strcasecmp(action, "global")) {
		for (a = args; a; a = a->next) {
			if (!strcasecmp(a->name, "policy")) {
				if (!strcasecmp(a->value, "allow"))
					mime_list->policy = POLICY_ALLOW;
				else if (!strcasecmp(a->value, "deny"))
					mime_list->policy = POLICY_DENY;
			} else if (!strcasecmp(a->name, "dtemplate")) {
				FREE_AND_NULL(mime_list->dtemplate);
				if (strcmp(a->name, ""))
					mime_list->dtemplate = xstrdup(a->value);
			}
		}
	}

	if (action == NULL || list == NULL)
		goto finish;
	if (!strcasecmp(action, "delete")) {
		if (!strcasecmp(list, "allow"))
			ml = mime_list->allow;
		else if (!strcasecmp(list, "deny"))
			ml = mime_list->deny;
		for (; ml; ml = ml->next) {
			if (ml->id == id) {
				if (!strcasecmp(list, "allow"))
					mime_list->allow = mime_ll_delete(ml);
				else if (!strcasecmp(list, "deny"))
					mime_list->deny = mime_ll_delete(ml);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			if (!strcasecmp(list, "allow")) {
				ml = mime_ll_new(mime_list->allow);
				if (mime_list->allow == NULL)
					mime_list->allow = ml;
			} else if (!strcasecmp(list, "deny")) {
				ml = mime_ll_new(mime_list->deny);
				if (mime_list->deny == NULL)
					mime_list->deny = ml;
			} else
				goto finish;
			ml->id = mime_list->id++;
		} else {
			if (!strcasecmp(list, "allow")) {
				for (ml = mime_list->allow; ml; ml = ml->next)
					if (ml->id == id)
						break;
			} else if (!strcasecmp(list, "deny")) {
				for (ml = mime_list->deny; ml; ml = ml->next)
					if (ml->id == id)
						break;
			}

			if (ml == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						ml->enabled = FALSE;
					else
						ml->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					mime_ll_insert(ml, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					mime_ll_insert(ml, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "template"))
					mime_ll_insert(ml, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "host"))
					mime_ll_insert(ml, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "mime"))
					mime_ll_insert(ml, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "file"))
					mime_ll_insert(ml, NULL, NULL, NULL, NULL, NULL, a->value);
				else if (!strcasecmp(a->name, "move")) {
					if (!strcasecmp(a->value, "allow") && strcasecmp(list, a->value)) {
						MOVENODE(struct MIME_LIST_LIST *, ml, mime_list->deny, mime_list->allow);
					} else if (!strcasecmp(a->value, "deny") && strcasecmp(list, a->value)) {
						MOVENODE(struct MIME_LIST_LIST *, ml, mime_list->allow, mime_list->deny);
					}
				}
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct MIME_LIST_LIST *, (!strcasecmp(list, "allow")) ? mime_list->allow : mime_list->deny, ml, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct MIME_LIST_LIST *, (!strcasecmp(list, "allow")) ? mime_list->allow : mime_list->deny, ml, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct MIME_LIST_LIST *, (!strcasecmp(list, "allow")) ? mime_list->allow : mime_list->deny, ml, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct MIME_LIST_LIST *, (!strcasecmp(list, "allow")) ? mime_list->allow : mime_list->deny, ml, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&mime_list->lock);
}

void interface_page_config_cookies(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;
	char *ptr;
	struct COOKIE_LIST_LIST *cl = NULL;

	if (cookie_list == NULL)
		return;

	pthread_rwlock_rdlock(&cookie_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"cookies\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Policy</td><td><select name=\"policy\"><option value=\"allow\" %s>allow<option value=\"deny\" %s>deny</select></td></tr>", (cookie_list->policy == POLICY_ALLOW) ? "selected" : "", (cookie_list->policy == POLICY_DENY) ? "selected" : "");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=cookies&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr> <tr><td><br></td></tr>\n");

	for (i = 0; i < 2; i++) {
		filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"30%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		switch (i) {
		case 0:
			cl = cookie_list->allow;
			if (cl != NULL)
				filebuf_addf(filebuf, "<tr><th>ALLOW</th></tr>\n");
			break;
		case 1:
			cl = cookie_list->deny;
			if (cl != NULL)
				filebuf_addf(filebuf, "<tr><th>DENY</th></tr>\n");
			break;
		}

		filebuf_addf(filebuf, "</table></td></tr>\n");
		if (cl != NULL)
			filebuf_addf(filebuf, "<tr><td><br></td></tr>\n");
		for (; cl; cl = cl->next) {
			filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
			filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (cl->enabled == TRUE) ? "yes" : "no");

			if (cl->comment != NULL) {
				ptr = string_to_html(cl->comment, HTML_NEWLINES);
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}

			if (cl->profiles != NULL) {
				ptr = array_merge(cl->profiles, ',');
				filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
				xfree(ptr);
			}


			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Direction</td><td>%s</td></tr>\n", (cl->direction == COOKIE_BOTH) ? "ANY" : (cl->direction == COOKIE_IN) ? "IN" : "OUT");
			filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
			filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=cookies&amp;dialog=show&amp;list=%s&amp;id=%d\">Edit</a> <a href=\"config?section=cookies&amp;action=delete&amp;list=%s&amp;id=%d\">Delete</a></td>\n", (i == 0) ? "allow" : "deny", cl->id, (i == 0) ? "allow" : "deny", cl->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=cookies&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=cookies&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=down\">Down</a></td>\n", (i == 0) ? "allow" : "deny", cl->id, (i == 0) ? "allow" : "deny", cl->id);
			filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=cookies&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=cookies&amp;action=shift&amp;list=%s&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", (i == 0) ? "allow" : "deny", cl->id, (i == 0) ? "allow" : "deny", cl->id);
			filebuf_addf(filebuf, "</table></td></tr>\n");
			filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
		}
	}

	pthread_rwlock_unlock(&cookie_list->lock);

	filebuf_addf(filebuf, "</table>\n");
}

void interface_page_config_cookies_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *ptr = NULL;
	struct COOKIE_LIST_LIST *cl = NULL;
	struct cgi_args_t *a = args;

	if (cookie_list == NULL)
		return;

	pthread_rwlock_rdlock(&cookie_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (id != -1 && list != NULL) {
		if (!strcasecmp(list, "allow"))
			for (cl = cookie_list->allow; cl && cl->id != id; cl = cl->next);
		else if (!strcasecmp(list, "deny"))
			for (cl = cookie_list->deny; cl && cl->id != id; cl = cl->next);
	}

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"cookies\">\n");

	if (list != NULL)
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"list\" value=\"%s\">\n", list);
	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (cl == NULL || cl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (cl != NULL && cl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (cl != NULL && cl->comment != NULL)
		ptr = string_to_html(cl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (cl != NULL && cl->profiles != NULL)
		ptr = array_merge(cl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Action</td><td><select name=\"%s\">\n", (list != NULL) ? "move" : "list");
	filebuf_addf(filebuf, "<option value=\"allow\" %s>allow\n", (list != NULL && !strcasecmp(list, "allow")) ? "selected" : "");
	filebuf_addf(filebuf, "<option value=\"deny\" %s>deny\n", (list != NULL && !strcasecmp(list, "deny")) ? "selected" : "");
	filebuf_addf(filebuf, "</select></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Direction</td><td align=\"center\">");
	filebuf_addf(filebuf, "Both: <input type=\"radio\" name=\"direction\" value=\"both\" %s> ", (cl == NULL || cl->direction == COOKIE_BOTH) ? "checked" : "");
	filebuf_addf(filebuf, "In: <input type=\"radio\" name=\"direction\" value=\"in\" %s> ", (cl != NULL && cl->direction == COOKIE_IN) ? "checked" : "");
	filebuf_addf(filebuf, "Out: <input type=\"radio\" name=\"direction\" value=\"out\" %s> ", (cl != NULL && cl->direction == COOKIE_OUT) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&cookie_list->lock);
}


void interface_page_config_cookies_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *list = NULL, *action = NULL;
	struct COOKIE_LIST_LIST *cl = NULL;
	struct cgi_args_t *a;

	if (cookie_list == NULL)
		return;
	pthread_rwlock_wrlock(&cookie_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
		if (!strcasecmp(a->name, "list"))
			list = a->value;
	}

	if (action != NULL && !strcasecmp(action, "global")) {
		for (a = args; a; a = a->next) {
			if (!strcasecmp(a->name, "policy")) {
				if (!strcasecmp(a->value, "allow"))
					cookie_list->policy = POLICY_ALLOW;
				else if (!strcasecmp(a->value, "deny"))
					cookie_list->policy = POLICY_DENY;
			}
		}
	}

	if (action == NULL || list == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		if (!strcasecmp(list, "allow"))
			cl = cookie_list->allow;
		else if (!strcasecmp(list, "deny"))
			cl = cookie_list->deny;
		for (; cl; cl = cl->next) {
			if (cl->id == id) {
				if (!strcasecmp(list, "allow"))
					cookie_list->allow = cookie_ll_delete(cl);
				else if (!strcasecmp(list, "deny"))
					cookie_list->deny = cookie_ll_delete(cl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			if (!strcasecmp(list, "allow")) {
				cl = cookie_ll_new(cookie_list->allow);
				if (cookie_list->allow == NULL)
					cookie_list->allow = cl;
			} else if (!strcasecmp(list, "deny")) {
				cl = cookie_ll_new(cookie_list->deny);
				if (cookie_list->deny == NULL)
					cookie_list->deny = cl;
			} else
				goto finish;
			cl->id = cookie_list->id++;
		} else {
			if (!strcasecmp(list, "allow")) {
				for (cl = cookie_list->allow; cl; cl = cl->next)
					if (cl->id == id)
						break;
			} else if (!strcasecmp(list, "deny")) {
				for (cl = cookie_list->deny; cl; cl = cl->next)
					if (cl->id == id)
						break;
			}

			if (cl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						cl->enabled = FALSE;
					else
						cl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					cookie_ll_insert(cl, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					cookie_ll_insert(cl, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "direction"))
					cookie_ll_insert(cl, NULL, NULL, a->value);
				else if (!strcasecmp(a->name, "move")) {
					if (!strcasecmp(a->value, "allow") && strcasecmp(list, a->value)) {
						MOVENODE(struct COOKIE_LIST_LIST *, cl, cookie_list->deny, cookie_list->allow);
					} else if (!strcasecmp(a->value, "deny") && strcasecmp(list, a->value)) {
						MOVENODE(struct COOKIE_LIST_LIST *, cl, cookie_list->allow, cookie_list->deny);
					}
				}
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct COOKIE_LIST_LIST *, (!strcasecmp(list, "allow")) ? cookie_list->allow : cookie_list->deny, cl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct COOKIE_LIST_LIST *, (!strcasecmp(list, "allow")) ? cookie_list->allow : cookie_list->deny, cl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct COOKIE_LIST_LIST *, (!strcasecmp(list, "allow")) ? cookie_list->allow : cookie_list->deny, cl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct COOKIE_LIST_LIST *, (!strcasecmp(list, "allow")) ? cookie_list->allow : cookie_list->deny, cl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&cookie_list->lock);
}

void interface_page_config_external(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct EXTERNAL_LIST_LIST *el = NULL;

	if (external == NULL)
		return;

	pthread_rwlock_rdlock(&external->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=external&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (el = external->external_list; el; el = el->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (el->enabled == TRUE) ? "yes" : "no");

		if (el->comment != NULL) {
			ptr = string_to_html(el->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (el->profiles != NULL) {
			ptr = array_merge(el->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (el->mime != NULL) {
			ptr = string_to_html(el->mime, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Mimetype</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (el->size != 0) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Size</td><td>%d</td></tr>\n", el->size);
		}

		if (el->newmime != NULL) {
			ptr = string_to_html(el->newmime, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">New mime</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (el->exec != NULL) {
			ptr = string_to_html(el->exec, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Executable</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Type</td><td>%s</td></tr>\n", (el->type == EXTERNAL_PIPE) ? "pipe" : "file");
		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=external&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=external&amp;action=delete&amp;id=%d\">Delete</a></td>\n", el->id, el->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=external&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=external&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", el->id, el->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=external&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=external&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", el->id, el->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&external->lock);
}

void interface_page_config_external_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL;
	struct EXTERNAL_LIST_LIST *el = NULL;
	struct cgi_args_t *a = args;

	if (external == NULL)
		return;

	pthread_rwlock_rdlock(&external->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (el = external->external_list; el && el->id != id; el = el->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"external\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (el == NULL || el->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (el != NULL && el->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (el != NULL && el->comment != NULL)
		ptr = string_to_html(el->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (el != NULL && el->profiles != NULL)
		ptr = array_merge(el->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (el != NULL && el->exec != NULL)
		ptr = string_to_html(el->exec, 0);
	filebuf_addf(filebuf, "<tr><td>Executable</td><td><input type=\"text\" name=\"exec\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (el != NULL && el->mime != NULL)
		ptr = string_to_html(el->mime, 0);
	filebuf_addf(filebuf, "<tr><td>Mimetype</td><td><input type=\"text\" name=\"mime\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Size</td><td><input type=\"text\" name=\"size\" value=\"%d\" size=\"20\"></td></tr>\n", (el != NULL) ? el->size : 0);
	if (el != NULL && el->newmime != NULL)
		ptr = string_to_html(el->newmime, 0);
	filebuf_addf(filebuf, "<tr><td>Newmime</td><td><input type=\"text\" name=\"newmime\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Type</td><td align=\"center\">");
	filebuf_addf(filebuf, "Pipe: <input type=\"radio\" name=\"type\" value=\"pipe\" %s> ", (el == NULL || el->type == EXTERNAL_PIPE) ? "checked" : "");
	filebuf_addf(filebuf, "File: <input type=\"radio\" name=\"type\" value=\"file\" %s> ", (el != NULL && el->type == EXTERNAL_FILE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&external->lock);
}


void interface_page_config_external_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct EXTERNAL_LIST_LIST *el = NULL;
	struct cgi_args_t *a;

	if (external == NULL)
		return;

	pthread_rwlock_wrlock(&external->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		for (el = external->external_list; el; el = el->next) {
			if (el->id == id) {
				external->external_list = external_list_delete(el);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			el = external_list_new(external->external_list);
			if (external->external_list == NULL)
				external->external_list = el;
			el->id = external->id++;
		} else {
			for (el = external->external_list; el; el = el->next)
				if (el->id == id)
					break;
			if (el == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						el->enabled = FALSE;
					else
						el->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					external_list_insert(el, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					external_list_insert(el, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "mime"))
					external_list_insert(el, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "exec"))
					external_list_insert(el, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "newmime"))
					external_list_insert(el, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "type"))
					external_list_insert(el, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "size"))
					external_list_insert(el, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
			}

		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct EXTERNAL_LIST_LIST *, external->external_list, el, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct EXTERNAL_LIST_LIST *, external->external_list, el, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct EXTERNAL_LIST_LIST *, external->external_list, el, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct EXTERNAL_LIST_LIST *, external->external_list, el, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&external->lock);
}

void interface_page_config_templates(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct TEMPLATE_LIST *tl = NULL;

	if (templates == NULL)
		return;

	pthread_rwlock_rdlock(&templates->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"templates\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Path</td><td><input type=\"text\" name=\"path\" value=\"%s\" size=\"20\"></td></tr>\n", (templates->path != NULL) ? templates->path : "");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=templates&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr> <tr><td><br></td></tr>\n");

	for (tl = templates->template_list; tl; tl = tl->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (tl->enabled == TRUE) ? "yes" : "no");

		if (tl->comment != NULL) {
			ptr = string_to_html(tl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (tl->profiles != NULL) {
			ptr = array_merge(tl->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (tl->name != NULL) {
			ptr = string_to_html(tl->name, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Name</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (tl->file != NULL) {
			ptr = string_to_html(tl->file, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">File</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (tl->mime != NULL) {
			ptr = string_to_html(tl->mime, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Mimetype</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (tl->code != -1)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Response code</td><td>%d</td></tr>\n", tl->code);

		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Type</td><td>%s</td></tr>\n", (tl->type == TEMPLATE_FILE) ? "file" : "executable");
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Parsable</td><td>%s</td></tr>\n", (tl->parsable == TRUE) ? "true" : "false");

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=templates&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=templates&amp;action=delete&amp;id=%d\">Delete</a></td>\n", tl->id, tl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=templates&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=templates&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", tl->id, tl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=templates&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=templates&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", tl->id, tl->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&templates->lock);
}

void interface_page_config_templates_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL, buf[16];
	struct TEMPLATE_LIST *tl = NULL;
	struct cgi_args_t *a = args;

	if (profiles_list == NULL)
		return;

	pthread_rwlock_rdlock(&templates->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (tl = templates->template_list; tl && tl->id != id; tl = tl->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"templates\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (tl == NULL || tl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (tl != NULL && tl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (tl != NULL && tl->comment != NULL)
		ptr = string_to_html(tl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (tl != NULL && tl->profiles != NULL)
		ptr = array_merge(tl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (tl != NULL && tl->name != NULL)
		ptr = string_to_html(tl->name, 0);
	filebuf_addf(filebuf, "<tr><td>Name</td><td><input type=\"text\" name=\"name\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (tl != NULL && tl->file != NULL)
		ptr = string_to_html(tl->file, 0);
	filebuf_addf(filebuf, "<tr><td>File</td><td><input type=\"text\" name=\"file\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (tl != NULL && tl->mime != NULL)
		ptr = string_to_html(tl->mime, 0);
	filebuf_addf(filebuf, "<tr><td>Mimetype</td><td><input type=\"text\" name=\"mime\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (tl != NULL)
		snprintf(buf, sizeof(buf), "%d", tl->code);

	filebuf_addf(filebuf, "<tr><td>Response code</td><td><input type=\"text\" name=\"code\" value=\"%s\"></td></tr>\n", (tl != NULL && tl->code != -1) ? buf : "");
	filebuf_addf(filebuf, "<tr><td>Type</td><td align=\"center\">");
	filebuf_addf(filebuf, "File: <input type=\"radio\" name=\"type\" value=\"file\" %s> ", (tl == NULL || tl->type == TEMPLATE_FILE) ? "checked" : "");
	filebuf_addf(filebuf, "Executable: <input type=\"radio\" name=\"type\" value=\"executable\" %s> ", (tl != NULL && tl->type == TEMPLATE_EXECUTABLE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Parsable</td><td align=\"center\">Yes: <input type=\"radio\" name=\"parsable\" value=\"yes\" %s> No: <input type=\"radio\" name=\"parsable\" value=\"no\" %s></td></tr>\n", (tl != NULL && tl->parsable == TRUE) ? "checked" : "", (tl == NULL || tl->parsable == FALSE) ? "checked" : "");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&templates->lock);
}

void interface_page_config_templates_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct TEMPLATE_LIST *tl = NULL;
	struct cgi_args_t *a;

	if (templates == NULL)
		return;

	pthread_rwlock_wrlock(&templates->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		for (tl = templates->template_list; tl; tl = tl->next) {
			if (tl->id == id) {
				templates->template_list = template_list_delete(tl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			tl = template_list_new(templates->template_list);
			if (templates->template_list == NULL)
				templates->template_list = tl;
			tl->id = templates->id++;
		} else {
			for (tl = templates->template_list; tl; tl = tl->next)
				if (tl->id == id)
					break;
			if (tl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						tl->enabled = FALSE;
					else
						tl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					template_list_insert(tl, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					template_list_insert(tl, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "name"))
					template_list_insert(tl, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "file"))
					template_list_insert(tl, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "type"))
					template_list_insert(tl, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "mime"))
					template_list_insert(tl, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "code"))
					template_list_insert(tl, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
				else if (!strcasecmp(a->name, "parsable")) {
					if (!strcasecmp(a->value, "yes"))
						tl->parsable = TRUE;
					else
						tl->parsable = FALSE;
				}
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct TEMPLATE_LIST *, templates->template_list, tl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct TEMPLATE_LIST *, templates->template_list, tl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct TEMPLATE_LIST *, templates->template_list, tl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct TEMPLATE_LIST *, templates->template_list, tl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&templates->lock);
}

void interface_page_config_network(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct LISTEN_LIST *ll = NULL;

	if (network == NULL)
		return;
	pthread_rwlock_rdlock(&network->lock);
	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=network&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (ll = network->listen_list; ll; ll = ll->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (ll->enabled == TRUE) ? "yes" : "no");

		if (ll->comment != NULL) {
			ptr = string_to_html(ll->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ll->ip != NULL) {
			ptr = string_to_html(ll->ip, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">IP</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ll->port != -1)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Port</td><td>%d</td></tr>\n", ll->port);

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=network&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=network&amp;action=delete&amp;id=%d\">Delete</a></td>\n", ll->id, ll->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=network&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=network&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", ll->id, ll->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=network&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=network&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", ll->id, ll->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&network->lock);
}

void interface_page_config_network_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL, buf[16];
	struct LISTEN_LIST *ll = NULL;
	struct cgi_args_t *a = args;

	if (network == NULL)
		return;

	pthread_rwlock_rdlock(&network->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (ll = network->listen_list; ll && ll->id != id; ll = ll->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"network\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");
	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (ll == NULL || ll->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (ll != NULL && ll->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (ll != NULL && ll->comment != NULL)
		ptr = string_to_html(ll->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (ll != NULL && ll->ip != NULL)
		ptr = string_to_html(ll->ip, 0);
	filebuf_addf(filebuf, "<tr><td>IP</td><td><input type=\"text\" name=\"ip\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (ll != NULL)
		snprintf(buf, sizeof(buf), "%d", ll->port);

	filebuf_addf(filebuf, "<tr><td>Port</td><td><input type=\"text\" name=\"port\" value=\"%s\"></td></tr>\n", (ll != NULL && ll->port != -1) ? buf : "");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");

	filebuf_addf(filebuf, "</table></td></tr>\n");
	pthread_rwlock_unlock(&network->lock);
}

void interface_page_config_network_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct LISTEN_LIST *ll = NULL;
	struct cgi_args_t *a;

	if (network == NULL)
		return;

	pthread_rwlock_wrlock(&network->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "delete")) {
		for (ll = network->listen_list; ll; ll = ll->next) {
			if (ll->id == id) {
				network->listen_list = listen_list_delete(ll);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			ll = listen_list_new(network->listen_list);
			if (network->listen_list == NULL)
				network->listen_list = ll;
			ll->id = network->id++;
		} else {
			for (ll = network->listen_list; ll; ll = ll->next)
				if (ll->id == id)
					break;
			if (ll == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						ll->enabled = FALSE;
					else
						ll->enabled = TRUE;
				} else if (!strcasecmp(a->name, "comment"))
					listen_list_insert(ll, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "ip"))
					listen_list_insert(ll, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "port"))
					listen_list_insert(ll, NULL, NULL, a->value);
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct LISTEN_LIST *, network->listen_list, ll, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct LISTEN_LIST *, network->listen_list, ll, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct LISTEN_LIST *, network->listen_list, ll, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct LISTEN_LIST *, network->listen_list, ll, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&network->lock);
}

void interface_page_config_redirect(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct REDIRECT_LIST_LIST *rl = NULL;

	if (redirect_list == NULL)
		return;

	pthread_rwlock_rdlock(&redirect_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=redirect&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (rl = redirect_list->redirect_list; rl; rl = rl->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (rl->enabled == TRUE) ? "yes" : "no");

		if (rl->comment != NULL) {
			ptr = string_to_html(rl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->profiles != NULL) {
			ptr = array_merge(rl->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->url != NULL) {
			ptr = string_to_html(rl->url, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">URL</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->redirect != NULL) {
			ptr = string_to_html(rl->redirect, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Redirect</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (rl->port != -1)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Port</td><td>%d</td></tr>\n", rl->port);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">302 redirect</td><td>%s</td></tr>\n", (rl->send302 == TRUE) ? "yes" : "no");

		if (rl->options != 0)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Options</td><td>%s,%s,%s</td></tr>\n", (rl->options & URL_ENCODE) ? "encode URL" : "", (rl->options & URL_DECODEBEFORE) ? "decode URL before" : "", (rl->options & URL_DECODEAFTER) ? "decode URL after" : "");

		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Applies to</td><td>%s</td></tr>\n", (rl->which == REDIRECT_BOTH) ? "both" : (rl->which == REDIRECT_URL) ? "URL" : "location");
		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=redirect&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=redirect&amp;action=delete&amp;id=%d\">Delete</a></td>\n", rl->id, rl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=redirect&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=redirect&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", rl->id, rl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=redirect&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=redirect&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", rl->id, rl->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&redirect_list->lock);
}

void interface_page_config_redirect_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL;
	struct REDIRECT_LIST_LIST *rl = NULL;
	struct cgi_args_t *a = args;

	if (redirect_list == NULL)
		return;
	pthread_rwlock_rdlock(&redirect_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (rl = redirect_list->redirect_list; rl && rl->id != id; rl = rl->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"redirect\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (rl == NULL || rl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (rl != NULL && rl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (rl != NULL && rl->comment != NULL)
		ptr = string_to_html(rl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (rl != NULL && rl->profiles != NULL)
		ptr = array_merge(rl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (rl != NULL && rl->url != NULL)
		ptr = string_to_html(rl->url, 0);
	filebuf_addf(filebuf, "<tr><td>URL</td><td><input type=\"text\" name=\"url\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (rl != NULL && rl->redirect != NULL)
		ptr = string_to_html(rl->redirect, 0);
	filebuf_addf(filebuf, "<tr><td>Redirect</td><td><input type=\"text\" name=\"redirect\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (rl != NULL && rl->port != -1)
		filebuf_addf(filebuf, "<tr><td>Port</td><td><input type=\"text\" name=\"port\" value=\"%d\" size=\"20\"></td></tr>\n", rl->port);
	else
		filebuf_addf(filebuf, "<tr><td>Port</td><td><input type=\"text\" name=\"port\" value=\"\" size\"20\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td>302 redirect</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"send302\" value=\"yes\" %s> ", (rl == NULL || rl->send302 == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"send302\" value=\"no\" %s> ", (rl != NULL && rl->send302 == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Options</td><td align=\"center\">");
	filebuf_addf(filebuf, "Encode URL: <input type=\"checkbox\" name=\"encode\" %s>   ", (rl != NULL && (rl->options & URL_ENCODE)) ? "checked" : "");
	filebuf_addf(filebuf, "Decode URL before: <input type=\"checkbox\" name=\"decodebefore\" %s>   ", (rl != NULL && (rl->options & URL_DECODEBEFORE)) ? "checked" : "");
	filebuf_addf(filebuf, "Decode URL after: <input type=\"checkbox\" name=\"decodeafter\" %s>", (rl != NULL && (rl->options & URL_DECODEAFTER)) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Applies to</td><td align=\"center\">");
	filebuf_addf(filebuf, "Both: <input type=\"radio\" name=\"which\" value=\"both\" %s> ", (rl != NULL && rl->which == REDIRECT_BOTH) ? "checked" : "");
	filebuf_addf(filebuf, "URL: <input type=\"radio\" name=\"which\" value=\"url\" %s> ", (rl == NULL || rl->which == REDIRECT_URL) ? "checked" : "");
	filebuf_addf(filebuf, "Location header: <input type=\"radio\" name=\"which\" value=\"location\" %s>   ", (rl != NULL && rl->which == REDIRECT_LOCATION) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&redirect_list->lock);
}


void interface_page_config_redirect_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct REDIRECT_LIST_LIST *rl = NULL;
	struct cgi_args_t *a;

	if (redirect_list == NULL)
		return;
	pthread_rwlock_wrlock(&redirect_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;
	if (!strcasecmp(action, "delete")) {
		for (rl = redirect_list->redirect_list; rl; rl = rl->next) {
			if (rl->id == id) {
				redirect_list->redirect_list = redirect_list_delete(rl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			rl = redirect_list_new(redirect_list->redirect_list);
			if (redirect_list->redirect_list == NULL)
				redirect_list->redirect_list = rl;
			rl->id = redirect_list->id++;
		} else {
			for (rl = redirect_list->redirect_list; rl; rl = rl->next)
				if (rl->id == id)
					break;
			if (rl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			rl->options = 0;
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						rl->enabled = FALSE;
					else
						rl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					redirect_list_insert(rl, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					redirect_list_insert(rl, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "url"))
					redirect_list_insert(rl, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "redirect"))
					redirect_list_insert(rl, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "port"))
					redirect_list_insert(rl, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "which"))
					redirect_list_insert(rl, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "send302"))
					redirect_list_insert(rl, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "encode"))
					rl->options |= URL_ENCODE;
				else if (!strcasecmp(a->name, "decodebefore"))
					rl->options |= URL_DECODEBEFORE;
				else if (!strcasecmp(a->name, "decodeafter"))
					rl->options |= URL_DECODEAFTER;
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct REDIRECT_LIST_LIST *, redirect_list->redirect_list, rl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct REDIRECT_LIST_LIST *, redirect_list->redirect_list, rl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct REDIRECT_LIST_LIST *, redirect_list->redirect_list, rl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct REDIRECT_LIST_LIST *, redirect_list->redirect_list, rl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&redirect_list->lock);
}
void interface_page_config_forward(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct FORWARD_LIST_LIST *fl = NULL;
	if (forward_list == NULL)
		return;
	pthread_rwlock_rdlock(&forward_list->lock);
	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=forward&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");

	for (fl = forward_list->forward_list; fl; fl = fl->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (fl->enabled == TRUE) ? "yes" : "no");

		if (fl->comment != NULL) {
			ptr = string_to_html(fl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->profiles != NULL) {
			ptr = array_merge(fl->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->proxy != NULL) {
			ptr = string_to_html(fl->proxy, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Proxy</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->username != NULL) {
			ptr = string_to_html(fl->username, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Username</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->password != NULL) {
			ptr = string_to_html(fl->password, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Password</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->domain != NULL) {
			ptr = string_to_html(fl->domain, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Domain</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->port != -1)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Port</td><td>%d</td></tr>\n", fl->port);

		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Type</td><td>%s</td></tr>\n", (fl->type == PROXY_NORMAL) ? "HTTP" : (fl->type == PROXY_CONNECT) ? "CONNECT" : "SOCKS4");

		if (fl->icptype != ICP_NONE) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">ICP peer type</td><td>%s</td></tr>\n", (fl->icptype == ICP_PARENT) ? "Parent" : "Sibling");
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">ICP port</td><td>%d</td></tr>\n", fl->icpport);
		}

		if (fl->which != 0)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Applies to</td><td>%s,%s,%s</td></tr>\n", (fl->which & FORWARD_HTTP) ? "HTTP requests" : "", (fl->which & FORWARD_FTP) ? "FTP requests" : "", (fl->which & FORWARD_CONNECT) ? "CONNECT requests" : "");

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=forward&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=forward&amp;action=delete&amp;id=%d\">Delete</a></td>\n", fl->id, fl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=forward&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=forward&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", fl->id, fl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=forward&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=forward&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", fl->id, fl->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&forward_list->lock);
}

void interface_page_config_forward_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL;
	struct FORWARD_LIST_LIST *fl = NULL;
	struct cgi_args_t *a = args;

	if (forward_list == NULL)
		return;
	pthread_rwlock_rdlock(&forward_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (fl = forward_list->forward_list; fl && fl->id != id; fl = fl->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"forward\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (fl == NULL || fl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (fl != NULL && fl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (fl != NULL && fl->comment != NULL)
		ptr = string_to_html(fl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->profiles != NULL)
		ptr = array_merge(fl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->proxy != NULL)
		ptr = string_to_html(fl->proxy, 0);
	filebuf_addf(filebuf, "<tr><td>Proxy</td><td><input type=\"text\" name=\"proxy\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->username != NULL)
		ptr = string_to_html(fl->username, 0);
	filebuf_addf(filebuf, "<tr><td>Username</td><td><input type=\"text\" name=\"username\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->password != NULL)
		ptr = string_to_html(fl->password, 0);
	filebuf_addf(filebuf, "<tr><td>Password</td><td><input type=\"text\" name=\"password\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->domain != NULL)
		ptr = string_to_html(fl->domain, 0);
	filebuf_addf(filebuf, "<tr><td>Domain</td><td><input type=\"text\" name=\"domain\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (fl != NULL && fl->port != -1)
		filebuf_addf(filebuf, "<tr><td>Port</td><td><input type=\"text\" name=\"port\" value=\"%d\" size=\"20\"></td></tr>\n", fl->port);
	else
		filebuf_addf(filebuf, "<tr><td>Port</td><td><input type=\"text\" name=\"port\" value=\"\" size\"20\"></td></tr>\n");


	filebuf_addf(filebuf, "<tr><td>ICP peer type</td><td align=\"center\">");
	filebuf_addf(filebuf, "None: <input type=\"radio\" name=\"icptype\" value=\"none\" %s> ", (fl == NULL || fl->icptype == ICP_NONE) ? "checked" : "");
	filebuf_addf(filebuf, "Parent: <input type=\"radio\" name=\"icptype\" value=\"parent\" %s> ", (fl != NULL && fl->icptype == ICP_PARENT) ? "checked" : "");
	filebuf_addf(filebuf, "Sibling: <input type=\"radio\" name=\"icptype\" value=\"sibling\" %s> ", (fl != NULL && fl->icptype == ICP_SIBLING) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (fl != NULL && fl->icpport != -1)
		filebuf_addf(filebuf, "<tr><td>ICP port</td><td><input type=\"text\" name=\"icpport\" value=\"%d\" size=\"20\"></td></tr>\n", fl->icpport);
	else
		filebuf_addf(filebuf, "<tr><td>ICP port</td><td><input type=\"text\" name=\"icpport\" value=\"\" size\"20\"></td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Type</td><td align=\"center\">");
	filebuf_addf(filebuf, "HTTP: <input type=\"radio\" name=\"type\" value=\"HTTP\" %s> ", (fl == NULL || fl->type == PROXY_NORMAL) ? "checked" : "");
	filebuf_addf(filebuf, "SOCKS4: <input type=\"radio\" name=\"type\" value=\"SOCKS4\" %s> ", (fl != NULL && fl->type == PROXY_SOCKS4) ? "checked" : "");
	filebuf_addf(filebuf, "CONNECT: <input type=\"radio\" name=\"type\" value=\"CONNECT\" %s> ", (fl != NULL && fl->type == PROXY_CONNECT) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td>Applies to</td><td align=\"center\">");
	filebuf_addf(filebuf, "HTTP requests: <input type=\"checkbox\" name=\"HTTP\" %s>   ", (fl == NULL || (fl->which & FORWARD_HTTP)) ? "checked" : "");
	filebuf_addf(filebuf, "FTP requests: <input type=\"checkbox\" name=\"FTP\" %s>   ", (fl == NULL || (fl->which & FORWARD_FTP)) ? "checked" : "");
	filebuf_addf(filebuf, "CONNECT requests: <input type=\"checkbox\" name=\"CONNECT\" %s>   ", (fl != NULL && (fl->which & FORWARD_CONNECT)) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&forward_list->lock);
}


void interface_page_config_forward_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct FORWARD_LIST_LIST *fl = NULL;
	struct cgi_args_t *a;

	if (forward_list == NULL)
		return;

	pthread_rwlock_wrlock(&forward_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;
	if (!strcasecmp(action, "delete")) {
		for (fl = forward_list->forward_list; fl; fl = fl->next) {
			if (fl->id == id) {
				forward_list->forward_list = forward_list_delete(fl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			fl = forward_list_new(forward_list->forward_list);
			if (forward_list->forward_list == NULL)
				forward_list->forward_list = fl;
			fl->id = forward_list->id++;
		} else {
			for (fl = forward_list->forward_list; fl; fl = fl->next)
				if (fl->id == id)
					break;
			if (fl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			fl->which = 0;
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						fl->enabled = FALSE;
					else
						fl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					forward_list_insert(fl, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					forward_list_insert(fl, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "proxy"))
					forward_list_insert(fl, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "port"))
					forward_list_insert(fl, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "type"))
					forward_list_insert(fl, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "username"))
					forward_list_insert(fl, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "password"))
					forward_list_insert(fl, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "domain"))
					forward_list_insert(fl, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "icptype"))
					forward_list_insert(fl, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "icpport"))
					forward_list_insert(fl, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
				else if (!strcasecmp(a->name, "HTTP"))
					fl->which |= FORWARD_HTTP;
				else if (!strcasecmp(a->name, "FTP"))
					fl->which |= FORWARD_FTP;
				else if (!strcasecmp(a->name, "CONNECT"))
					fl->which |= FORWARD_CONNECT;
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct FORWARD_LIST_LIST *, forward_list->forward_list, fl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct FORWARD_LIST_LIST *, forward_list->forward_list, fl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct FORWARD_LIST_LIST *, forward_list->forward_list, fl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct FORWARD_LIST_LIST *, forward_list->forward_list, fl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&forward_list->lock);
}

void interface_page_config_keywords(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr;
	struct KEYWORD_LIST_LIST *kl = NULL;

	if (keyword_list == NULL)
		return;

	pthread_rwlock_rdlock(&keyword_list->lock);

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"60%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"keywords\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"global\">\n");
	filebuf_addf(filebuf, "<tr><td>Template</td><td><input type=\"text\" name=\"template\" value=\"%s\" size=\"20\"></td></tr>\n", (keyword_list->template != NULL) ? keyword_list->template : "");
	filebuf_addf(filebuf, "<tr><td>Threshold</td><td><input type=\"text\" name=\"threshold\" value=\"%d\" size=\"20\"></td></tr>\n", keyword_list->threshold);
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><a href=\"config?section=keywords&amp;dialog=show\">Add</a></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr> <tr><td><br></td></tr>\n");

	for (kl = keyword_list->keyword_list; kl; kl = kl->next) {
		filebuf_addf(filebuf, "<tr><td align=\"center\">\n");
		filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Enabled</td><td>%s</td></tr>\n", (kl->enabled == TRUE) ? "yes" : "no");

		if (kl->comment != NULL) {
			ptr = string_to_html(kl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (kl->profiles != NULL) {
			ptr = array_merge(kl->profiles, ',');
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Profiles</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (kl->mime != NULL) {
			ptr = string_to_html(kl->mime, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Mimetype</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (kl->size != 0) {
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Size</td><td>%d</td></tr>\n", kl->size);
		}

		if (kl->keyword != NULL) {
			ptr = string_to_html(kl->keyword, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Keyword</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		filebuf_addf(filebuf, "<tr><td width=\"10%%\">Score</td><td>%d</td></tr>\n", kl->score);
		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"config?section=keywords&amp;dialog=show&amp;id=%d\">Edit</a> <a href=\"config?section=keywords&amp;action=delete&amp;id=%d\">Delete</a></td>\n", kl->id, kl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=keywords&amp;action=shift&amp;id=%d&amp;direction=up\">Up</a> <a href=\"config?section=keywords&amp;action=shift&amp;id=%d&amp;direction=down\">Down</a></td>\n", kl->id, kl->id);
		filebuf_addf(filebuf, "<td align=\"center\"><a href=\"config?section=keywords&amp;action=shift&amp;id=%d&amp;direction=top\">Top</a> <a href=\"config?section=keywords&amp;action=shift&amp;id=%d&amp;direction=bottom\">Bottom</a></td></tr>\n", kl->id, kl->id);
		filebuf_addf(filebuf, "</table></td></tr>\n");
		filebuf_addf(filebuf, "</table></td></tr> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>\n");

	pthread_rwlock_unlock(&keyword_list->lock);
}

void interface_page_config_keywords_dialog(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *ptr = NULL;
	struct KEYWORD_LIST_LIST *kl = NULL;
	struct cgi_args_t *a = args;

	if (keyword_list == NULL)
		return;

	pthread_rwlock_rdlock(&keyword_list->lock);

	for (; a; a = a->next) {
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (id != -1)
		for (kl = keyword_list->keyword_list; kl && kl->id != id; kl = kl->next);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"keywords\">\n");

	if (id != -1) {
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"edit\">\n");
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"id\" value=\"%d\">\n", id);
	} else
		filebuf_addf(filebuf, "<input type=\"hidden\" name=\"action\" value=\"add\">\n");

	filebuf_addf(filebuf, "<tr><td>Enabled</td><td align=\"center\">");
	filebuf_addf(filebuf, "Yes: <input type=\"radio\" name=\"enabled\" value=\"yes\" %s> ", (kl == NULL || kl->enabled == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "No: <input type=\"radio\" name=\"enabled\" value=\"no\" %s> ", (kl != NULL && kl->enabled == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "</td></tr>\n");

	if (kl != NULL && kl->comment != NULL)
		ptr = string_to_html(kl->comment, 0);
	filebuf_addf(filebuf, "<tr><td>Comment</td><td><input type=\"text\" name=\"comment\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (kl != NULL && kl->profiles != NULL)
		ptr = array_merge(kl->profiles, ',');
	filebuf_addf(filebuf, "<tr><td>Profiles</td><td><input type=\"text\" name=\"profiles\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (kl != NULL && kl->mime != NULL)
		ptr = string_to_html(kl->mime, 0);
	filebuf_addf(filebuf, "<tr><td>Mimetype</td><td><input type=\"text\" name=\"mime\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Size</td><td><input type=\"text\" name=\"size\" value=\"%d\" size=\"20\"></td></tr>\n", (kl != NULL) ? kl->size : 0);

	if (kl != NULL && kl->keyword != NULL)
		ptr = string_to_html(kl->keyword, 0);
	filebuf_addf(filebuf, "<tr><td>Keyword</td><td><input type=\"text\" name=\"keyword\" value=\"%s\" size=\"80\"></td></tr>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Score</td><td><input type=\"text\" name=\"score\" value=\"%d\" size=\"20\"></td></tr>\n", (kl != NULL) ? kl->score : 0);

	filebuf_addf(filebuf, "<tr><td colspan=\"2\"><br></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</table></td></tr>\n");

	pthread_rwlock_unlock(&keyword_list->lock);
}


void interface_page_config_keywords_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int id = -1;
	char *action = NULL;
	struct KEYWORD_LIST_LIST *kl = NULL;
	struct cgi_args_t *a;

	if (keyword_list == NULL)
		return;

	pthread_rwlock_wrlock(&keyword_list->lock);

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "action"))
			action = a->value;
		if (!strcasecmp(a->name, "id"))
			id = atoi(a->value);
	}

	if (action == NULL)
		goto finish;

	if (!strcasecmp(action, "global")) {
		for (a = args; a; a = a->next) {
			if (!strcasecmp(a->name, "template")) {
				FREE_AND_NULL(keyword_list->template);
				if (strcmp(a->value, ""))
					keyword_list->template = xstrdup(a->value);
			} else if (!strcasecmp(a->name, "threshold"))
				keyword_list->threshold = atoi(a->value);
		}
	}

	if (!strcasecmp(action, "delete")) {
		for (kl = keyword_list->keyword_list; kl; kl = kl->next) {
			if (kl->id == id) {
				keyword_list->keyword_list = keyword_list_delete(kl);
				break;
			}
		}
	} else if (!strcasecmp(action, "add") || !strcasecmp(action, "edit") || !strcasecmp(action, "shift")) {
		if (!strcasecmp(action, "add")) {
			kl = keyword_list_new(keyword_list->keyword_list);
			if (keyword_list->keyword_list == NULL)
				keyword_list->keyword_list = kl;
			kl->id = keyword_list->id++;
		} else {
			for (kl = keyword_list->keyword_list; kl; kl = kl->next)
				if (kl->id == id)
					break;
			if (kl == NULL)
				goto finish;
		}

		if (strcasecmp(action, "shift")) {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "enabled")) {
					if (!strcasecmp(a->value, "no"))
						kl->enabled = FALSE;
					else
						kl->enabled = TRUE;
				} else if (!strcasecmp(a->name, "profiles"))
					keyword_list_insert(kl, a->value, NULL, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "comment"))
					keyword_list_insert(kl, NULL, a->value, NULL, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "mime"))
					keyword_list_insert(kl, NULL, NULL, a->value, NULL, NULL, NULL);
				else if (!strcasecmp(a->name, "keyword"))
					keyword_list_insert(kl, NULL, NULL, NULL, a->value, NULL, NULL);
				else if (!strcasecmp(a->name, "score"))
					keyword_list_insert(kl, NULL, NULL, NULL, NULL, a->value, NULL);
				else if (!strcasecmp(a->name, "size"))
					keyword_list_insert(kl, NULL, NULL, NULL, NULL, NULL, a->value);
			}
		} else {
			for (a = args; a; a = a->next) {
				if (!strcasecmp(a->name, "direction")) {
					if (!strcasecmp(a->value, "up")) {
						SHIFTNODE(struct KEYWORD_LIST_LIST *, keyword_list->keyword_list, kl, UP);
					} else if (!strcasecmp(a->value, "down")) {
						SHIFTNODE(struct KEYWORD_LIST_LIST *, keyword_list->keyword_list, kl, DOWN);
					} else if (!strcasecmp(a->value, "top")) {
						SETNODE(struct KEYWORD_LIST_LIST *, keyword_list->keyword_list, kl, TOP);
					} else if (!strcasecmp(a->value, "bottom")) {
						SETNODE(struct KEYWORD_LIST_LIST *, keyword_list->keyword_list, kl, BOTTOM);
					}
				}
			}
		}
	}

      finish:

	pthread_rwlock_unlock(&keyword_list->lock);
}

void filter_check_show(CONNECTION * connection)
{
	char *ptr;
	struct FILTER_LIST_LIST *fl;
	FILEBUF *filebuf;
	HEADER *header;

	filebuf = filebuf_new();

	fl = filter_check(filter_list, connection);

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>URL filter matche for %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_BG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><th align=\"center\" colspan=\"2\">URL filter match for %s%s</th></tr>\n", connection->header->host, connection->header->file);

	if (fl == NULL)
		filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\">None</td></tr>\n");
	else {
		if (fl->comment != NULL) {
			ptr = string_to_html(fl->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->host != NULL) {
			ptr = string_to_html(fl->host, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Host</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->file != NULL) {
			ptr = string_to_html(fl->file, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">File</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (fl->template != NULL)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Template</td><td>%s</td></tr>\n", fl->template);

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"http://%s/config?section=filter&amp;dialog=show&amp;list=deny&amp;id=%d\">Edit</a> <a href=\"http://%s/config?section=filter&amp;action=delete&amp;list=deny&amp;id=%d\">Delete</a></td>\n", INTERFACEURL, fl->id, INTERFACEURL, fl->id);
	}

	filebuf_addf(filebuf, "</table></td></tr></table></body></html>\n");

	header = header_new();;
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);

	pthread_rwlock_unlock(&filter_list->lock);
}

void mime_check_show(CONNECTION * connection)
{
	char *ptr;
	struct MIME_LIST_LIST *ml;
	FILEBUF *filebuf;
	HEADER *header;

	filebuf = filebuf_new();

	ml = mime_check(mime_list, connection);

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>MIME filter match for %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_BG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><th align=\"center\" colspan=\"2\">MIME filter match for %s%s</th></tr>\n", connection->header->host, connection->header->file);

	if (ml == NULL)
		filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\">None</td></tr>\n");
	else {
		if (ml->comment != NULL) {
			ptr = string_to_html(ml->comment, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Comment</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ml->host != NULL) {
			ptr = string_to_html(ml->host, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Host</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ml->file != NULL) {
			ptr = string_to_html(ml->file, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">File</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ml->mime != NULL) {
			ptr = string_to_html(ml->mime, HTML_NEWLINES);
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Mimetype</td><td>%s</td></tr>\n", ptr);
			xfree(ptr);
		}

		if (ml->template != NULL)
			filebuf_addf(filebuf, "<tr><td width=\"10%%\">Template</td><td>%s</td></tr>\n", ml->template);

		filebuf_addf(filebuf, "<tr><td colspan=\"2\" align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
		filebuf_addf(filebuf, "<tr><td align=\"center\"><a href=\"http://%s/config?section=mime&amp;dialog=show&amp;list=deny&amp;id=%d\">Edit</a> <a href=\"http://%s/config?section=mime&amp;action=delete&amp;list=deny&amp;id=%d\">Delete</a></td>\n", INTERFACEURL, ml->id, INTERFACEURL, ml->id);
	}

	filebuf_addf(filebuf, "</table></td></tr></table></body></html>\n");

	header = header_new();;
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);

	pthread_rwlock_unlock(&mime_list->lock);
}

void score_show(CONNECTION * connection, int score)
{
	FILEBUF *filebuf;
	HEADER *header;

	filebuf = filebuf_new();

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Keyword score for %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_BG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><th align=\"center\">Keyword score for %s%s</th></tr>\n", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<tr><td align=\"center\">%d</td></tr>\n", score);
	filebuf_addf(filebuf, "</table></td></tr></table></body></html>\n");

	header = header_new();;
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);
}

void cache_show(CONNECTION * connection)
{
	char buf[4096], *ptr;
	FILEBUF *filebuf;
	HEADER *header;
	CACHEMAP *cachemap;
	struct tm t;

	filebuf = filebuf_new();

	snprintf(buf, sizeof(buf), "%s://%s:%d%s", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);
	cachemap = cache_open(NULL, global->cache, buf, NULL, CACHE_OFFLINE);

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Cache information for %s://%s:%d%s</title></head>", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_BG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\" width=\"100%%\"><tr><td align=\"center\">\n");
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><th align=\"center\" colspan=\"2\">Cache information for %s://%s:%d%s</th></tr>\n", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);

	pthread_mutex_lock(&global->cache->lock);

	if (cachemap == NULL)
		filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\">Nothing found</td></tr>\n");
	else {
		if (cachemap->store != NULL)
			filebuf_addf(filebuf, "<tr><td>Path</td><td>%s</td></tr>\n", cachemap->store->path);

		filebuf_addf(filebuf, "<tr><td>Size</td><td>%d bytes</td></tr>\n", cachemap->size);

		localtime_r(&cachemap->mtime, &t);
		strftime(buf, sizeof(buf), HTTPTIMEFORMAT, &t);

		filebuf_addf(filebuf, "<tr><td>Retrieved</td><td>%s</td></tr>\n", buf);

		if (cachemap->lmtime != 0) {
			localtime_r(&cachemap->lmtime, &t);
			strftime(buf, sizeof(buf), HTTPTIMEFORMAT, &t);
			filebuf_addf(filebuf, "<tr><td>Last modified</td><td>%s</td></tr>\n", buf);
		}

		if (cachemap->etime != 0) {
			localtime_r(&cachemap->etime, &t);
			strftime(buf, sizeof(buf), HTTPTIMEFORMAT, &t);
			filebuf_addf(filebuf, "<tr><td>Expires</td><td>%s</td></tr>\n", buf);
		}

		if (cachemap->ftime != 0) {
			localtime_r(&cachemap->ftime, &t);
			strftime(buf, sizeof(buf), HTTPTIMEFORMAT, &t);
			filebuf_addf(filebuf, "<tr><td>Fresh until</td><td>%s</td></tr>\n", buf);
		}


		ptr = string_to_html(cachemap->header->raw, 0);
		filebuf_addf(filebuf, "<tr><td colspan=\"2\"><font size=\"3\"><b>Header:</b></font>\n<pre>%s</pre></td></tr>\n", ptr);
		xfree(ptr);
	}

	filebuf_addf(filebuf, "</table></td></tr></table></body></html>\n");

	pthread_mutex_unlock(&global->cache->lock);

	if (cachemap != NULL)
		cache_close(global->cache, cachemap);

	header = header_new();;
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);
}


void show_filebuf_diff(CONNECTION * connection, FILEBUF * orig, FILEBUF * new)
{
	int oline = 0, nline = 0, matches;
	char *ptr;
	FILEBUF *filebuf;
	HEADER *header;
	STACK *ostack = NULL, *ostackstart = NULL, *nstack = NULL, *nstackstart = NULL, *otmpstack = NULL, *ntmpstack = NULL, *tmpstack, *tmpstack2;

	filebuf = filebuf_new();

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Rewrite diff for %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_TABLEBG, INTERFACE_LINK, INTERFACE_VLINK);

	/* load both FILEBUF's into a linked list where each node is a seperate line */
	ostackstart = filebuf_to_stack(orig);
	nstackstart = filebuf_to_stack(new);

	ostack = ostackstart;
	nstack = nstackstart;

	while (ostack != NULL && nstack != NULL) {
		for (; ostack != NULL && nstack != NULL && !strcmp(ostack->data, nstack->data); ostack = ostack->next, nstack = nstack->next, oline++, nline++);
		/* no differences */
		if (ostack == NULL && nstack == NULL)
			break;

		/* find end of changed area by searching for first DIFF_HUNK lines which are the same in both lists */
		for (matches = 0, otmpstack = ostack, ntmpstack = NULL; otmpstack != NULL; otmpstack = otmpstack->next) {
			for (ntmpstack = nstack; ntmpstack != NULL; ntmpstack = ntmpstack->next) {
				for (matches = 0, tmpstack = otmpstack, tmpstack2 = ntmpstack; tmpstack != NULL && tmpstack2 != NULL && matches < DIFF_HUNK && !strcmp(tmpstack->data, tmpstack2->data); tmpstack = tmpstack->next, tmpstack2 = tmpstack2->next, matches++);
				if (matches >= DIFF_HUNK)
					break;
			}
			if (matches >= DIFF_HUNK)
				break;
		}

		filebuf_addf(filebuf, "<b>Line %d:</b><br>\n", oline);

		for (; ostack != NULL && (ostack != otmpstack || otmpstack == NULL); ostack = ostack->next, oline++) {
			/* show lines removed from original file */
			ptr = string_to_html(ostack->data, 0);
			filebuf_addf(filebuf, "<b>-  </b> %s<br>\n", ptr);
			xfree(ptr);
		}

		for (; nstack != NULL && (nstack != ntmpstack || ntmpstack == NULL); nstack = nstack->next, nline++) {
			/* show lines added to original file */
			ptr = string_to_html(nstack->data, 0);
			filebuf_addf(filebuf, "<b>+  </b> %s<br>\n", ptr);
			xfree(ptr);
		}

		filebuf_add(filebuf, "<br>", 4);
	}

	/* free up the linked lists */
	stack_free(ostackstart);
	stack_free(nstackstart);

	filebuf_addf(filebuf, "</body></html>\n");

	header = header_new();;
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);
}

void interface_page_global(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	int i;
	char *subsection = NULL;
	struct cgi_args_t *a;

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "subsection"))
			subsection = a->value;
	}

	if (subsection != NULL) {
		for (i = 0; global_action_pages[i].file != NULL; i++) {
			if (!strcasecmp(global_action_pages[i].file, subsection))
				global_action_pages[i].handler(filebuf, args, connection);
		}
	}

	filebuf_addf(filebuf, "<table cellspacing=\"0\" cellpadding=\"0\" border=\"0\" width=\"100%%\">\n");
	for (i = 0; global_pages[i].file != NULL; i++) {
		filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"30%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
		filebuf_addf(filebuf, "<tr><th>%s</th></tr></table></td></tr> <tr><td><br></td></tr>\n", global_pages[i].file);
		filebuf_addf(filebuf, "<tr><td>\n");
		global_pages[i].handler(filebuf, args, connection);
		filebuf_addf(filebuf, "</tr></td><td> <tr><td><br></td></tr>\n");
	}

	filebuf_addf(filebuf, "</table>");
}

void interface_page_global_general(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr = NULL;
	pthread_rwlock_rdlock(&global->general->lock);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"global\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"subsection\" value=\"general\">\n");
	filebuf_addf(filebuf, "<tr><td>Connection timeout</td><td><input type=\"text\" name=\"ctimeout\" value=\"%d\" size=\"20\"></td>\n", global->general->ctimeout);
	filebuf_addf(filebuf, "<tr><td>Header timeout</td><td><input type=\"text\" name=\"timeout\" value=\"%d\" size=\"20\"></td>\n", global->general->timeout);
	filebuf_addf(filebuf, "<tr><td>Keepalive timeout</td><td><input type=\"text\" name=\"keeptimeout\" value=\"%d\" size=\"20\"></td>\n", global->general->keeptimeout);
	filebuf_addf(filebuf, "<tr><td>Maximum buffer size</td><td><input type=\"text\" name=\"maxbuffer\" value=\"%d\" size=\"20\"></td>\n", global->general->maxbuffer);

	if (global->general->tempdir != NULL)
		ptr = string_to_html(global->general->tempdir, 0);
	filebuf_addf(filebuf, "<tr><td>Temporary directory</td><td><input type=\"text\" name=\"tempdir\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Web interface line length</td><td><input type=\"text\" name=\"linelength\" value=\"%d\" size=\"20\"></td>\n", global->general->linelength);
	if (global->general->cportrange != NULL)
		ptr = string_to_html(global->general->cportrange, 0);
	filebuf_addf(filebuf, "<tr><td>CONNECT ports</td><td><input type=\"text\" name=\"cportrange\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Connection pool size</td><td><input type=\"text\" name=\"poolsize\" value=\"%d\" size=\"20\"></td>\n", global->general->poolsize);
	filebuf_addf(filebuf, "<tr><td>Connection pool timeout</td><td><input type=\"text\" name=\"pooltimeout\" value=\"%d\" size=\"20\"></td>\n", global->general->pooltimeout);

	filebuf_addf(filebuf, "<tr><td>Always compress mimetype</td><td><input type=\"text\" name=\"encodemime\" value=\"%s\" size=\"80\"></td>\n", (global->general->encodemime != NULL) ? global->general->encodemime : "");
	filebuf_addf(filebuf, "<tr><td>Compress outgoing</td><td align=\"center\">Yes: <input type=\"radio\" name=\"compressout\" value=\"yes\" %s> No: <input type=\"radio\" name=\"compressout\" name=\"no\" %s></td></tr>\n", (global->general->compressout == TRUE) ? "checked" : "", (global->general->compressout == FALSE) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Compress incoming</td><td align=\"center\">Yes: <input type=\"radio\" name=\"compressin\" value=\"yes\" %s> No: <input type=\"radio\" name=\"compressin\" name=\"no\" %s></td></tr>\n", (global->general->compressin == TRUE) ? "checked" : "", (global->general->compressin == FALSE) ? "checked" : "");


	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr>\n");

	pthread_rwlock_unlock(&global->general->lock);
}

void interface_page_global_general_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	struct cgi_args_t *a;

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "ctimeout"))
			global_general_insert(a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "timeout"))
			global_general_insert(NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "keeptimeout"))
			global_general_insert(NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "maxbuffer"))
			global_general_insert(NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "tempdir"))
			global_general_insert(NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "linelength"))
			global_general_insert(NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "cportrange"))
			global_general_insert(NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "poolsize"))
			global_general_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL, NULL);
		else if (!strcasecmp(a->name, "pooltimeout"))
			global_general_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value, NULL);
		else if (!strcasecmp(a->name, "encodemime"))
			global_general_insert(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, a->value);
		else if (!strcasecmp(a->name, "compressout")) {
			pthread_rwlock_wrlock(&global->general->lock);
			if (!strcasecmp(a->value, "yes"))
				global->general->compressout = TRUE;
			else
				global->general->compressout = FALSE;
			pthread_rwlock_unlock(&global->general->lock);
		} else if (!strcasecmp(a->name, "compressin")) {
			pthread_rwlock_wrlock(&global->general->lock);
			if (!strcasecmp(a->value, "yes"))
				global->general->compressin = TRUE;
			else
				global->general->compressin = FALSE;
			pthread_rwlock_unlock(&global->general->lock);
		}
	}
}

void interface_page_global_ftp(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr = NULL;

	pthread_rwlock_rdlock(&global->ftp->lock);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"global\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"subsection\" value=\"ftp\">\n");
	filebuf_addf(filebuf, "<tr><td>Passive mode</td><td><input type=\"checkbox\" name=\"passive\" %s></td>\n", (global->ftp->pasv == TRUE) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Timeout</td><td><input type=\"text\" name=\"timeout\" value=\"%d\" size=\"20\"></td>\n", global->ftp->timeout);

	if (global->ftp->anonlogin != NULL)
		ptr = string_to_html(global->ftp->anonlogin, 0);
	filebuf_addf(filebuf, "<tr><td>Anonymous login</td><td><input type=\"text\" name=\"anonlogin\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (global->ftp->anonpass != NULL)
		ptr = string_to_html(global->ftp->anonpass, 0);
	filebuf_addf(filebuf, "<tr><td>Anonymous password</td><td><input type=\"text\" name=\"anonpass\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td>Sort order</td><td>Ascending: <input type=\"radio\" name=\"sortorder\" value=\"ascending\" %s> Descending: <input type=\"radio\" name=\"sortorder\" value=\"descending\" %s></td></tr>\n", (global->ftp->sortorder == SORT_ASCENDING) ? "checked" : "", (global->ftp->sortorder == SORT_DESCENDING) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td>Sort field</td><td>None: <input type=\"radio\" name=\"sortfield\" %s> Name: <input type=\"radio\" name=\"sortfield\" value=\"name\" %s> Size: <input type=\"radio\" name=\"sortfield\" value=\"size\" %s> Date: <input type=\"radio\" name=\"sortfield\" value=\"date\" %s></td></tr>\n", (global->ftp->sortfield == SORT_NONE) ? "checked" : "", (global->ftp->sortfield == SORT_NAME) ? "checked" : "", (global->ftp->sortfield == SORT_SIZE) ? "checked" : "", (global->ftp->sortfield == SORT_DATE) ? "checked" : "");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr>\n");

	pthread_rwlock_unlock(&global->ftp->lock);
}

void interface_page_global_ftp_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	struct cgi_args_t *a;

	global->ftp->pasv = FALSE;

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "passive")) {
			pthread_rwlock_wrlock(&global->ftp->lock);
			global->ftp->pasv = TRUE;
			pthread_rwlock_unlock(&global->ftp->lock);
		} else if (!strcasecmp(a->name, "anonlogin"))
			global_ftp_insert(a->value, NULL, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "anonpass"))
			global_ftp_insert(NULL, a->value, NULL, NULL, NULL);
		else if (!strcasecmp(a->name, "timeout"))
			global_ftp_insert(NULL, NULL, a->value, NULL, NULL);
		else if (!strcasecmp(a->name, "sortorder"))
			global_ftp_insert(NULL, NULL, NULL, a->value, NULL);
		else if (!strcasecmp(a->name, "sortfield"))
			global_ftp_insert(NULL, NULL, NULL, NULL, a->value);
	}
}

void interface_page_global_dnsbl(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr = NULL;

	pthread_rwlock_rdlock(&global->dnsbl->lock);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"global\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"subsection\" value=\"dnsbl\">\n");

	if (global->dnsbl->template != NULL)
		ptr = string_to_html(global->dnsbl->template, 0);
	filebuf_addf(filebuf, "<tr><td>Template</td><td><input type=\"text\" name=\"template\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (global->dnsbl->domain != NULL)
		ptr = string_to_html(global->dnsbl->domain, 0);
	filebuf_addf(filebuf, "<tr><td>Domain</td><td><input type=\"text\" name=\"domain\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	if (global->dnsbl->badips != NULL)
		ptr = array_merge(global->dnsbl->badips, ',');
	filebuf_addf(filebuf, "<tr><td>Blocked IP addresses</td><td><input type=\"text\" name=\"badips\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr>\n");

	pthread_rwlock_unlock(&global->dnsbl->lock);
}

void interface_page_global_dnsbl_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	struct cgi_args_t *a;

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "template"))
			global_dnsbl_insert(a->value, NULL, NULL);
		else if (!strcasecmp(a->name, "domain"))
			global_dnsbl_insert(NULL, a->value, NULL);
		else if (!strcasecmp(a->name, "badips"))
			global_dnsbl_insert(NULL, NULL, a->value);
	}
}

void interface_page_global_prefetch(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	char *ptr = NULL;

	pthread_mutex_lock(&global->prefetch->lock);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellspacing=\"0\" cellpadding=\"0\" border=\"1\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<form action=\"config\" method=\"POST\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"section\" value=\"global\">\n");
	filebuf_addf(filebuf, "<input type=\"hidden\" name=\"subsection\" value=\"prefetch\">\n");

	filebuf_addf(filebuf, "<tr><td>Threads</td><td><input type=\"text\" name=\"threads\" value=\"%d\" size=\"20\"></td>\n", global->prefetch->threads);
	filebuf_addf(filebuf, "<tr><td>Queue size</td><td><input type=\"text\" name=\"queuesize\" value=\"%d\" size=\"20\"></td>\n", global->prefetch->queuesize);

	if (global->prefetch->tags != NULL)
		ptr = array_merge(global->prefetch->tags, ',');
	filebuf_addf(filebuf, "<tr><td>Tags</td><td><input type=\"text\" name=\"tags\" value=\"%s\" size=\"80\"></td>\n", (ptr != NULL) ? ptr : "");
	FREE_AND_NULL(ptr);

	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"2\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form></table></td></tr>\n");

	pthread_mutex_unlock(&global->prefetch->lock);
}

void interface_page_global_prefetch_action(FILEBUF * filebuf, struct cgi_args_t *args, CONNECTION * connection)
{
	struct cgi_args_t *a;

	for (a = args; a; a = a->next) {
		if (!strcasecmp(a->name, "threads"))
			global_prefetch_insert(a->value, NULL, NULL);
		else if (!strcasecmp(a->name, "queuesize"))
			global_prefetch_insert(NULL, a->value, NULL);
		else if (!strcasecmp(a->name, "tags"))
			global_prefetch_insert(NULL, NULL, a->value);
	}
}

void show_dirlist(CONNECTION * connection, struct dir_t *dir)
{
	char *ptr, buf[128];
	FILEBUF *filebuf = filebuf_new();
	HEADER *header = NULL;

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Directory listing for ftp://%s:%d%s</title></head>", connection->header->host, connection->header->port, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_BG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"0\" width=\"100%%\">\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"80%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	ptr = path_to_url(connection->header->file);
	filebuf_addf(filebuf, "<tr><td align=\"center\"><font size=\"5\">%s</font></td></tr></table></td></tr><tr><td><br></td></tr>\n", ptr);
	xfree(ptr);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr>");
	filebuf_addf(filebuf, "<td></td>\n");
	filebuf_addf(filebuf, "<td align=\"center\">");
	filebuf_addf(filebuf, "<a href=\"?order=descending&field=name\"><img src=\"http://%s/template/uparrow\" border=\"0\"></a>", INTERFACEURL);
	filebuf_addf(filebuf, "<a href=\"?order=ascending&field=name\"><img src=\"http://%s/template/downarrow\" border=\"0\"></a>", INTERFACEURL);
	filebuf_addf(filebuf, "</td>");
	filebuf_addf(filebuf, "<td align=\"center\">");
	filebuf_addf(filebuf, "<a href=\"?order=descending&field=size\"><img src=\"http://%s/template/uparrow\" border=\"0\"></a>", INTERFACEURL);
	filebuf_addf(filebuf, "<a href=\"?order=ascending&field=size\"><img src=\"http://%s/template/downarrow\" border=\"0\"></a>", INTERFACEURL);
	filebuf_addf(filebuf, "</td>");
	filebuf_addf(filebuf, "<td align=\"center\">");
	filebuf_addf(filebuf, "<a href=\"?order=descending&field=date\"><img src=\"http://%s/template/uparrow\" border=\"0\"></a>", INTERFACEURL);
	filebuf_addf(filebuf, "<a href=\"?order=ascending&field=date\"><img src=\"http://%s/template/downarrow\" border=\"0\"></a>", INTERFACEURL);
	filebuf_addf(filebuf, "</td>");
	filebuf_addf(filebuf, "</tr>");
	filebuf_addf(filebuf, "<form action=\"%s\" method=\"GET\">\n", connection->header->file);

	for (; dir; dir = dir->next) {
		filebuf_addf(filebuf, "<tr>");
		filebuf_addf(filebuf, "<td align=\"center\"><input type=\"checkbox\" name=\"filename\" value=\"%s\"></td>\n", dir->name);

		ptr = string_to_html(dir->name, 0);
		filebuf_addf(filebuf, "<td><a href=\"%s\">%s</a></td>", ptr, ptr);
		xfree(ptr);

		if (dir->isdir == TRUE)
			filebuf_addf(filebuf, "<td>DIR</td>");
		else {
			ptr = filesize(dir->size);
			filebuf_addf(filebuf, "<td>%s</td>", ptr);
			xfree(ptr);
		}

		strftime(buf, sizeof(buf), HTTPTIMEFORMAT, &dir->time);
		filebuf_addf(filebuf, "<td>%s</td>", buf);
		filebuf_addf(filebuf, "</tr>\n");
	}

	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"4\">Delete: <input type=\"radio\" name=\"action\" value=\"delete\"> Rename: <input type=\"radio\" name=\"action\" value=\"rename\"> Mkdir: <input type=\"radio\" name=\"action\" value=\"mkdir\"> Rmdir: <input type=\"radio\" name=\"action\" value=\"rmdir\"> Regexp match: <input type=\"radio\" name=\"action\" value=\"rematch\"> Raw command: <input type=\"radio\" name=\"action\" value=\"raw\"></td></tr>\n");
	filebuf_addf(filebuf, "<tr><td align=\"center\" colspan=\"4\">Argument: <input type=\"text\" name=\"argument\" size=\"80\"><input type=\"submit\" value=\"Submit\"></td></tr>\n");
	filebuf_addf(filebuf, "</form>\n");
	filebuf_addf(filebuf, "<form enctype=\"multipart/form-data\" action=\"%s\" method=\"POST\">\n", connection->header->file);
	filebuf_addf(filebuf, "<tr><td colspan=\"4\" align=\"center\">Upload file: <input type=\"file\" name=\"upload\"><input type=\"submit\" value=\"Upload\"></td></tr>\n");
	filebuf_addf(filebuf, "</form>\n");
	filebuf_addf(filebuf, "</td></tr></table></table></body></html>\n");

	header = header_new();;
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);
}

void show_headers(CONNECTION * connection)
{
	char *ptr;
	FILEBUF *filebuf = filebuf_new();
	HEADER *header = NULL;

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>HTTP headers to/from %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_TABLEBG, INTERFACE_LINK, INTERFACE_VLINK);

	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);
	filebuf_addf(filebuf, "<tr><td align=\"center\"><b>Client header</b></td></tr>\n");

	ptr = string_to_html(connection->header->raw, 0);
	filebuf_addf(filebuf, "<tr><td><pre>%s</pre></td></tr>\n", ptr);
	xfree(ptr);

	filebuf_addf(filebuf, "<tr><td align=\"center\"><b>Server header</b></td></tr>\n");
	ptr = string_to_html(connection->rheader->raw, 0);
	filebuf_addf(filebuf, "<tr><td><pre>%s</pre></td></tr>\n", ptr);
	xfree(ptr);

	filebuf_addf(filebuf, "</table></body></html>\n");

	header = header_new();;
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);
}

void show_filebuf(CONNECTION * connection, FILEBUF * fb)
{
	char *ptr;
	FILEBUF *filebuf;
	HEADER *header;

	filebuf = filebuf_new();
	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Raw contents of %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_TABLEBG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_add(fb, "", 1);

	ptr = string_to_html(fb->data, 0);
	filebuf_addf(filebuf, "<pre>\n");
	filebuf_add(filebuf, ptr, strlen(ptr));
	filebuf_addf(filebuf, "</pre>\n");
	xfree(ptr);

	filebuf_resize(fb, fb->size - 1);
	filebuf_addf(filebuf, "</body></html>\n");

	header = header_new();
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);
}

void show_profiles(CONNECTION * connection)
{
	char **args;
	FILEBUF *filebuf;
	HEADER *header;

	filebuf = filebuf_new();
	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Active profiles for %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_TABLEBG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	args = connection->profiles;
	for (; args && *args; args++)
		filebuf_addf(filebuf, "<tr><td>%s</td></tr>\n", *args);

	filebuf_addf(filebuf, "</table></body></html>\n");

	header = header_new();
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	filebuf_free(filebuf);
}

void show_proxytest(CONNECTION * connection)
{
	char *ptr;
	HEADER *header;
	FILEBUF *filebuf;

	filebuf = filebuf_new();

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>Proxy test for %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_TABLEBG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	if (connection->proxy_type != PROXY_NORMAL)
		filebuf_addf(filebuf, "<tr><td align=\"center\">Not forwarding through HTTP proxy</td></tr>\n");
	else {
		proxy_test(connection);

		if (connection->rheader == NULL)
			filebuf_addf(filebuf, "<tr><td align=\"center\">%s:%d failed to connect back</td></tr>\n", connection->proxy_host, connection->proxy_port);
		else {
			filebuf_addf(filebuf, "<tr><td align=\"center\">Headers received from %s:%d</td></tr>\n", connection->proxy_host, connection->proxy_port);
			ptr = string_to_html(connection->rheader->raw, 0);
			filebuf_addf(filebuf, "<tr><td><pre>%s</pre></td></tr>\n", ptr);
			xfree(ptr);
		}
	}

	filebuf_addf(filebuf, "</table></body></html>\n");

	header = header_new();
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);
	http_header_free(connection->rheader);

	filebuf_free(filebuf);
}

void show_htmltree(CONNECTION * connection)
{
	FILEBUF *filebuf;
	HEADER *header;

	filebuf = filebuf_new();

	filebuf_addf(filebuf, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
	filebuf_addf(filebuf, "<html><head><title>HTML tree for %s%s</title></head>", connection->header->host, connection->header->file);
	filebuf_addf(filebuf, "<body text=\"%s\" bgcolor=\"%s\" link=\"%s\" vlink=\"%s\">\n", INTERFACE_TEXT, INTERFACE_TABLEBG, INTERFACE_LINK, INTERFACE_VLINK);
	filebuf_addf(filebuf, "<table cellpadding=\"0\" cellspacing=\"0\" border=\"1\" width=\"100%%\" bgcolor=\"%s\" bordercolor=\"%s\">\n", INTERFACE_TABLEBG, INTERFACE_TABLEBORDER);

	show_htmltree_node(filebuf, connection->htmlstream->tree, 0);

	filebuf_addf(filebuf, "</table></body></html>\n");

	header = header_new();
	header->type = HTTP_RESP;
	header->code = 200;
	header->content_type = xstrdup("text/html");
	header->content_length = filebuf->size;

	header_send(header, connection, CLIENT, HEADER_RESP);
	net_filebuf_send(filebuf, connection, CLIENT);

	http_header_free(header);

	filebuf_free(filebuf);
}

void show_htmltree_node(FILEBUF * filebuf, struct htmlstream_node *node, int indent)
{
	char *ptr;
	struct htmlstream_node *n;
	struct htmlstream_tag_property *p;

	for (n = node; n; n = n->next) {
		filebuf_addf(filebuf, "<tr><td>", indent);
		filebuf_addf(filebuf, "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">");

		filebuf_addf(filebuf, "<tr><td width=\"%d\"></td><td>", indent * 36);

		if (n->tag != NULL) {
			filebuf_addf(filebuf, "%s ", n->tag->name);

			for (p = n->tag->properties; p; p = p->next)
				filebuf_addf(filebuf, "%s=\"%s\" ", p->name, p->value);
		} else if (n->text != NULL) {
			ptr = string_to_html(n->text, HTML_NEWLINES | HTML_SPACES);
			filebuf_addf(filebuf, "\"%s\"", ptr);
			xfree(ptr);
		}

		filebuf_addf(filebuf, "</td></tr></table></td></tr>\n");

		if (n->up != NULL)
			show_htmltree_node(filebuf, n->up, indent + 1);
	}
}
